/*
 * Copyright (C) 2008 Dell Inc.
 * Chieng-Hwa Lin <chieng-hwa_lin@dell.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*--------------------------------------------------------------------
 * Linux include files
 *--------------------------------------------------------------------*/
#include <linux/version.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/timer.h>
#include <linux/mtio.h>
#include <linux/ioctl.h>
#include <linux/fcntl.h>
// 2.6.24
#include <linux/interrupt.h>
#include <linux/vmalloc.h>
#include <linux/poll.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <../arch/ppc/platforms/4xx/sequoia.h>
/*
// chl chl
//#include <linux/blk.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#include <linux/vmalloc.h>
#include <linux/spinlock.h>
#include <linux/wait.h>
#endif
*/
/*--------------------------------------------------------------------
 * Dell include files
 *--------------------------------------------------------------------*/
//#include "./include/gpiotypes.h"
//#include "./include/gpio_struct.h"
#include <gpiotypes.h>
#include <gpio_struct.h>
#include <dcgpio.h>
#include <cmc_gpio.h>
MODULE_LICENSE("GPL");

/*--------------------------------------------------------------------
 * Local Variables used for this module
 *--------------------------------------------------------------------*/

//--------------------------------------------------------------------
// Functions Prototypes
//--------------------------------------------------------------------

int init_device_driver(void);
void GpioUnload(PGPIO_DEVICE_EXTENSION pGPIOCard);
s32 GpioInitDriver(PGPIO_DEVICE_EXTENSION pGPIOCard);
int GpioDeviceIoControl(struct inode * inode,
        struct file * filp,
        PGPIO_DEVICE_EXTENSION pGPIOCard,
        unsigned int cmd_in,
        unsigned long arg,
        int *lpBytesReturned);

extern int  driver_exited;

extern irqreturn_t GpioIsr( int irq, void *dev_id, struct pt_regs *regs);
u32 XIic_check_and_heal_stuck_scl(int bus_index);
u32 XIic_check_and_disable_i2c_bus(int bus_index);
u32 PPC_check_and_heal_stuck_scl(int bus_index);
u32 XIic_check_and_heal_stuck_bus_reset(int bus_index);
u32 PPC_check_and_heal_stuck_bus_reset(int bus_index);
s32 GpioLoadDevInterrupt(PGPIO_DEVICE_EXTENSION pGpioCard);
void GpioUnloadDevInterrupt(PGPIO_DEVICE_EXTENSION pGpioCard);
int PPC_check_and_heal_stuck_sda(int BusIndex);
int XIic_check_and_heal_stuck_sda(int BusIndex);

//--------------------------------------------------------------------
// Global data
//--------------------------------------------------------------------
short fan_fifo[FAN_FIFOSIZE];
short shifty_fifo[SHIFTY_FIFOSIZE];
short kvm_fifo[KVM_FIFOSIZE];
short blade_fifo[BLADE_FIFOSIZE];
short iom_fifo[IOM_FIFOSIZE];
short psu_fifo[PSU_FIFOSIZE];
short cp_fifo[CP_FIFOSIZE];
struct intr_status_t blade_intr_queue;
struct intr_status_t fan_intr_queue;
struct intr_status_t iom_intr_queue;
struct psu_intr_status_t psu_intr_queue;





struct intr_status_t kvm_intr_queue;
struct intr_status_t shifty_intr_queue;
struct intr_status_t cp_intr_queue;

int RecoveryThreadExitingFlag = FALSE;
GPIO_DEVICE_EXTENSION gpio_card;
typedef struct _dcd_thread_data
{
	// Thread ID #
	pid_t ID;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,7)
	// completion pointer
	struct completion compl;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	wait_queue_head_t compl;  // Sempahore for I/O completions
#else
	struct wait_queue *compl;  // Sempahore for watch dog thread signalling
#endif// #if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,7)
	// Pointer to name
	void *pName;
}dcd_thread_data;
extern wait_queue_head_t XilinxRecoverySemaphore;  // Sempahore for I2C1 recovery
extern int xilinx_recovery_event_present;
extern wait_queue_head_t XilinxDisableSemaphore;  // Sempahore for I2C1 Disable
extern int xilinx_disable_event_present;
//--------------------------------------------------------------------
// Driver error messages
//--------------------------------------------------------------------

char EMsgDevNotFound[] = "\r\nDCGPIO driver could not find the device.\r\n";

int gpio_major = 253;

int  
gpio_i2c_dbg(
    const char *myformat,
    ...
    )
{
#ifdef CMC_DRIVER_DBG
    va_list args;
    int     rc=0;

    va_start(args, myformat);
    rc = vprintk(myformat, args);
    va_end(args);
    return(rc);
#else
    return(0);
#endif
} /* end ppc_i2c_dbg */

static int gpio_open(struct inode * inode, struct file * filp)
{


    return 0;
}

/* Following function was formerly called 'gpio_close' */
static int gpio_release(struct inode * inode, struct file * filp)
{

    return 0;
}


static int gpio_ioctl(struct inode * inode, struct file * filp,
                    unsigned int cmd_in, unsigned long arg)
{
    int	IoCmdStatus = 0; 
    u32 BytesReturned;
    GPIO_DEVICE_EXTENSION *pGPIOCard = &gpio_card;

    pGPIOCard->GPIOFirstIO = 1;

    IoCmdStatus = GpioDeviceIoControl( inode, filp, pGPIOCard, cmd_in, arg,
		&BytesReturned);

    return (IoCmdStatus);
}

static struct file_operations gpio_fops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	    owner: 		THIS_MODULE,
	    ioctl:		gpio_ioctl,
	    open:		gpio_open,
	    release:		gpio_release,
#else
	    NULL,            /* lseek */
	    NULL,            /* read */
	    NULL,            /* write */
	    NULL,            /* readdir */
	    NULL,            /* poll */
	    gpio_ioctl,       /* ioctl */
	    NULL,            /* mmap */
	    gpio_open,        /* open */
	    NULL,            /* flush */
	    gpio_release,     /* release, was formerly sg_close */
	    NULL,            /* fsync */
	    NULL,            /* fasync */
	    NULL,            /* check_media_change */
	    NULL,            /* revalidate */
	    NULL,            /* lock */
#endif
	};

/* Driver initialization */
static int __init
gpio_init(void)
{
	static int gpio_registered = 0;
	int result, rc;

	if(!gpio_registered) 
	{
		result = register_chrdev(gpio_major,"gpio",&gpio_fops);
		if (result < 0)
		{
			printk("Unable to get major %d for Dell RAC GPIO device\n",gpio_major);
			return 1;
		}
		if (gpio_major == 0) gpio_major = result; /* dynamic */
		{
			gpio_registered++;
		}
	}

	rc = init_device_driver();
	if (rc) 
	{
		printk("Unable to initialize the CMC GPIO device driver, major number: %d\n",gpio_major);
		unregister_chrdev(gpio_major, "gpio");
		return 1;
	}

	return 0;
}

/* Inititialize the driver and ESM card */
int init_device_driver()
{
	PGPIO_DEVICE_EXTENSION pGPIOCard;

	/*  return the input request packet as the output request packet with */

	/* clear the card structure first */
	memset(&gpio_card, 0, sizeof(GPIO_DEVICE_EXTENSION));

	// Point to GPIO Driver Structure
	pGPIOCard = &gpio_card;

	// initialize driver structure and device
	if (GpioInitDriver(pGPIOCard) != GPIO_STATUS_CMD_SUCCESS)
	{
		return GPIO_STATUS_CMD_NO_SUCH_DEVICE;
	}

	return (GPIO_STATUS_CMD_SUCCESS);
} /* init_device_driver */

//---------------------------------------------------------------------
// GpioUnload - stop processing ESM Commands; stop monitoring; stop watchdog
//---------------------------------------------------------------------
void GpioUnload(PGPIO_DEVICE_EXTENSION pGPIOCard)
{

#ifdef GPIO_WITH_INTERRUPTS
        GpioUnloadDevInterrupt(pGPIOCard);
#endif
	unregister_chrdev(gpio_major, "gpio");


} // GpioUnload 

//---------------------------------------------------------------------
// GpioDDStart
//---------------------------------------------------------------------
s32 GpioDDStart(PGPIO_DEVICE_EXTENSION pGpioCard)
{


//	DrvDbgMsg3(("GpioDDStart: entry\n"));

	return GPIO_STATUS_CMD_SUCCESS;
} // GpioDDStart

//--------------------------------------------------------------------
// gpio_complete_thread
//--------------------------------------------------------------------
void gpio_complete_thread(void* pThreadData)
{
	dcd_thread_data *pThread = (dcd_thread_data *)pThreadData;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,7)
	complete(
		&pThread->compl
		);
#else
		wake_up(
			&pThread->compl
			);

#endif// #if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,7)
}// gpio_complete_thread
//--------------------------------------------------------------------
// gpio_kernel_thread
//--------------------------------------------------------------------
void* gpio_kernel_thread(int (*fn)(void *), 
						void *arg,
						void *pThreadName)
{

	dcd_thread_data* pThreadData;

	// Allocate thread data space...
	pThreadData = vmalloc(
		sizeof(dcd_thread_data)
		);

	if (pThreadData == NULL )
	{
		// failed to allocate memory
		printk(
			"gpio_kernel_thread: failed to allocate memory\n"
			);
		return NULL;
	}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,7)
	// Initialize the thread data...
	init_completion(
		&pThreadData->compl
		);
	
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
        init_waitqueue_head(
			&pThreadData->compl
			);
#else
        // catch GPFs
        pThreadData->compl = NULL;
#endif
	// create kernel thread...
	pThreadData->ID = kernel_thread(
		fn, 
		arg, 
		(CLONE_FS | CLONE_FILES | CLONE_SIGHAND)
		);

	if ( pThreadData->ID < 0)
	{
		// failed to create thread.
		// free the thread data.
		vfree (
			pThreadData
			);

		printk(
			"gpio_kernel_thread: failed to create thread\n"
			);

		return NULL;
	}

	// store kernel thread name...
	pThreadData->pName = pThreadName;

	// return pointer to the thread structure.
	return (void *)pThreadData;
}// gpio_kernel_thread

void Xiic_mask_interrupts(int BusIndex)
{
	u32 GpioCntlReg;
	// Mask the interrupt for the specific bus
	switch (BusIndex)
	{
		case 0:
		// IOM1
			CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, 0x0462, 0xfc3f);
			CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, 0x0462, &GpioCntlReg);
		break;
		case 1:
		// IOM2
			CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, 0x0462, 0xfc3f); 	
		break;
		case 2:
		// PSU 1
			CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, 0x0442, 0xfc3f); 	
		break;
		case 3:
		// PSU 2
			CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, 0x0442, 0xfc3f); 	
		break;
		case 4:
		// KVM
			CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, 0x04A2, 0xe001); 	
		break;
		default:
		break;
	}
}

void Xiic_unmask_interrupts(int BusIndex)
{
	u32 GpioCntlReg;

	// Mask the interrupt for the specific bus
	switch (BusIndex)
	{
		case 0:
		// IOM1
			CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, 0x0462, &GpioCntlReg);
			GpioCntlReg &= ~(0xfc3f);
			CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, 0x0462, GpioCntlReg|DONT_ENABLE_SMBUS_TIMEOUT_FOR_IOMS);
		break;
		case 1:
		// IOM2
			CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, 0x0462, &GpioCntlReg);
			GpioCntlReg &= ~(0xfc3f);
			CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, 0x0462, GpioCntlReg|DONT_ENABLE_SMBUS_TIMEOUT_FOR_IOMS);
		break;
		case 2:
		// PSU 1
			CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, 0x0442, &GpioCntlReg);
			GpioCntlReg &= ~(0xfc3f);
			CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, 0x0442, GpioCntlReg); 	
		break;
		case 3:
		// PSU 2
			CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, 0x0442, &GpioCntlReg);
			GpioCntlReg &= ~(0xfc3f);
			CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, 0x0442, GpioCntlReg); 	
		break;
		case 4:
		// KVM
			CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, 0x04a2, &GpioCntlReg);
			GpioCntlReg &= ~(0xe001);
			CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, 0x04A2, GpioCntlReg); 	
		break;
		default:
		break;
	}
}

/*****************************************************************************/
/**
*
* Unlocks the IIC device. Unravel must only be called after the driver has been
* initialized. The configuration after this reset is as follows:
*   - Repeated start is disabled
*   - General call is disabled
*
* The upper layer software is responsible for initializing and re-configuring
* (if necessary) and restarting the IIC device after the recovery.
*
* @param    InstancePtr is a pointer to the XIic instance to be worked on.
*
* @return
*
* None.
*
* @note
*
* None.
*
* @internal
*
* The I2C unravel clocks the respective I2Cbus.  This is accomplished by using
* the IIC_GPIO_CNTL register.  This takes enables the software control of the
* I2C core.  The I2C clock and data lines are directly controlled by this routine.
* The clocking uses busy waits since the entire task takes around 200 usecs.
*
****************************************************************************/
u32
PPC_Recover(int BusIndex)
{
	int status;

	status = PPC_check_and_heal_stuck_sda(BusIndex);
	printk(KERN_INFO "PPC_Recover: check PPC i2c bus, index %d, status 0x%x\n",BusIndex, status);

	/* hold the SDA low for > 35 msec to see if slave recovers */
	if (status != 0)
	{
		//printk(KERN_INFO "PPC_Recover: check and heal stuck SCL for index: %d\n", BusIndex);
		status = PPC_check_and_heal_stuck_scl(BusIndex);
		//printk(KERN_INFO "PPC_Recover: check and heal stuck SCL status: 0x%x\n", status);

		printk(KERN_INFO "PPC_Recover: reset PPC i2c bus, index %d\n", BusIndex);
   		status = PPC_check_and_heal_stuck_bus_reset(BusIndex);
	}

    return status;
}
/*****************************************************************************/
/**
*
* Unlocks the IIC device. Unravel must only be called after the driver has been
* initialized. The configuration after this reset is as follows:
*   - Repeated start is disabled
*   - General call is disabled
*
* The upper layer software is responsible for initializing and re-configuring
* (if necessary) and restarting the IIC device after the recovery.
*
* @param    InstancePtr is a pointer to the XIic instance to be worked on.
*
* @return
*
* None.
*
* @note
*
* None.
*
* @internal
*
* The I2C unravel clocks the respective I2Cbus.  This is accomplished by using
* the IIC_GPIO_CNTL register.  This takes enables the software control of the
* I2C core.  The I2C clock and data lines are directly controlled by this routine.
* The clocking uses busy waits since the entire task takes around 200 usecs.
*
****************************************************************************/
u32
XIic_Recover(int BusIndex)
{
	int I2CBusBaseAddressOffset;
	int RegOffset;
	int status;
	

	I2CBusBaseAddressOffset = 0x1600 + 0x200 * BusIndex;
	RegOffset = I2CBusBaseAddressOffset + 0x40;


    /* Now the the buffers are reset on the I2C core unravel the slaves. */
      /* can only clear busy AFTER the master port has been reset 
     * and hence a DONE interrupt cannot arrive 
     */
    Xiic_mask_interrupts(BusIndex);	
   
    gpio_i2c_dbg("XIic_Recovery, calling XIic_check_and_heal_stuck_sda.\n");
    gpio_i2c_dbg("I2CBusBaseAddressOffset: 0x%x\n", I2CBusBaseAddressOffset); 
    gpio_i2c_dbg("RegOffset: 0x%x\n", RegOffset); 
    status = XIic_check_and_heal_stuck_sda(BusIndex);

    /* hold the SDA low for > 35 msec to see if slave recovers */
    if (status != 0)
    {

    	status = XIic_check_and_heal_stuck_scl(BusIndex);
     	/*
     	* We hold the rest line low for > 100 msec as a last resort.
     	* (GPIO 59 IOMSET1, GPIO 60 IOMSET2)
     	* (GPIO 57 PSU123, GPIO 58 PSU456)
     	* (GPIO 61 KVM)
     	*/
	if (status != 0)

     		status = XIic_check_and_heal_stuck_bus_reset(BusIndex);
     } else
	status = XIic_check_and_heal_stuck_bus_reset(BusIndex);
    Xiic_unmask_interrupts(BusIndex);	

    return status;

}
/*****************************************************************************/
/**
*
* This function disable the I2C Bus for the index
* Currently this is enabled only for IOMs only
*
* @param    InstancePtr is a pointer to the XIic instance to be worked on.
*
* @return
*
* None.
*
* @note
*
* None. BITS067915:Disable IOM1 and IOM2 Bus before reset
*
* @internal
*
* This function should be called just before the cmc reboot operation
****************************************************************************/
u32
XIic_Disable(int BusIndex)
{
	int I2CBusBaseAddressOffset;
	int RegOffset;
	int status;
	

	I2CBusBaseAddressOffset = 0x1600 + 0x200 * BusIndex;
	RegOffset = I2CBusBaseAddressOffset + 0x40;


    /* Now the the buffers are reset on the I2C core unravel the slaves. */
    /* can only clear busy AFTER the master port has been reset 
     	* and hence a DONE interrupt cannot arrive 
     	*/
    Xiic_mask_interrupts(BusIndex);
	if (0 == BusIndex || 1 == BusIndex)
	{
		XIic_check_and_disable_i2c_bus(BusIndex);
	}

    return status;

}
/*******************************************************************************
* FUNCTION: PPC_check_and_heal_stuck_sda()
* ____________________________________________________________________________
*
* DESCRIPTION:  This function attempts to wake up TWI slave devices that
*               are holding SDA low.
*
* INPUTS:
*   base             The ID of the port to heal
*   timeout_int_status  The time-out interrupt status flags
*
* OUTPUTS:      
*   none
*
* RETURNS:
*   nothing
*
* NOTES:
*   Healing a stuck-SDA TWI bus is attempted by toggling SDA at 100kHz rates. 
*
*******************************************************************************/
int PPC_check_and_heal_stuck_sda(int BusIndex)
{
	int attempt;
	u32 clr_channel_reg;
	u32 GpioCntlReg;
	u32 I2CBusBaseAddressOffset;
	u32 I2CBusSoftResetAddress;
	u32 BusIsHung = 1;
	u32 status = 0;

	if (BusIndex == 5) 
	{
		I2CBusBaseAddressOffset = 0xef600710;
		I2CBusSoftResetAddress  = 0xef60070c;
	}
	else 
	{
		I2CBusBaseAddressOffset = 0xef600810;
		I2CBusSoftResetAddress  = 0xef60080c;
	}

	// IIC controller must be placed in the reset state
	CMCGpioReadReg32Unmapped(I2CBusSoftResetAddress, 0, &GpioCntlReg);
	GpioCntlReg |= 0x00000001;
  	CMCGpioWriteReg32Unmapped(I2CBusSoftResetAddress, 0, GpioCntlReg); 	

    if ((BusIsHung & 1) != 0)
	{
		CMCGpioReadReg32Unmapped(I2CBusBaseAddressOffset, 0, &GpioCntlReg);
        GpioCntlReg |= 0x0C000000; /* Make SCL and SDA high */
		CMCGpioWriteReg32Unmapped(I2CBusBaseAddressOffset, 0, GpioCntlReg); 	

		attempt = IIC_SDA_LOCKUP_RECOVERY_ITER; // try 27 times
		do
		{
	    	CMCGpioReadReg32Unmapped(I2CBusBaseAddressOffset, 0, &GpioCntlReg);
			GpioCntlReg |= 0x0C000000; /* Make SCL and SDA high */
	    	CMCGpioWriteReg32Unmapped(I2CBusBaseAddressOffset, 0, GpioCntlReg); 	

            udelay(IIC_SDA_LOCKUP_RECOVERY_US_DELAY);

	    	CMCGpioReadReg32Unmapped(I2CBusBaseAddressOffset, 0, &GpioCntlReg);
            GpioCntlReg &= ~0x04000000; /* Make SCL low */
            GpioCntlReg |= 0x08000000;  /* Make SDA high */
	    	CMCGpioWriteReg32Unmapped(I2CBusBaseAddressOffset, 0, GpioCntlReg); 	

            mdelay(35);	// 35 msec

	    	CMCGpioReadReg32Unmapped(I2CBusBaseAddressOffset, 0, &GpioCntlReg);
            GpioCntlReg |= 0x0C000000; /* Make SDA and SCL high */
	    	CMCGpioWriteReg32Unmapped(I2CBusBaseAddressOffset, 0, GpioCntlReg); 	

            udelay(IIC_SDA_LOCKUP_RECOVERY_US_DELAY);

	    	CMCGpioReadReg32Unmapped(I2CBusBaseAddressOffset, 0, &clr_channel_reg);
		}
		while((--attempt > 0) && ((clr_channel_reg & 0x02000000) == 0));
		if (clr_channel_reg & 0x02000000)
		{
			status = 0;
		}
		else
		{
			status = -1;
		}
	}

	// set IIC controller back to normal operation state
	CMCGpioReadReg32Unmapped(I2CBusSoftResetAddress, 0, &GpioCntlReg);
	GpioCntlReg &= ~(0x00000001);
  	CMCGpioWriteReg32Unmapped(I2CBusSoftResetAddress, 0, GpioCntlReg); 	

    return status;
}

/*******************************************************************************
* FUNCTION: XIic_check_and_heal_stuck_sda()
* ____________________________________________________________________________
*
* DESCRIPTION:  This function attempts to wake up TWI slave devices that
*               are holding SDA low.
*
* INPUTS:
*   base             The ID of the port to heal
*   timeout_int_status  The time-out interrupt status flags
*
* OUTPUTS:      
*   none
*
* RETURNS:
*   nothing
*
* NOTES:
*   Healing a stuck-SDA TWI bus is attempted by toggling SDA at 100kHz rates. 
*
*******************************************************************************/
int XIic_check_and_heal_stuck_sda(int BusIndex)
{
   int               attempt;
   u32   clr_channel_reg;
   u32   GpioCntlReg;
   u32 I2CBusBaseAddressOffset;
   u16 RegOffset;
   u32 BusIsHung = 1;
   u32 status = 0;

	
	I2CBusBaseAddressOffset = CMC_XILINX_I2C_BUS_0_OFFSET + 0x200 * BusIndex;
	RegOffset = I2CBusBaseAddressOffset + CMC_I2C_GPIO_CONTROL_OFFSET;
	CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, RegOffset, &GpioCntlReg);
	gpio_i2c_dbg("XIic_check_and_heal_stuck_sda,  reg 0x%x: 0x%x\n", RegOffset, GpioCntlReg);

    /*
     * Timeout can only be detected if the master port is enabled.
     * If port has SDA stuck low AND SCL stuck high: 
     *
     *  [a] go into clear channel mode 
     *  [b] toggle SCL.  
     *  [c] exit clear channel mode
     * 
     * Could also generate a stop condition, but this
     * is not implemented.
     */
    if ((BusIsHung & 1) != 0)
    {
	RegOffset = I2CBusBaseAddressOffset + CMC_I2C_GPIO_CONTROL_OFFSET;
	CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, RegOffset, &GpioCntlReg);
	gpio_i2c_dbg("XIic_check_and_heal_stuck_sda 1,  reg 0x%x: 0x%x\n", RegOffset, GpioCntlReg);
        GpioCntlReg &= ~IIC_SDA_T; /* un tri-state SDA */
        GpioCntlReg &= ~IIC_SCL_T; /* un tri-state SCL */
        GpioCntlReg |=  IIC_SDA_O | IIC_SCL_O; /* Make SCL and SDA high */
	gpio_i2c_dbg("XIic_check_and_heal_stuck_sda,  writing reg 0x%x with 0x%x\n", RegOffset, GpioCntlReg);
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg); 	

        GpioCntlReg |= IIC_EN_SOFT; /* Take over the bus */
	gpio_i2c_dbg("XIic_check_and_heal_stuck_sda,  taking over, writing reg 0x%x with 0x%x\n", RegOffset, GpioCntlReg);
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg); 	
        /*
         * make TWI_SDA_LOCKUP_RECOVERY_ITER attempts to get SDA to rise,
         * stop if SDA is already back up
         */
	// Try 27 times
        attempt = IIC_SDA_LOCKUP_RECOVERY_ITER;
        do
        {
	CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, RegOffset, &GpioCntlReg);
        GpioCntlReg &= ~IIC_SDA_T; /* un tri-state SDA */
        GpioCntlReg &= ~IIC_SCL_T; /* un tri-state SCL */
        GpioCntlReg |= IIC_SDA_O | IIC_SCL_O; /* Make SCL and SDA high */
	gpio_i2c_dbg("XIic_check_and_heal_stuck_sda,  SCL and SDA HI, writing reg 0x%x with 0x%x\n", RegOffset, GpioCntlReg);
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg); 	

        GpioCntlReg |= IIC_EN_SOFT; /* Take over the bus */
            /* drive SCL low */

	    CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, RegOffset, &GpioCntlReg);
            GpioCntlReg &= ~IIC_SCL_O;
            GpioCntlReg |= IIC_SDA_O | IIC_EN_SOFT;
	gpio_i2c_dbg("XIic_check_and_heal_stuck_sda,  SDA and ENABLE HI, writing reg 0x%x with 0x%x\n", RegOffset, GpioCntlReg);
	    // SDA is high, enable software is high
	    CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);

            /* delay */
		
            udelay(5);

            /* release SCL */
            GpioCntlReg |= IIC_EN_SOFT | IIC_SCL_O | IIC_SDA_O;
	gpio_i2c_dbg("XIic_check_and_heal_stuck_sda,  DATA, Clock and ENABLE HI, writing reg 0x%x with 0x%x\n", RegOffset, GpioCntlReg);
	    // Data is high, clock is high and enable software is high
	    CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);

            /* delay microseconds*/
            udelay(IIC_SDA_LOCKUP_RECOVERY_US_DELAY);

	    // Tri-state all the bus pins and enable software control
            GpioCntlReg |= IIC_SDA_T | IIC_SCL_T | IIC_SDA_O | IIC_SCL_O | IIC_EN_SOFT;
	gpio_i2c_dbg("XIic_check_and_heal_stuck_sda,  tri-state and ENABLE, writing reg 0x%x with 0x%x\n", RegOffset, GpioCntlReg);
	    CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);
            /* read SCL and SDA */
	    CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, RegOffset, &clr_channel_reg);

        }
        while((--attempt > 0) &&
              ((clr_channel_reg & IIC_SDA_O) == 0));
	if (clr_channel_reg & IIC_SDA_O)
	{
	gpio_i2c_dbg("XIic_check_and_heal_stuck_sda,  GOOD clr_channel_reg: 0x%x\n", clr_channel_reg);
		status = 0;
	}
	else
	{
	gpio_i2c_dbg("XIic_check_and_heal_stuck_sda,  BAD clr_channel_reg: 0x%x\n", clr_channel_reg);
		
		status = -1;
	}

        /* return to normal operating mode */
        GpioCntlReg |= IIC_SDA_T; /* Tri-state SDA */
        GpioCntlReg |= IIC_SCL_T; /* Tri-state SCL */
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);

        GpioCntlReg &= ~IIC_EN_SOFT; /* Release the bus */
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);
    }
    return status;
}
/*******************************************************************************
* FUNCTION: PPC_check_and_heal_stuck_scl()
* ____________________________________________________________________________
*
* DESCRIPTION:  This function attempts to wake up TWI slave devices that
*               are holding SCL low.
*
* @param    InstancePtr is a pointer to the XIic instance to be worked on.
*
* @return
*
* None.
*
* @note
*
* None.
*
* @internal
*
* NOTES:
*   Healing a stuck-SCL IIC bus is attempted by holding SDA low for >35msec. 
*
*******************************************************************************/
u32 PPC_check_and_heal_stuck_scl(int bus_index)
{
	u32 GpioCntlReg;
	u32 I2CBusBaseAddressOffset;
	u32 I2CBusSoftResetAddress;
	u32 status = 0;

	if (bus_index == 5) 
	{
		I2CBusBaseAddressOffset = 0xef600710;
		I2CBusSoftResetAddress  = 0xef60070c;
	}
	else 
	{
		I2CBusBaseAddressOffset = 0xef600810;
		I2CBusSoftResetAddress  = 0xef60080c;
	}
	// IIC controller must be placed in the reset state
	CMCGpioReadReg32Unmapped(I2CBusSoftResetAddress, 0, &GpioCntlReg);
	GpioCntlReg |= 0x00000001;
	CMCGpioWriteReg32Unmapped(I2CBusSoftResetAddress, 0, GpioCntlReg);

	CMCGpioReadReg32Unmapped(I2CBusBaseAddressOffset, 0, &GpioCntlReg);
	GpioCntlReg |= 0x0C000000; // Make SCL and SDA high
	CMCGpioWriteReg32Unmapped(I2CBusBaseAddressOffset, 0, GpioCntlReg);

	CMCGpioReadReg32Unmapped(I2CBusBaseAddressOffset, 0, &GpioCntlReg);

	// Test to see if the SCL clock is low
	if(0 == (GpioCntlReg & 0x01000000)) 
	{
		CMCGpioReadReg32Unmapped(I2CBusBaseAddressOffset, 0, &GpioCntlReg);
		GpioCntlReg &= ~(0x08000000); /* drive SDA low */
		GpioCntlReg |= 0x04000000;    /* drive SCL high */
		CMCGpioWriteReg32Unmapped(I2CBusBaseAddressOffset, 0, GpioCntlReg);

		mdelay(100); /* delay , milliseconds*/

    	GpioCntlReg |= 0x0C000000; /* release SDA and SCL */
    	CMCGpioWriteReg32Unmapped(I2CBusBaseAddressOffset, 0, GpioCntlReg);

		CMCGpioReadReg32Unmapped(I2CBusBaseAddressOffset, 0, &GpioCntlReg);
		if (GpioCntlReg & 0x01000000)
		{
			status = 0;
		}
		else
		{
			status = -1;
		}
	}

	// set IIC controller back to normal operation state
	CMCGpioReadReg32Unmapped(I2CBusSoftResetAddress, 0, &GpioCntlReg);
	GpioCntlReg &= ~(0x00000001);
  	CMCGpioWriteReg32Unmapped(I2CBusSoftResetAddress, 0, GpioCntlReg); 	

    return status;
}

/*******************************************************************************
* FUNCTION: XIic_check_and_heal_stuck_scl()
* ____________________________________________________________________________
*
* DESCRIPTION:  This function attempts to wake up TWI slave devices that
*               are holding SCL low.
*
* @param    InstancePtr is a pointer to the XIic instance to be worked on.
*
* @return
*
* None.
*
* @note
*
* None.
*
* @internal
*
* NOTES:
*   Healing a stuck-SCL IIC bus is attempted by holding SDA low for >35msec. 
*
*******************************************************************************/
u32 XIic_check_and_heal_stuck_scl(int bus_index)
{
   u32 GpioCntlReg;
   u32 I2CBusBaseAddressOffset;
   u16 RegOffset;
   u32 status = 0;



   I2CBusBaseAddressOffset = CMC_XILINX_I2C_BUS_0_OFFSET + 0x200 * bus_index;
   RegOffset = I2CBusBaseAddressOffset + 0x128;

   CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, RegOffset, &GpioCntlReg);
   
           GpioCntlReg |= IIC_SDA_T; /* tri-state SDA */
           GpioCntlReg |= IIC_SCL_T; /* tri-state SCL */
	   gpio_i2c_dbg("XIic_check_and_heal_stuck_scl, writing IIC_SDA_T, IIC_SCL_T Hi, reg: 0x%x\n", GpioCntlReg);
	   CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);
        GpioCntlReg |= IIC_EN_SOFT; /* Take over the bus */
	   gpio_i2c_dbg("XIic_check_and_heal_stuck_scl, taking over the bus, writing reg: 0x%x\n", GpioCntlReg);
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);
   	CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, RegOffset, &GpioCntlReg);
         if(0 == (GpioCntlReg & IIC_SCL_O)) {
	   gpio_i2c_dbg("XIic_check_and_heal_stuck_scl, IIC_SCL_O is hi, do the recovery\n");
       	   GpioCntlReg &= ~IIC_SDA_T; /* un tri-state SDA */
           GpioCntlReg &= ~IIC_SCL_T; /* un tri-state SCL */
	   CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);
        
            /* drive SDA low */

   	    CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, RegOffset, &GpioCntlReg);
            GpioCntlReg &= ~(IIC_SDA_O);
            GpioCntlReg |= IIC_SCL_O | IIC_EN_SOFT;
	   gpio_i2c_dbg("XIic_check_and_heal_stuck_scl, drive SDA low, writing reg: 0x%x\n", GpioCntlReg);
	    CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);

            /* delay , milliseconds*/

            msleep(40); 

            /* release SDA */
   	    CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, RegOffset, &GpioCntlReg);
	    // Clock, Data and Soft control == 1
            GpioCntlReg |= IIC_SDA_O | IIC_SCL_O | IIC_EN_SOFT;
	   gpio_i2c_dbg("XIic_check_and_heal_stuck_scl, release SDA, writing reg: 0x%x\n", GpioCntlReg);
	    CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);
	    // Tri-state all the bus pins and enable software control
            GpioCntlReg |= IIC_SDA_T | IIC_SCL_T | IIC_SDA_O | IIC_SCL_O | IIC_EN_SOFT;
	   gpio_i2c_dbg("XIic_check_and_heal_stuck_scl, tri-state and enable software, writing reg: 0x%x\n", GpioCntlReg);
	    CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);

   	CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, RegOffset, &GpioCntlReg);
	if (GpioCntlReg & IIC_SCL_O)
	{
		status = 0;
	   gpio_i2c_dbg("XIic_check_and_heal_stuck_scl, IIC_SCL_O is set, good status, reg: 0x%x\n", GpioCntlReg);
	}
	else
	{
		status = -1;
	   gpio_i2c_dbg("XIic_check_and_heal_stuck_scl, IIC_SCL_O is not set, bad status, reg: 0x%x\n", GpioCntlReg);
	}

        /* return to normal operating mode */
        GpioCntlReg |= IIC_SDA_T; /* Tri-state SDA */
        GpioCntlReg |= IIC_SCL_T; /* Tri-state SCL */
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);

        GpioCntlReg &= ~(IIC_EN_SOFT); /* Release the bus */
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);
    } else if (0 == (GpioCntlReg & IIC_SDA_O))
    { // SDA is stucked low, we need to reset
	status = -1;
    }
    return status;
}

/*******************************************************************************
* FUNCTION: XIic_check_and_disable_i2c_bus()
* ____________________________________________________________________________
*
* DESCRIPTION:  This function attempts to check I2C bus state and disable it safely.
*
* @param    InstancePtr is a pointer to the XIic instance to be worked on.
*
* @return
*
* None.
*
* @note
*
* None.
*
* @internal
*
* NOTES:
*   This function should be called only before reboot to make the I2C Bus in Idle condition
*   at the shutdown. Defect BITS067915
*******************************************************************************/
u32 XIic_check_and_disable_i2c_bus(int bus_index)
{
   u32 GpioCntlReg;
   u32 I2CBusBaseAddressOffset;
   u16 RegOffset;
   u32 status = 0;

	I2CBusBaseAddressOffset = CMC_XILINX_I2C_BUS_0_OFFSET + 0x200 * bus_index;
	RegOffset = I2CBusBaseAddressOffset + 0x0100;
	CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, RegOffset, &GpioCntlReg);
	while (GpioCntlReg & 0x04){
		udelay(1000);/* Wait for completion of I2C operation */
		CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, RegOffset, &GpioCntlReg);
		}
	/*Disable the I2C Bus*/
	GpioCntlReg &=~(0x01);
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);
	return status;
}

/*******************************************************************************
* FUNCTION: PPC_check_and_heal_stuck_bus_reset()
* ____________________________________________________________________________
*
* DESCRIPTION:  This function attempts to wake up I2C slave devices that
*               are holding SCL or SDA line low.
*
*
* @return
*
* None.
*
* @note
*
* None.
*
* @internal
*
* NOTES:
*   Healing a stuck IIC bus is attempted by holding reset line  low for >100msec. 
*
*******************************************************************************/
u32 PPC_check_and_heal_stuck_bus_reset(int bus_index)
{
   u32 GpioCntlReg;
   u32 status = 0;

   switch (bus_index)
   {	
	case 5:
    	// PPC i2c0
		// There is no reset function for bus 0
		break;

	case 6:
    	// PPC i2c1
		CMCGpioReadReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, &GpioCntlReg);
		GpioCntlReg &= ~(0x0000000000000002);
		CMCGpioWriteReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, GpioCntlReg);

		msleep(110); 

		GpioCntlReg |= 0x0000000000000002;
		CMCGpioWriteReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, GpioCntlReg);

		CMCGpioReadReg32Unmapped(0xef600810, 0, &GpioCntlReg);
		if ((GpioCntlReg & 0x03000000) == 0x03000000)
			status = 0;
		else  
			status = -1;
		break;

	default:
		break;
   }
   msleep(100);

   return status; 
}

/*******************************************************************************
* FUNCTION: XIic_check_and_heal_stuck_bus_reset()
* ____________________________________________________________________________
*
* DESCRIPTION:  This function attempts to wake up I2C slave devices that
*               are holding SCL or SDA line low.
*
*
* @return
*
* None.
*
* @note
*
* None.
*
* @internal
*
* NOTES:
*   Healing a stuck IIC bus is attempted by holding reset line  low for >100msec. 
*
*******************************************************************************/
u32 XIic_check_and_heal_stuck_bus_reset(int bus_index)
{
   u32 GpioCntlReg;
   u32 I2CBusBaseAddressOffset;
   u16 RegOffset;
   u32 status = 0;
   u32 temp;

   I2CBusBaseAddressOffset = 0x1600 + 0x200 * bus_index;
   RegOffset = I2CBusBaseAddressOffset + 0x128;
   
   switch (bus_index)
   {	
	case 0:
    	// IOM1, GPIO 59
	CMCGpioReadReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, &GpioCntlReg);
	temp = ~(0x00000010);
	GpioCntlReg &= temp;
	CMCGpioWriteReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, GpioCntlReg);
            msleep(110); 
	GpioCntlReg |= 0x00000010;
	CMCGpioWriteReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, GpioCntlReg);
	break;
	case 1:
    	// IOM2, GPIO 60
	CMCGpioReadReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, &GpioCntlReg);
	temp = ~(0x00000008);
	GpioCntlReg &= temp;
	CMCGpioWriteReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, GpioCntlReg);
            msleep(110); 
	GpioCntlReg |= 0x00000008;
	CMCGpioWriteReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, GpioCntlReg);
	break;
	case 2:
    	// PSU1, GPIO 57
	CMCGpioReadReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, &GpioCntlReg);
	temp = ~(0x00000040);
	GpioCntlReg &= temp;
	CMCGpioWriteReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, GpioCntlReg);
            msleep(110); 
	GpioCntlReg |= 0x00000040;
	CMCGpioWriteReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, GpioCntlReg);
	break;
	case 3:
    	// PSU2, GPIO 58
	gpio_i2c_dbg("XIic_check_and_heal_stuck_bus_reset, PSU2 GPIO 58.\n");
	CMCGpioReadReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, &GpioCntlReg);
	temp = ~(0x00000020);
	GpioCntlReg &= temp;
	CMCGpioWriteReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, GpioCntlReg);
            msleep(110); 
	GpioCntlReg |= 0x00000020;
	CMCGpioWriteReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, GpioCntlReg);
	break;
	case 4:
    	// KVM, GPIO 61
	CMCGpioReadReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, &GpioCntlReg);
	temp = ~(0x00000004);
	GpioCntlReg &= temp;
	CMCGpioWriteReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, GpioCntlReg);
            msleep(110); 
	GpioCntlReg |= 0x00000004;
	CMCGpioWriteReg32Unmapped(CMC_PPC_GPIO_0_BASE_ADDRESS_1, 0, GpioCntlReg);
	break;
	default:
	break;
   }
   msleep(100);

   	CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, RegOffset, &GpioCntlReg);
      	// Tri-state all the bus pins and enable software control
        GpioCntlReg |= IIC_SDA_T | IIC_SCL_T | IIC_SDA_O | IIC_SCL_O | IIC_EN_SOFT;
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);

   	CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, RegOffset, &GpioCntlReg);
	if (GpioCntlReg & 0x000a)
	{
		gpio_i2c_dbg("XIic_check_and_heal_stuck_bus_reset, good status; 0x%x\n", GpioCntlReg);
		status = 0;
	}
	else
	{
		gpio_i2c_dbg("XIic_check_and_heal_stuck_bus_reset, bad status; 0x%x\n", GpioCntlReg);
		status = -1;
	}

        /* return to normal operating mode */
        GpioCntlReg |= IIC_SDA_T; /* Tri-state SDA */
        GpioCntlReg |= IIC_SCL_T; /* Tri-state SCL */
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);

        GpioCntlReg &= ~(IIC_EN_SOFT); /* Release the bus */
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, RegOffset, GpioCntlReg);
   return status;

}

/* the name of this function should be less than 16 bytes (sizeof(current->comm)) */
#define GPIO_RECOVERY_THREAD "gpio_recov_thr"

void gpio_recov_thr(void *data)
{
	int recovered_index;
	int status;
	PGPIO_DEVICE_EXTENSION pGpioCard = (PGPIO_DEVICE_EXTENSION)data;

	#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
    	daemonize (
			"%s", 
			GPIO_RECOVERY_THREAD
			);
	#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,4,7)
    	daemonize ();
		// need to do this to avoid zombi threads in kernel...
    	reparent_to_init ();
	#else
    	daemonize ();
	#endif// #if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)

	strcpy (
		current->comm, 
		GPIO_RECOVERY_THREAD
		);

	while(RecoveryThreadExitingFlag == FALSE)
	{
		pGpioCard->recovery_thread_sleeping = TRUE;
		xilinx_recovery_event_present = 0;
               	wait_event_interruptible (XilinxRecoverySemaphore, (xilinx_recovery_event_present != 0));
                if (driver_exited)
                {
	            RecoveryThreadExitingFlag = TRUE; 
                    break;
                }
                
	        recovered_index = pGpioCard->PPCBusIndex;
	        if ((recovered_index == 5) || (recovered_index == 6))
		       status = PPC_Recover(recovered_index);
	        else
		       status = XIic_Recover(recovered_index);
                
	}//while(ThreadExitingFlag == FALSE)
	gpio_complete_thread(pGpioCard->recovery_thread_data);
}

/* the name of this function should be less than 16 bytes (sizeof(current->comm)) */
#define GPIO_DISABLE_IOM_I2C_THREAD "gpio_disi2c_thr"

void gpio_disi2c_thr(void *data)
{
	int disable_index;
	int status;
	PGPIO_DEVICE_EXTENSION pGpioCard = (PGPIO_DEVICE_EXTENSION)data;

	#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
    	daemonize (
			"%s", 
			GPIO_RECOVERY_THREAD
			);
	#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,4,7)
    	daemonize ();
		// need to do this to avoid zombi threads in kernel...
    	reparent_to_init ();
	#else
    	daemonize ();
	#endif// #if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)

	strcpy (
		current->comm, 
		GPIO_RECOVERY_THREAD
		);

	while(RecoveryThreadExitingFlag == FALSE)
	{
		pGpioCard->recovery_thread_sleeping = TRUE;
		xilinx_disable_event_present = 0;
               	wait_event_interruptible (XilinxDisableSemaphore, (xilinx_disable_event_present != 0));
                if (driver_exited)
                {
	            RecoveryThreadExitingFlag = TRUE; 
                    break;
                }
                
	        disable_index = pGpioCard->PPCBusIndex;
            status = XIic_Disable(disable_index);
                
	}//while(ThreadExitingFlag == FALSE)
	gpio_complete_thread(pGpioCard->recovery_thread_data);
}

//---------------------------------------------------------------------
// GpioInitDriver
//---------------------------------------------------------------------
s32 GpioInitDriver(PGPIO_DEVICE_EXTENSION pGPIOCard)
{

	int status = 0;
	int temp;
	int blade_present;
	int Status;
	int i =0;

        driver_exited = 0;
	blade_present = 0;

	gpio_card.GPIOInitComplete = 0;
	// initialize driver structure
        if (GpioDDStart(pGPIOCard) != GPIO_STATUS_CMD_SUCCESS)
        {
                printk(EMsgDevNotFound);
                return GPIO_STATUS_CMD_NO_SUCH_DEVICE;
        }
        /* Allocate Spin lock */
        spin_lock_init(&ControlLock_bhlock);
        spin_lock_init(&ControlLock_irqlock);
        spin_lock_init(&pGPIOCard->blade_controlLock_irqlock);
        spin_lock_init(&pGPIOCard->fan_controlLock_irqlock);
        spin_lock_init(&pGPIOCard->psu_controlLock_irqlock);
        spin_lock_init(&pGPIOCard->iom_controlLock_irqlock);
        spin_lock_init(&pGPIOCard->cp_controlLock_irqlock);
        spin_lock_init(&pGPIOCard->kvm_controlLock_irqlock);
        spin_lock_init(&pGPIOCard->shifty_controlLock_irqlock);
	// Allocate I/O IRP completion semaphores, unlocked
	init_waitqueue_head(&FanInterruptSemaphore);
	init_waitqueue_head(&TempInterruptSemaphore);
	init_waitqueue_head(&ShftgInterruptSemaphore);
	init_waitqueue_head(&KVMInterruptSemaphore);
	init_waitqueue_head(&BladeInterruptSemaphore);
	init_waitqueue_head(&IOMInterruptSemaphore);
	init_waitqueue_head(&ECCInterruptSemaphore);
	init_waitqueue_head(&XilinxRecoverySemaphore);
	init_waitqueue_head(&PSUInterruptSemaphore);
	init_waitqueue_head(&CPInterruptSemaphore);
	init_waitqueue_head(&recovery_thread_semaphore);
	init_waitqueue_head(&XilinxDisableSemaphore);		
	pGPIOCard->InterruptRequiresDpc = TRUE;
	/* Allocate IRQ for interrupts */
	pGPIOCard->PPCFanIrqLevel = PPC_GPIO_FAN_IRQ;
	pGPIOCard->PPCTempIrqLevel = PPC_GPIO_TEMP_IRQ;
	pGPIOCard->PPCShftgIrqLevel = PPC_GPIO_SHIFTG_IRQ;
	pGPIOCard->PPCKVMIrqLevel = PPC_GPIO_KVM_IRQ;
	pGPIOCard->PPCBladeIrqLevel = PPC_GPIO_BLADE_IRQ;
	pGPIOCard->PPCIOMIrqLevel = PPC_GPIO_IOM_IRQ;
	pGPIOCard->PPCECCIrqLevel = PPC_GPIO_ECC_IRQ;
	pGPIOCard->PPCPSUIrqLevel = PPC_GPIO_PSU_IRQ;
	pGPIOCard->PPCCPIrqLevel = PPC_GPIO_CP_IRQ;
	pGPIOCard->GPIOFanInterruptCount = 0;
	pGPIOCard->GPIOBladeInterruptCount = 0;
	pGPIOCard->GPIOShiftyInterruptCount = 0;
	pGPIOCard->GPIOIOMInterruptInitialStatus = 0;
	pGPIOCard->GPIOPSUInterruptInitialStatus = 0;
	pGPIOCard->GPIOKVMInterruptInitialStatus = 0;
	pGPIOCard->GPIOKVMInterruptCount = 0;
	pGPIOCard->GPIOIOMInterruptCount = 0;
	pGPIOCard->GPIOECCInterruptCount = 0;
	pGPIOCard->GPIOPSUInterruptCount = 0;
	pGPIOCard->GPIOCPInterruptCount = 0;
	pGPIOCard->GPIOIOMInterruptCount = 0;
	pGPIOCard->GPIOKVMSMBUSTimeoutCount = 0;
	pGPIOCard->GPIOIOM1SMBUSTimeoutCount = 0;
	pGPIOCard->GPIOIOM2SMBUSTimeoutCount = 0;
	pGPIOCard->GPIOPSU1SMBUSTimeoutCount = 0;
	pGPIOCard->GPIOPSU2SMBUSTimeoutCount = 0;
    	pGPIOCard->GPIOFirstIO = 0;
	memset(&blade_intr_queue, 0, sizeof(blade_intr_queue));
	blade_intr_queue.fifo = blade_fifo;
	blade_intr_queue.fifosize = BLADE_FIFOSIZE;
	memset(&fan_intr_queue, 0, sizeof(fan_intr_queue));
	fan_intr_queue.fifo = fan_fifo;
	fan_intr_queue.fifosize = FAN_FIFOSIZE;
	memset(&iom_intr_queue, 0, sizeof(iom_intr_queue));
	iom_intr_queue.fifo = blade_fifo;
	iom_intr_queue.fifosize = IOM_FIFOSIZE;
	memset(&kvm_intr_queue, 0, sizeof(kvm_intr_queue));
	kvm_intr_queue.fifo = kvm_fifo;
	kvm_intr_queue.fifosize = KVM_FIFOSIZE;
	memset(&psu_intr_queue, 0, sizeof(psu_intr_queue));
	psu_intr_queue.fifo = psu_fifo;
	psu_intr_queue.fifosize = PSU_FIFOSIZE;

	for(i=0;i<7;i++) psu_intr_queue.psu_in_count[i]=0;
	

	memset(&cp_intr_queue, 0, sizeof(cp_intr_queue));
	cp_intr_queue.fifo = cp_fifo;
	cp_intr_queue.fifosize = CP_FIFOSIZE;
	memset(&shifty_intr_queue, 0, sizeof(shifty_intr_queue));
	shifty_intr_queue.fifo = shifty_fifo;
	shifty_intr_queue.fifosize = SHIFTY_FIFOSIZE;
	// Map 6K space for GPIO use
        gpio_card.GPIOMappedBaseAddress = (u32)fpga_base_vaddr;
        gpio_card.IIC0MappedBaseAddress = (u32)iic0_base_vaddr;
        gpio_card.IIC1MappedBaseAddress = (u32)iic1_base_vaddr;
        gpio_card.GPIO0MappedBaseAddress = (u32)gpio0_base_vaddr;
        gpio_card.GPIO1MappedBaseAddress = (u32)gpio1_base_vaddr;

	init_waitqueue_head(&recovery_thread_semaphore);

	// Create kernel thread
	pGPIOCard->recovery_thread_data = gpio_kernel_thread((void *)gpio_recov_thr,
								(void *)pGPIOCard,
								GPIO_RECOVERY_THREAD);
	if(pGPIOCard->recovery_thread_data == NULL)
	{
		printk("kernel thread creation failed\n");
	}
	// Create kernel thread
	pGPIOCard->recovery_thread_data = gpio_kernel_thread((void *)gpio_disi2c_thr,
								(void *)pGPIOCard,
								GPIO_DISABLE_IOM_I2C_THREAD);
	if(pGPIOCard->recovery_thread_data == NULL)
	{
		printk("kernel thread creation failed\n");
	}	
#ifdef GPIO_WITH_INTERRUPTS
	// Disable the shifty interrupt first
	temp = 0;
	printk("Disabling shifty bus interrupt for now.\n");
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_SHIFTG_INT_MASK_BASE, temp);

	// Mask the IOM interrupts bit 15 I2C_IOM123_ALERT and bit 14
	// I2C_IOM456_ALERT
	CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_IOM_INT_MASK_BASE, &temp);
	temp |= 0xc000;
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_IOM_INT_MASK_BASE, temp|DONT_ENABLE_SMBUS_TIMEOUT_FOR_IOMS);
	// Mask the PSU interrupts bit 15 I2C_PSU123_ALERT and bit 14
	// I2C_PSU456_ALERT
	CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_PSU_INT_MASK_BASE, &temp);
	temp |= 0xc000;
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_PSU_INT_MASK_BASE, temp);
	// Mask the KVM interrupts bit 15 I2C_KVM_ALERT
	CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_KVM_INT_MASK_BASE, &temp);
	temp |= 0x8000;
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_KVM_INT_MASK_BASE, temp);

	// Load the interrupts for all buses
        status = GpioLoadDevInterrupt(pGPIOCard);
	// Check if any blade is present
	// Before we can read the blade present status, we need to send the 
	// command to turn on the blade.
	// We don't need to check the blade presence.
	// Removed the following codes
#ifdef chl
	// Power on the blade
	blade_status = 0xffff;
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_TURN_BLADE_POWER_SUPPLY_ON_OFF, blade_status);
        for ( i = 0; i < 16; i++)
	{
		temp = CMC_GPIO_BLADE_STATUS_BASE;
		blade_address = temp + 2 * i;
		CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, blade_address, &blade_status);
		if (blade_status & 0x8000)
		{
			blade_present = 1;
			break;
		}
	}
#endif
	// Clear shifty interrupts
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_SHIFTG_INT_BASE, 0x07f0);
	msleep (50);
	// Enable the shifty bus primary registers
	CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_SHIFTG_INT_PRIMARY_ENABLE, 1);

#ifdef chl
	if (blade_present)
	{	
		// Enable Shifty bus master interrupt.
		//printk("CMC Gpio: Blade present.\n");
		//printk("Blade present, enabling shifty bus interrupt.\n");
		//CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, 0x1412, 0xc3ff);
	}
	temp =  mfdcr(0xe0);
	printk("UIC2 status register e0: 0x%x\n", temp);
	temp =  mfdcr(0xe2);
	printk("UIC2 enable register e2: 0x%x\n", temp);
	temp =  mfdcr(0xe6);
	printk("UIC2 Mask register e6: 0x%x\n", temp);
#endif
	printk("GpioLoadDevInterrupt, registering interrupt for blade, irq: %d\n",
					pGPIOCard->PPCBladeIrqLevel);
        if ((Status = request_irq((unsigned int)pGPIOCard->PPCBladeIrqLevel, GpioIsr,
                //SA_SHIRQ, "gpio", pGPIOCard))
		// 2.6.24
                IRQF_SHARED, "gpio", pGPIOCard))
		!= GPIO_STATUS_CMD_SUCCESS)
        {
              		printk(KERN_WARNING "gpio_init_card: unable to install interrupt handler - skipping controller\n");
			init_waitqueue_head(&BladeInterruptSemaphore);
			return Status;
        }

#endif
        // Initialize GPIO registers
	// Tell ISR that init is done
	gpio_card.GPIOInitComplete = 1;
	return status;
} // GpioInitDriver
 

static void __exit
gpio_exit(void)
{
	int exit_delay = 900000;
        driver_exited = 1;
	GpioUnload(&gpio_card);
        udelay (exit_delay);
}

// Lock the spinlock

void gpio_spin_lock(PGPIO_DEVICE_EXTENSION pGpioCard)
{
	   spin_lock(&ControlLock_bhlock);
}

// Unlock the spinlock

void gpio_spin_unlock(PGPIO_DEVICE_EXTENSION pGpioCard)
{
	spin_unlock(&ControlLock_bhlock);
	
}


module_init(gpio_init);
module_exit(gpio_exit);
                                                                                  
