/*
 * Copyright (C) 2007 Dell Inc.
 * Chieng-Hwa Lin <chieng-hwa_lin@dell.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifdef __KERNEL__
#include <linux/spinlock.h>
#endif /* __KERNEL__ */

//#define CMC_DRIVER_DBG 1
// typedef unsigned int u32;
// typedef unsigned char u8;
// typedef signed int s32;
typedef int booln;
typedef long             LONG;
typedef unsigned long   ULONG;
                                                                                
#define GPIO_WITH_INTERRUPTS 1
#define TRUE    1
#define FALSE   0
#define HANDLE  int

#define GPIO_STATUS_CMD_UNSUCCESSFUL             (-1)
#define GPIO_STATUS_CMD_SUCCESS                  (0)
#define GPIO_STATUS_CMD_NO_SUCH_DEVICE           (7)
// IRQ definitions
#define PPC_GPIO_FAN_IRQ           		(60)
#define PPC_GPIO_TEMP_IRQ           		(67)
#define PPC_GPIO_SHIFTG_IRQ           		(68)
#define PPC_GPIO_KVM_IRQ           		(27)
#define PPC_GPIO_BLADE_IRQ           		(64)
#define PPC_GPIO_IOM_IRQ           		(65)
#define PPC_GPIO_PSU_IRQ           		(50)
#define PPC_GPIO_CP_IRQ           		(51)
#define PPC_GPIO_ECC_IRQ           		(69)

#define GPIO_INTERRUPT				1 
#define NON_GPIO_INTERRUPT			2 

#define GPIO_LOSS_SYNC_STATUS			0x40
#define GPIO_LOSS_CLK_STATUS			0x20
#define GPIO_UNEXP_SYNC_STATUS			0x80
#define GPIO_SYNC_CONT_STATUS			0x200
#define GPIO_STALE_DATA				0x0100

//-------------------------------------------------------------------
// Ecc data
//-------------------------------------------------------------------a
#define DDR0_00                         0x00
#define DDR0_22                         0x16
// Power PC Spec. p. 949
#define DDR_DCR_BASE 0x10
#define ddrcfga  (DDR_DCR_BASE+0x0)   /* DDR configuration address reg */
#define ddrcfgd  (DDR_DCR_BASE+0x1)   /* DDR configuration data reg    */

// Power PC Spec. p. 318
/* Status */
#define DDR0_00_INT_ACK_ALL               0x7F000000
#define DDR0_00_INT_STATUS_MASK           0x00FF0000 /* Read only */
/* Bit0. A single access outside the defined PHYSICAL memory space detected. */
#define DDR0_00_INT_STATUS_BIT0           0x00010000
/* Bit1. Multiple accesses outside the defined PHYSICAL memory space detected. */
#define DDR0_00_INT_STATUS_BIT1           0x00020000
/* Bit2. Single correctable ECC event detected */
#define DDR0_00_INT_STATUS_BIT2           0x00040000
/* Bit3. Multiple correctable ECC events detected. */
#define DDR0_00_INT_STATUS_BIT3           0x00080000
/* Bit4. Single uncorrectable ECC event detected. */
#define DDR0_00_INT_STATUS_BIT4           0x00100000
/* Bit5. Multiple uncorrectable ECC events detected. */
#define DDR0_00_INT_STATUS_BIT5           0x00200000
/* Bit6. DRAM initialization complete. */
#define DDR0_00_INT_STATUS_BIT6           0x00400000
/* Bit7. Logical OR of all lower bits. */
#define DDR0_00_INT_STATUS_BIT7           0x00800000

#define DDR0_22_CTRL_RAW_MASK             0x03000000
#define DDR0_22_CTRL_RAW_ECC_DISABLE      0x00000000 /* ECC not being used */
#define DDR0_22_CTRL_RAW_ECC_CHECK_ONLY   0x01000000 /* ECC checking is on, but no attempts to correct*/
#define DDR0_22_CTRL_RAW_NO_ECC_RAM       0x02000000 /* No ECC RAM storage available */
#define DDR0_22_CTRL_RAW_ECC_ENABLE       0x03000000 /* ECC checking and correcting on */
#define DDR0_22_CTRL_RAW_ENCODE(n)          ((((unsigned long)(n))&0x3)<<24)
#define DDR0_22_CTRL_RAW_DECODE(n)          ((((unsigned long)(n))>>24)&0x3)

#define FAN_FIFOSIZE           		256
#define SHIFTY_FIFOSIZE        		256
#define KVM_FIFOSIZE           		256
#define BLADE_FIFOSIZE         		256
#define IOM_FIFOSIZE           		256
#define PSU_FIFOSIZE           		256
#define CP_FIFOSIZE           		256

/* interrupt queue structure. */
#define ISRFULLFIFO(p) \
        (((p)->in == (p)->out-1) || ((p)->in == ((p)->fifosize-1) && (p)->out == 0))
#define ISREMPTYFIFO(p) ((p)->in == (p)->out)

struct psu_intr_status_t{
    int waiting;
    int incnt;
    int outcnt;
    int overflows;
    int watermark;
    int in;
    int out;
    short *fifo;
    int fifosize;
    int psu_in_count[7];
};



struct intr_status_t{
    int waiting;
    int incnt;
    int outcnt;
    int overflows;
    int watermark;
    int in;
    int out;
    short *fifo;
    int fifosize;
};


#define PSU_1 1
#define PSU_2 2
#define PSU_3 4
#define PSU_4 8
#define PSU_5 16
#define PSU_6 32
#define EPP_ENABLE 64


//--------------------------------------------------------------------
// Defines and strucutures
//--------------------------------------------------------------------

#ifdef __KERNEL__

typedef struct _GPIO_DEVICE_EXTENSION
{
	unsigned int  istat;
	unsigned int  BaseAddress;
	unsigned int Offset;
	unsigned int DeviceBitmap;
	unsigned int InterruptRequiresDpc;
	unsigned int GPIOMappedBaseAddress;
	unsigned int IIC0MappedBaseAddress;
	unsigned int IIC1MappedBaseAddress;
	unsigned int GPIO0MappedBaseAddress;
	unsigned int GPIO1MappedBaseAddress;
	unsigned int GPIOFanInterruptStatus;
	unsigned int GPIOShiftyInterruptStatus;
	unsigned int GPIOKVMInterruptStatus;
	unsigned int GPIOBladeInterruptStatus;
	unsigned int GPIOPSUInterruptStatus;
	unsigned int GPIOCPInterruptStatus;
	unsigned int GPIOIOMInterruptStatus;
	unsigned int GPIOECCInterruptStatus;
	unsigned int GPIOPSUInterruptInitialStatus;
	unsigned int GPIOIOMInterruptInitialStatus;
	unsigned int GPIOKVMInterruptInitialStatus;
	unsigned int GPIOFanInterruptCount;
	unsigned int GPIOBladeInterruptCount;
	unsigned int GPIOShiftyInterruptCount;
	unsigned int GPIOIOMInterruptCount;
	unsigned int GPIOECCInterruptCount;
	unsigned int GPIOKVMInterruptCount;
	unsigned int GPIOPSUInterruptCount;
	unsigned int GPIOCPInterruptCount;
	unsigned int GPIOIOM1SMBUSTimeoutCount;
	unsigned int GPIOIOM2SMBUSTimeoutCount;
	unsigned int GPIOPSU1SMBUSTimeoutCount;
	unsigned int GPIOPSU2SMBUSTimeoutCount;
	unsigned int GPIOKVMSMBUSTimeoutCount;
	unsigned int GPIOFirstIO;
	unsigned int GPIOInitComplete;
	LONG FanInterruptSemaphore;  // Sempahore for ISR completions to fire DPCForISR
	LONG IOIRPCompleteSemaphore;  // Sempahore for I/O completions


	unsigned int InterruptCount;
	unsigned int PPCFanIrqLevel;
	unsigned int PPCTempIrqLevel;
	unsigned int PPCShftgIrqLevel;
	unsigned int PPCKVMIrqLevel;
	unsigned int PPCBladeIrqLevel;
	unsigned int PPCIOMIrqLevel;
	unsigned int PPCECCIrqLevel;
	unsigned int PPCPSUIrqLevel;
	unsigned int PPCCPIrqLevel;
	unsigned int PPCBusIndex;
	void *recovery_thread_data;
	unsigned char  recovery_thread_sleeping;
	spinlock_t    blade_controlLock_irqlock;
	spinlock_t    fan_controlLock_irqlock;
	spinlock_t    psu_controlLock_irqlock;
	spinlock_t    iom_controlLock_irqlock;
	spinlock_t    cp_controlLock_irqlock;
	spinlock_t    kvm_controlLock_irqlock;
	spinlock_t    shifty_controlLock_irqlock;
	
} GPIO_DEVICE_EXTENSION, *PGPIO_DEVICE_EXTENSION;

// GPIO_DEVICE_EXTENSION gpio_card; 

#endif /* __KERNEL__ */

/*--------------------------------------------------------------------
 * Packet definitions for GPIO
 *--------------------------------------------------------------------*/
typedef struct _GpioCmdIoctlReq
{
        unsigned int ReqType;
        unsigned int BaseAddress;
        unsigned int Offset;
	unsigned int Mask;
	unsigned int Value;
        unsigned int Result;
        signed int Status;
} GpioCmdIoctlReq;

typedef struct
{
	GpioCmdIoctlReq CmdBuf;
	unsigned int CmdBufLen;
	GpioCmdIoctlReq RespBuf;
	unsigned int RespBufLen;
	unsigned int ReturnedLen;

} GpioCmdArg;

