/*
 * Copyright (C) 2007 Dell Inc.
 * Chieng-Hwa Lin <chieng-hwa_lin@dell.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
//--------------------------------------------------------------------
// Defines to customize headers
//--------------------------------------------------------------------

#if (!defined SYSTEM_DEBUG_LEVEL) && (defined _DEBUG)
#define SYSTEM_DEBUG_LEVEL      3
#endif

/*--------------------------------------------------------------------
 * Linux include files
 *--------------------------------------------------------------------*/
#include <linux/version.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/timer.h>
#include <linux/mtio.h>
#include <linux/ioctl.h>
#include <linux/fcntl.h>
// 2.6.24
#include <linux/interrupt.h>
#include <linux/poll.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
/*
// chl chl
//#include <linux/blk.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#include <linux/vmalloc.h>
#include <linux/spinlock.h>
#include <linux/wait.h>
#endif
*/
//--------------------------------------------------------------------
// Standard C include files
//--------------------------------------------------------------------


//--------------------------------------------------------------------
// Dell specific headers
//--------------------------------------------------------------------


//--------------------------------------------------------------------
// Shared include files
//--------------------------------------------------------------------
// Local include files
// Function prototypes and other local includes
//--------------------------------------------------------------------
//--------------------------------------------------------------------
// Prototypes
//--------------------------------------------------------------------
void CMCGpioReadReg(int base_address, int offset, int *reg_result);
void CMCGpioWriteReg(int base_address, int offset, u16 reg_input);
//--------------------------------------------------------------------
// External defines
//--------------------------------------------------------------------

extern int ECC_event_present;
extern int xilinx_recovery_event_present;
int  
extern gpio_i2c_dbg(
    const char *myformat,
    ...
    );
//--------------------------------------------------------------------
// Local defines
//--------------------------------------------------------------------

u32 returned_blade_intr_list_tail;
u32 returned_fan_intr_list_tail;
u32 returned_iom_intr_list_tail;
u32 returned_psu_intr_list_tail;
u32 returned_cp_intr_list_tail;
u32 returned_kvm_intr_list_tail;
u32 returned_shifty_intr_list_tail;

//--------------------------------------------------------------------
// Global Variables used for this module
//--------------------------------------------------------------------
extern spinlock_t    ControlLock_bhlock;
extern spinlock_t    ControlLock_irqlock;
extern wait_queue_head_t FanInterruptSemaphore;  // Sempahore for fan interrupts
extern wait_queue_head_t TempInterruptSemaphore;  // Sempahore for temperature
extern wait_queue_head_t ShftgInterruptSemaphore;  // Sempahore for shifty interrupt
extern wait_queue_head_t BladeInterruptSemaphore;  // Sempahore for blade interrupt
extern wait_queue_head_t IOMInterruptSemaphore;  // Sempahore for OM interrupt
extern wait_queue_head_t ECCInterruptSemaphore;  // Sempahore for OM interrupt
extern wait_queue_head_t XilinxRecoverySemaphore;  // Sempahore for I2C1 recovery
extern wait_queue_head_t PSUInterruptSemaphore;  // Sempahore for PSU interrupt
extern wait_queue_head_t CPInterruptSemaphore;  // Sempahore for CP interrupt
extern wait_queue_head_t KVMInterruptSemaphore;  // Sempahore for CP interrupt
extern struct intr_status_t *blade_intr_list_tail;
extern struct intr_status_t *fan_intr_list_tail;
extern struct intr_status_t *psu_intr_list_tail;
extern struct intr_status_t *iom_intr_list_tail;
extern struct intr_status_t *cp_intr_list_tail;
extern struct intr_status_t *kvm_intr_list_tail;
extern struct intr_status_t *shifty_intr_list_tail;

//-------------------------------------------------------------------
// Code start
//--------------------------------------------------------------------
#include <gpiotypes.h>
#include <gpio_struct.h>
#include <cmc_gpio.h>

int GpioDpcForISR( int irq);
int IntCount  = 0;
int IntCountA = 0;
int IntCountB = 0;
int IntCountC = 0;
int IntCountD = 0;

u32 QueueISRStatus(struct intr_status_t *intr_queue, u32 intr_status,
			wait_queue_head_t *semaphore);
void updatePSUCounter(struct psu_intr_status_t* psu_intr_queue , unsigned int int_register_value);
extern struct intr_status_t blade_intr_queue;
extern struct intr_status_t fan_intr_queue;
extern struct intr_status_t iom_intr_queue;
extern struct psu_intr_status_t psu_intr_queue;
extern struct intr_status_t cp_intr_queue;
extern struct intr_status_t kvm_intr_queue;
extern struct intr_status_t shifty_intr_queue;

extern GPIO_DEVICE_EXTENSION gpio_card;

#ifdef GPIO_WITH_INTERRUPTS
//---------------------------------------------------------------------
// GpioIsr
// Description: Interrupt serivce routine
//---------------------------------------------------------------------
irqreturn_t GpioIsr( int irq, void *dev_id, struct pt_regs *regs)
{
	u32	status;
	u32 	temp;
	GPIO_DEVICE_EXTENSION *pGpioPtr = dev_id;

	IntCount++;
#if 0
	if(IntCount > 1)
		printk(KERN_WARNING "Uh oh, Re-entry [Across IRQs], %d.\n",IntCount);
#endif

	status = IRQ_NONE;

	if (irq == pGpioPtr->PPCKVMIrqLevel)
	{
		IntCountA++;
		if(IntCountA > 1)
			printk(KERN_WARNING "Uh-Oh, Re-entry on KVM %d.\n",IntCountA);
		
		(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_KVM_INT_BASE, &temp);
		if (temp & 0x6000)
		{	
			// Don't remove these codes to other location.
			// We need to return IRQ_NONE as soon as possible
			// Not our interrupt, it's Xilinx I2C driver's interrupt.
			// We won't clear interrupts here, because it's Xilinx
			// I2C driver's job to clear the interrupts.
			// Don't print SMBUS timeout message here.
			// Xilinx driver will print out messages.
			// Don't print SMBUS timeout message here.
			// Xilinx driver will print out messages.
			if (temp & 0x2000)
				pGpioPtr->GPIOKVMSMBUSTimeoutCount += 1;
		}
	        
                if (temp & 0x8001)
                {
                   GpioDpcForISR(irq);
		   IntCount--;
		   IntCountA--;
		   return IRQ_HANDLED;
                }

		if (temp & 0x6000)
		{
		  IntCount--;
		  IntCountA--; 
		  return IRQ_NONE;
		}
		printk(KERN_WARNING "Unexpected bits in KVM int register-> %x\n",temp);
		IntCount--;
		IntCountA--; 
                return IRQ_HANDLED;
	}
	else if (irq == pGpioPtr->PPCIOMIrqLevel)
	{
		IntCountB++;
		if(IntCountB > 1) 
			printk(KERN_WARNING "Uh Oh, Re-entry on IOM, %d.\n",IntCountB);

		(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_IOM_INT_BASE, &temp);

#if 0	//we shouldn't be getting any interrupts for SMBUS timeout, the bit can still be set
		//so don't do anything based on that.

		// Don't remove these codes to other location.
		// We need to return IRQ_NONE as soon as possible
		// Not our interrupt, it's I2C's interrupt
		if (temp & 0x0c00) /*I2C SMbus Timeout on IOMSET1 and/or IOMSET2*/
		{
			// SMBUS timeout interrupt occurred.
			// We need to tell the IOM management that
			// SMBUS timeout happened by sending
			// the status.
			pGpioPtr->GPIOIOMInterruptStatus = temp;

			// Don't pass the SMBUS timeout interrupt
			// to IOM since we have the global variable
			// to keep track of the interrupt.

			//IOM_event_present = 1;
			//wake_up_interruptible(&IOMInterruptSemaphore);

			// Don't print SMBUS timeout message here.
			// Xilinx driver will print out messages.

			/* This is wrong, what if both happen? */
			//if (temp & 0x0800)
			//	pGpioPtr->GPIOIOM1SMBUSTimeoutCount += 1;
			//else
			//	pGpioPtr->GPIOIOM2SMBUSTimeoutCount += 1;

			if(temp & 0x0800) /*IOMSET1 SMBus Timeout*/
				pGpioPtr->GPIOIOM1SMBUSTimeoutCount += 1;
			if(temp & 0x0400) /*IOMSET2 SMBus Timeout*/
				pGpioPtr->GPIOIOM2SMBUSTimeoutCount += 1;
		}
#endif 	//don't expect smbus timeouts

		if(temp & 0xC03F) /*If Alerts, or IOM GPIOs*/
		{
			status = GpioDpcForISR(irq);
			//return IRQ_HANDLED;	//always give teh i2c driver a crack at it
		}

//		if(temp & 0x3C00) /*If I2C Irq's pending*/
		if(temp & 0x3000) /*If I2C Irq's pending*/	//the SMBUS timeout won't genterate an interrupt, but the 
													//the bit is still set, don't send it to the i2c driver
		{
			IntCount--;
			IntCountB--;
			return IRQ_NONE; /*Send to next driver*/
		}

//this is now very likely		printk(KERN_WARNING "Unexpected bits in IOM int register -> %0x\n",temp);
		IntCount--;
		IntCountB--;
		return IRQ_HANDLED; 
	}
	else if (irq == pGpioPtr->PPCPSUIrqLevel)
	{
		IntCountC++;
		if(IntCountC > 1)
			printk(KERN_WARNING "Uh oh, re-entry on PSU %d.\n",IntCountC);

		(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_PSU_INT_BASE, &temp);
		// Don't remove these codes to other location.
		// We need to return IRQ_NONE as soon as possible
		// Not our interrupt, it's I2C's interrupt
		if (temp & 0x0c00)
		{
			// SMBUS timeout interrupt occurred.
			// We need to tell the power management that
			// SMBUS timeout happened by sending
			// the status.
			// Don't tell power daemon about this interrupts
			// since we already keep track of the interrupts
			// in the global variable 
#if 0 
			pGpioPtr->GPIOPSUInterruptStatus = temp;
			status = QueueISRStatus(pGpioPtr, psu_intr_queue,
				pGpioPtr->GPIOPSUInterruptStatus,
			        &PSUInterruptSemaphore);
#endif
			// Don't print SMBUS timeout message here.
			// Xilinx driver will print out messages.
			if (temp & 0x0800) /*SMBus timeout PSU123*/
				pGpioPtr->GPIOPSU1SMBUSTimeoutCount += 1;	
			if (temp & 0x0400) /*SMBus timeout PSU456*/
				pGpioPtr->GPIOPSU2SMBUSTimeoutCount += 1;	
		}
		
		if (temp & 0xC07F) /* PSUxxx GPIO interrupts */
		{
			status = GpioDpcForISR(irq);
			//return IRQ_HANDLED;  //always give the i2c driver a crack at it.
		}

		if (temp & 0x3000) /* If any I2C IRQs pending, return NONE*/
		{
			IntCount--;
			IntCountC--;
			return IRQ_NONE;
		}
//this is now very likely		printk(KERN_WARNING "Unexpected bits in PSU interrupt ->%x\n", temp);
		IntCount--;
		IntCountC--;
		return IRQ_HANDLED; 
	}
	/* If we make it here, this means a Fan or Blade interrupt has occurred
 	   and are handled only by the GPIO driver */

	IntCountD++;
	if(IntCountD > 1) 
		printk(KERN_WARNING "Uh oh, re-entry detected on misc GPIO, %d.\n",IntCountD);

	status = GpioDpcForISR(irq);

	// clear the pending device interrupt
        // as DPC will complete the request

	// Need to clear the interrupt here

	if (status == GPIO_INTERRUPT)
	{
		status = IRQ_HANDLED;
	}
	else if (status == NON_GPIO_INTERRUPT)
	{
		status = IRQ_NONE;
	}

	IntCount--;
  	IntCountD--;
        return status;
}

u32
GpioClearDevInt( int irq)
{
	u32 temp;
	u32 interrupt_to_mask;
	u32 orig_mask;
	int status = 0;
	GPIO_DEVICE_EXTENSION *pGpioPtr;
	u32 val, new_val;
	u32 eppenablebit = 0;
	u32 intrstatus = 0;
	u32 intrmask = 0;

	pGpioPtr = &gpio_card;

	if (irq == pGpioPtr->PPCFanIrqLevel)
	{

		// Clear the fan interrupt
		(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_FAN_INT_STATUS, &temp);
	   	(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_FAN_INT_MASK, &orig_mask);
		pGpioPtr->GPIOFanInterruptStatus = temp;
		if (temp == 0)
		{
			// Spurious interrupt
			printk("Fan spurious interrupt occurred, interrupt status: 0x%x.\n", temp);
			status = NON_GPIO_INTERRUPT;
		}
		temp &= ~orig_mask;
		if(temp != 0)		//only add new events to thw queue
		{
			// Alway queue the status, so the ioctl can get them

        		//spin_lock_irqsave(pGpioPtr->fan_controlLock_irqlock, gpioflags);
			status = QueueISRStatus(&fan_intr_queue,
				pGpioPtr->GPIOFanInterruptStatus,
				&FanInterruptSemaphore);

			// Clear the fan interrupt
			(void) CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_FAN_INT_STATUS, temp); 	
		}
	}
	if (irq == pGpioPtr->PPCShftgIrqLevel)
	{
		(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_SHIFTG_INT_BASE, &temp);
		(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_SHIFTG_INT_MASK_BASE, &orig_mask);	//this is really an enable not a mask!
		pGpioPtr->GPIOShiftyInterruptStatus = temp;
		interrupt_to_mask = ~temp & orig_mask; 
		(void) CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_SHIFTG_INT_MASK_BASE, interrupt_to_mask); //disable the incomming bits
		(void) CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_SHIFTG_INT_BASE, temp); 				  //clear the register
	
		if (temp == 0)
		{
			//printk("Shifty spurious interrupt occurred, interrupt status: 0x%x.\n", temp);
			// Spurious interrupt
			status = NON_GPIO_INTERRUPT;
		}
		if(temp	& orig_mask) 
		{
			// Clear the shifty interrupt
			// Write 1 will clear the interrupt
			pGpioPtr->GPIOShiftyInterruptStatus = temp;
			status = QueueISRStatus(&shifty_intr_queue,
				pGpioPtr->GPIOShiftyInterruptStatus,
				&ShftgInterruptSemaphore);
		}
	}
	if (irq == pGpioPtr->PPCKVMIrqLevel)
	{
		(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_KVM_INT_BASE, &temp);
	   	(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_KVM_INT_MASK_BASE, &orig_mask);

		if (temp & 0x8000)
		{
			// We need to mask these interrupts for IOM 
			// for now.
			// Save the status and tell the applications
			//printk("GPIO ISR:  masking KVM interrupts: 0x%x\n", temp);
			if (pGpioPtr->GPIOFirstIO == 0)
			   pGpioPtr->GPIOKVMInterruptInitialStatus = temp;
			(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_KVM_INT_MASK_BASE, &interrupt_to_mask);
			(void) CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_KVM_INT_MASK_BASE, interrupt_to_mask | 0x8000);
		}

#if 0
		if (temp & 0x2000)
		{
			// KVM has SMTO error
			// Need to recover the bus
			printk("GPIO ISR, IOM 1, status: 0x%x\n", temp);
			printk("GPIO ISR, waking up kernel thread\n");
			// Don't mask the interrupt yet
			// We need the interrupts when the bus
			// fails again
			pGpioPtr->PPCBusIndex = 4;
			xilinx_recovery_event_present = 1;
			wake_up_interruptible(&XilinxRecoverySemaphore);
		}
#endif
		pGpioPtr->GPIOKVMInterruptStatus = temp; /* Status register to queue*/
		if (temp == 0)
		{
			// Spurious interrupt
			//printk("KVM spurious interrupt occurred, interrupt status: 0x%x.\n", temp);
			status = NON_GPIO_INTERRUPT;
		}
		temp &= ~orig_mask;
		if(temp != 0)		//only add new events to thw queue
		{
			gpio_i2c_dbg("Driver, ISR, CP calling QueueISRStatus: 0x%x\n", pGpioPtr->GPIOKVMInterruptStatus); 
			status = QueueISRStatus(&kvm_intr_queue,
				pGpioPtr->GPIOKVMInterruptStatus,
				&KVMInterruptSemaphore);

			(void) CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_KVM_INT_BASE, temp & 0x8001); 	
		}

	}
	if (irq == pGpioPtr->PPCBladeIrqLevel)
	{
		status = GPIO_INTERRUPT;
		(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_BLADE_INT_BASE, &temp);
	   	(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_BLADE_INT_MASK_BASE, &orig_mask);
		pGpioPtr->GPIOBladeInterruptStatus = temp;
		if (temp == 0)
		{
			// Spurious interrupt
			//printk("Blade spurious interrupt occurred, interrupt status: 0x%x.\n", temp);
			status = NON_GPIO_INTERRUPT;
		}
		temp &= ~orig_mask;
		if(temp != 0)		//only add new events to thw queue
		{
			gpio_i2c_dbg("Driver, ISR, blade calling QueueISRStatus: 0x%x\n", pGpioPtr->GPIOBladeInterruptStatus); 
			status = QueueISRStatus(&blade_intr_queue,
				pGpioPtr->GPIOBladeInterruptStatus,
				&BladeInterruptSemaphore);

			// Clear the blade interrupt
			(void) CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_BLADE_INT_BASE, temp); 	
		}
	}
	if (irq == pGpioPtr->PPCPSUIrqLevel)
	{
		// Check for PSU alert first
		(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_PSU_INT_BASE, &temp);
		(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_PSU_INT_MASK_BASE, &orig_mask);
		if (temp & 0xc040)
		{
			// We need to mask these interrupts for PSU 
			// for now.
			// GPIO_PRINTF("GPIO ISR:  masking PSU interrupts: 0x%x\n", temp);
			if (pGpioPtr->GPIOFirstIO == 0)
			   pGpioPtr->GPIOPSUInterruptInitialStatus = temp;
			// EXPECT: Application to unmask on service [BLR]
			(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_PSU_INT_MASK_BASE, &interrupt_to_mask);
			interrupt_to_mask |= 0xc000;
			interrupt_to_mask |= 0x0040; // disable epp interrupt
			(void) CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_PSU_INT_MASK_BASE, interrupt_to_mask);
		}
 
#if 0
		if (temp & 0x0c00)
		{
			// Either PSU1 or PSU2 has SMTO error
			// Need to recover the bus
			if (temp & 0x0800)
			{ // SMTO error for PSU 1
				printk("GPIO ISR, PSU 1, status: 0x%x\n", temp);
				printk("GPIO ISR, waking up kernel thread\n");
				// Don't mask the interrupt yet
				// We need the interrupts when the bus
				// fails again
				pGpioPtr->PPCBusIndex = 2;
				xilinx_recovery_event_present = 1;
				wake_up_interruptible(&XilinxRecoverySemaphore);
			}
			else
			{
				printk("GPIO ISR, PSU 2, status: 0x%x\n", temp);
				printk("GPIO ISR, waking up kernel thread\n");
				// Don't mask the interrupt yet
				// We need the interrupts when the bus
				// fails again
				pGpioPtr->PPCBusIndex = 3;
				xilinx_recovery_event_present = 1;
				wake_up_interruptible(&XilinxRecoverySemaphore);
			}
		}
#endif

		// At this point, the I2c interrupts are all out, so
		// it's only our interrupt.
		pGpioPtr->GPIOPSUInterruptStatus = temp;
		if (temp == 0)
		{
			// Spurious interrupt
			//printk("PSU spurious interrupt occurred, interrupt status: 0x%x.\n", temp);
			status = NON_GPIO_INTERRUPT;
		}
		temp &= 0xC07F;	//mask off i2c type interrupts
		temp &= ~orig_mask;
		if(temp != 0)		//only add new events to the queue
		{
			// Clear the PSU interrupt
			// Write 1 will clear the interrupt
			gpio_i2c_dbg("Driver, ISR, calling QueueISRStatus: 0x%x\n", pGpioPtr->GPIOPSUInterruptStatus); 
			GPIO_PRINTF("Test \n"); 
			
			updatePSUCounter(&psu_intr_queue,pGpioPtr->GPIOPSUInterruptStatus);
			status = QueueISRStatus((struct intr_status_t*)&psu_intr_queue,
				pGpioPtr->GPIOPSUInterruptStatus,
				&PSUInterruptSemaphore);
			(void) CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_PSU_INT_BASE, temp & 0xC07F);
			// logging epp interrupt on GRID failure
			(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_EPP_EN_BASE, &eppenablebit);
			if (eppenablebit) {
                        	(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_EPP_EN_BASE, &eppenablebit);
                        	(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_PSU_INT_BASE, &intrstatus);
                        	(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_PSU_INT_MASK_BASE, &intrmask);
                        	GPIO_PRINTF("GPIO ISR (epp enable bit: 0x%x, interrupt status: 0x%x, interrupt mask: 0x%x)\n", 
					eppenablebit, 
					intrstatus, 
					intrmask);
			}

			status = GPIO_INTERRUPT;
		}

	}
	if (irq == pGpioPtr->PPCCPIrqLevel)
	{
		//heres the deal:
		//on interrupt write the status register back to clear all but the alert
		//no interrutps should be masked except the alert
		//the app needs to clear the masked alert when it has taken care of it
					
		
		(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_CP_INT_BASE, &temp);
		pGpioPtr->GPIOCPInterruptStatus = temp;
		(void) CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_CP_INT_BASE, temp & 0xfb);	//clear the edge interrupts 	

	   	//gpio_i2c_dbg("Driver, ISR, CP calling QueueISRStatus: 0x%x\n", pGpioPtr->GPIOCPInterruptStatus); 
	   	(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_CP_INT_MASK_BASE, &orig_mask);
	   	pGpioPtr->GPIOCPInterruptStatus = temp;	
//	   	printk("Driver, ISR, CP calling QueueISRStatus: 0x%x [mask: %x]\n", pGpioPtr->GPIOCPInterruptStatus, orig_mask); 
		if(temp & ~orig_mask)	//only add new events to the queue
		{
			if(temp & 0x04)
			{
	   			(void) CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_CP_INT_MASK_BASE, orig_mask|0x04);	//mask off the alert

			}
			status = QueueISRStatus(&cp_intr_queue,
				pGpioPtr->GPIOCPInterruptStatus,
				&CPInterruptSemaphore);
		}
	   	status = GPIO_INTERRUPT;
	}
	if (irq == pGpioPtr->PPCIOMIrqLevel)
	{
		// Clear the IOM interrupt
		(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_IOM_INT_BASE, &temp);
		(void) CMCGpioReadReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_IOM_INT_MASK_BASE, &orig_mask);

		temp &= 0xC03F; /*Hide I2C events from GPIO Driver!*/
		if (temp & 0xc000)
		{
			// Save the status and tell the applications
			//printk("GPIO ISR:  masking IOM interrupts: 0x%x\n", temp);
			pGpioPtr->GPIOIOMInterruptStatus = temp;
			if (pGpioPtr->GPIOFirstIO == 0)
				pGpioPtr->GPIOIOMInterruptInitialStatus = temp;
			interrupt_to_mask = (temp & 0xc000) | orig_mask;
			(void) CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_IOM_INT_MASK_BASE, interrupt_to_mask);
		}
#if 0
		if (temp & 0x0c00)
		{
			// Either IOM1 or IOM2 has SMTO error
			// Need to recover the bus
			if (temp & 0x0800)
			{ // SMTO error for IOM 1
				printk("GPIO ISR, IOM 1, status: 0x%x\n", temp);
				printk("GPIO ISR, waking up kernel thread\n");
				// Don't mask the interrupt yet
				// We need the interrupts when the bus
				// fails again
				pGpioPtr->PPCBusIndex = 0;
				xilinx_recovery_event_present = 1;
				wake_up_interruptible(&XilinxRecoverySemaphore);
			}
			else
			{
				printk("GPIO ISR, IOM 2, status: 0x%x\n", temp);
				printk("GPIO ISR, waking up kernel thread\n");
				// Don't mask the interrupt yet
				// We need the interrupts when the bus
				// fails again
				pGpioPtr->PPCBusIndex = 1;
				xilinx_recovery_event_present = 1;
				wake_up_interruptible(&XilinxRecoverySemaphore);
			}
		}
#endif

		pGpioPtr->GPIOIOMInterruptStatus = temp;
		// Write 1 will clear the interrupt
		if (temp == 0)
		{
			// Spurious interrupt
			//printk("IOM spurious interrupt occurred, interrupt status: 0x%x.\n", temp);
			status = NON_GPIO_INTERRUPT;
			return status;
		}
		temp &= 0xC03F;
		temp &= ~orig_mask;
		if(temp != 0)
		{
			gpio_i2c_dbg("Driver, ISR, IOM calling QueueISRStatus: 0x%x\n", pGpioPtr->GPIOIOMInterruptStatus); 
			// Clear the IOM interrupt
			// Write 1 will clear the interrupt
			status = QueueISRStatus(&iom_intr_queue,
				pGpioPtr->GPIOIOMInterruptStatus,
				&IOMInterruptSemaphore);

			(void) CMCGpioWriteReg(CMC_GPIO_BASE_ADDRESS, CMC_GPIO_IOM_INT_BASE, temp & 0xC03F); 	
			status = GPIO_INTERRUPT;
		}
	}
	if (irq == pGpioPtr->PPCECCIrqLevel)
	{
			// Need to disable the ecc temporarily
			mtdcr(ddrcfga, DDR0_22);
			val = mfdcr(ddrcfgd);
			new_val = (~DDR0_22_CTRL_RAW_MASK & val);
			mtdcr(ddrcfgd, new_val);

			// Now get the ecc status register
			mtdcr(ddrcfga, DDR0_00);
			temp = mfdcr(ddrcfgd);

			// Clear ecc interrupt, change INT_ACK parameter
			mtdcr(ddrcfga, DDR0_00);
			new_val = (mfdcr(ddrcfgd) | DDR0_00_INT_ACK_ALL);
			mtdcr(ddrcfgd, new_val);

			// Re-enable ecc interrupt, restore the CTRL_RAW
			// parameter
			mtdcr(ddrcfga, DDR0_22);
			mtdcr(ddrcfgd, val);

		pGpioPtr->GPIOECCInterruptStatus = temp;

		ECC_event_present = 1;
		wake_up_interruptible(&ECCInterruptSemaphore);
		status = GPIO_INTERRUPT;
	}
	return status;
		
}

int
GpioDpcForISR( int irq)
{
	u32	status;
	GPIO_DEVICE_EXTENSION *pGpioPtr;

	//spin_lock_irqsave(&ControlLock_irqlock, gpioflags);

	pGpioPtr = &gpio_card;
	pGpioPtr->InterruptCount++;


	// Clear the interrupt
       	status = GpioClearDevInt(irq);


	//spin_unlock_irqrestore(&ControlLock_irqlock, gpioflags);
	return status;

} /* GpioDpcForISR */
//---------------------------------------------------------------------
// GpioLoadDevInterrup
// Description: Enable the interrupt
//---------------------------------------------------------------------
s32 GpioLoadDevInterrupt(PGPIO_DEVICE_EXTENSION pGpioCard)
{
	int Status;


	// Enable Interrupts if applicable


        /*
         * Enable the interrupt handler
         */
	printk("GpioLoadDevInterrupt, registering interrupt for Fan, irq: %d\n",
					pGpioCard->PPCFanIrqLevel);
        if ((Status = request_irq((unsigned int)pGpioCard->PPCFanIrqLevel, GpioIsr,
                //SA_SHIRQ, "gpio", pGpioCard))
		// 2.6.24
                IRQF_SHARED, "gpio", pGpioCard))
		!= GPIO_STATUS_CMD_SUCCESS)
        {
              		printk(KERN_WARNING "gpio_init_card: unable to install interrupt handler - skipping controller\n");
			//init_waitqueue_head(&FanInterruptSemaphore);
			return Status;
        }
	printk("GpioLoadDevInterrupt, registering interrupt for shifty, irq: %d\n",
					pGpioCard->PPCShftgIrqLevel);
        if ((Status = request_irq((unsigned int)pGpioCard->PPCShftgIrqLevel, GpioIsr,
                //SA_SHIRQ, "gpio", pGpioCard))
		// 2.6.24
                IRQF_SHARED, "gpio", pGpioCard))
		!= GPIO_STATUS_CMD_SUCCESS)
        {
              		printk(KERN_WARNING "gpio_init_card: unable to install interrupt handler - skipping controller\n");
			//init_waitqueue_head(&ShftgInterruptSemaphore);
			return Status;
        }
	printk("GpioLoadDevInterrupt, registering interrupt for KVM, irq: %d\n",
					pGpioCard->PPCKVMIrqLevel);
        if ((Status = request_irq((unsigned int)pGpioCard->PPCKVMIrqLevel, GpioIsr,
                //SA_SHIRQ, "gpio", pGpioCard))
		// 2.6.24
                IRQF_SHARED, "gpio", pGpioCard))
		!= GPIO_STATUS_CMD_SUCCESS)
        {
              		printk(KERN_WARNING "gpio_init_card: unable to install interrupt handler - skipping controller\n");
			//init_waitqueue_head(&KVMInterruptSemaphore);
			return Status;
        }
	printk("GpioLoadDevInterrupt, registering interrupt for IOM, irq: %d\n",
					pGpioCard->PPCIOMIrqLevel);
        if ((Status = request_irq((unsigned int)pGpioCard->PPCIOMIrqLevel, GpioIsr,
                //SA_SHIRQ, "gpio", pGpioCard))
		// 2.6.24
                IRQF_SHARED, "gpio", pGpioCard))
		!= GPIO_STATUS_CMD_SUCCESS)
        {
              		printk(KERN_WARNING "gpio_init_card: unable to install interrupt handler - skipping controller\n");
			//init_waitqueue_head(&IOMInterruptSemaphore);
			return Status;
        }
	printk("GpioLoadDevInterrupt, registering interrupt for ECC, irq: %d\n",
					pGpioCard->PPCECCIrqLevel);
        if ((Status = request_irq((unsigned int)pGpioCard->PPCECCIrqLevel, GpioIsr,
                //SA_SHIRQ, "gpio", pGpioCard))
		// 2.6.24
                IRQF_SHARED, "gpio", pGpioCard))
		!= GPIO_STATUS_CMD_SUCCESS)
        {
              		printk(KERN_WARNING "gpio_init_card: unable to install interrupt handler - skipping controller\n");
			//init_waitqueue_head(&ECCInterruptSemaphore);
			return Status;
        }
	printk("GpioLoadDevInterrupt, registering interrupt for PSU, irq: %d\n",
					pGpioCard->PPCPSUIrqLevel);
        if ((Status = request_irq((unsigned int)pGpioCard->PPCPSUIrqLevel, GpioIsr,
                //SA_SHIRQ, "gpio", pGpioCard))
		// 2.6.24
                IRQF_SHARED, "gpio", pGpioCard))
		!= GPIO_STATUS_CMD_SUCCESS)
        {
              		printk(KERN_WARNING "gpio_init_card: unable to install interrupt handler - skipping controller\n");
			//init_waitqueue_head(&PSUInterruptSemaphore);
			return Status;
        }
	printk("GpioLoadDevInterrupt, registering interrupt for CP, irq %d\n",
					pGpioCard->PPCCPIrqLevel);
        if ((Status = request_irq((unsigned int)pGpioCard->PPCCPIrqLevel, GpioIsr,
                //SA_SHIRQ, "gpio", pGpioCard))
		// 2.6.24
                IRQF_SHARED, "gpio", pGpioCard))
		!= GPIO_STATUS_CMD_SUCCESS)
        {
              		printk(KERN_WARNING "gpio_init_card: unable to install interrupt handler - skipping controller\n");
			//init_waitqueue_head(&CPInterruptSemphore);
			return Status;
        }

	
	return GPIO_STATUS_CMD_SUCCESS;
} // GpioLoadDevInterrupt

//---------------------------------------------------------------------
// GpioUnloadDevInterrupt
//---------------------------------------------------------------------
void GpioUnloadDevInterrupt(PGPIO_DEVICE_EXTENSION pGpioCard)
{
	// disconnect ISR
	// so that we do not get sprurios interrupts
	// and try to service it

	// disconnect the interrupt first
	free_irq((unsigned int)pGpioCard->PPCFanIrqLevel, pGpioCard);
	free_irq((unsigned int)pGpioCard->PPCShftgIrqLevel, pGpioCard);
	free_irq((unsigned int)pGpioCard->PPCKVMIrqLevel, pGpioCard);
	free_irq((unsigned int)pGpioCard->PPCBladeIrqLevel, pGpioCard);
	free_irq((unsigned int)pGpioCard->PPCIOMIrqLevel, pGpioCard);
	free_irq((unsigned int)pGpioCard->PPCECCIrqLevel, pGpioCard);
	free_irq((unsigned int)pGpioCard->PPCPSUIrqLevel, pGpioCard);
	free_irq((unsigned int)pGpioCard->PPCCPIrqLevel, pGpioCard);

	//init_waitqueue_head(&FanInterruptSemaphore);

} // EsmUnloadDevInterrupt

char *fifotype(struct intr_status_t *intr_queue)
{
	if (intr_queue == &blade_intr_queue)
		return "Blade";
	else if (intr_queue == &fan_intr_queue)
		return "Fan";
	else if (intr_queue == &iom_intr_queue)
		return "IOM";
	else if (intr_queue == (struct intr_status_t*)&psu_intr_queue)
		return "PSU";
	else if (intr_queue == &cp_intr_queue)
		return "CP";
	else if (intr_queue == &kvm_intr_queue)
		return "KVM";
	return "Shifty";
}

void updatePSUCounter(struct psu_intr_status_t* psu_intr_queue , unsigned int int_register_value)
{
	int count = 0;
	//printk("Interrupt value is 0x%x\n",int_register_value);
	if(int_register_value & EPP_ENABLE ) psu_intr_queue->psu_in_count[0] ++; 
	if(int_register_value & PSU_1  ) psu_intr_queue->psu_in_count[1] ++; 
	if(int_register_value & PSU_2  ) psu_intr_queue->psu_in_count[2] ++;
	if(int_register_value & PSU_3  ) psu_intr_queue->psu_in_count[3] ++;
	if(int_register_value & PSU_4  ) psu_intr_queue->psu_in_count[4] ++;
	if(int_register_value & PSU_5  ) psu_intr_queue->psu_in_count[5] ++;
	if(int_register_value & PSU_6  ) psu_intr_queue->psu_in_count[6] ++;

}
u32 QueueISRStatus(struct intr_status_t *intr_queue, u32 intr_status,
			wait_queue_head_t *semp)
{
	int i;

	if (ISRFULLFIFO(intr_queue)) {
		if (!intr_queue->overflows++)
			printk("hapi %s isr fifo overflowed\n", fifotype(intr_queue));
		/* increment and wrap pointer, if necessary */
		if (++intr_queue->out >= intr_queue->fifosize)
			intr_queue->out = 0;
	}

	/* store value */
	intr_queue->fifo[intr_queue->in] = intr_status;

	/* increment and wrap pointer, if necessary */
	if (++intr_queue->in >= intr_queue->fifosize)
		intr_queue->in = 0;

	if (intr_queue->waiting) {
		intr_queue->waiting = 0;
		wake_up_interruptible(semp);
	}

	/* statistics */
	i = intr_queue->in - intr_queue->out;
	if (i < 0)
		i = intr_queue->fifosize + i;
	if (i > intr_queue->watermark)
		intr_queue->watermark = i;
	intr_queue->incnt++;

	return GPIO_INTERRUPT;

}
#endif // GPIO with interrupt
