/*
 * (C) Copyright 2006
 * Stefan Roese, DENX Software Engineering, sr@denx.de.
 *
 * (C) Copyright 2006
 * Jacqueline Pira-Ferriol, AMCC/IBM, jpira-ferriol@fr.ibm.com
 * Alain Saurel,	    AMCC/IBM, alain.saurel@fr.ibm.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <asm/processor.h>
#include <ppc440.h>
#include <util_defs.h>
#include <util_flashmap.h>
#include <util_fpga.h>
#include <util_ret_codes.h>

#include <sdram.h>


/*****************************************************************************
*	Prototypes
*****************************************************************************/
static u32 membus(char *msg);
static u32 memaddr(char *msg);
static u32 mempat(char *msg);
static u32 core_death(char *msg);

extern s32  g_post_rc;
extern char g_post_msg[];



DECLARE_GLOBAL_DATA_PTR;

extern flash_info_t flash_info[CFG_MAX_FLASH_BANKS]; /* info for FLASH chips	*/

/* pw start adds for HW flow control*/
/* Remove unfortunate and unused (in this file) #defines from ppc440.h */
#undef tcr
#undef tsr
struct config_gpio_regs {
	u32 tcr;
	u32 odr;
	u32 tsr;
	u32 osr;
	u32 isr1;
	u32 isr2;
	u32 isr3;
};

static void config_gpio_init(int num, u32 *bit, u32 *mask, uint *shift,
			struct config_gpio_regs *cfg)
{
	if (num < 16) {
		*bit = PPC_BIT(num);
		*mask = PPC_BIT(2*num) | PPC_BIT(2*num + 1);
		*shift = 31 - (2*num + 1);
		cfg->tcr = GPIO0_TCR;
		cfg->odr = GPIO0_ODR;
		cfg->tsr = GPIO0_TSRL;
		cfg->osr = GPIO0_OSRL;
		cfg->isr1 = GPIO0_ISR1L;
		cfg->isr2 = GPIO0_ISR2L;
		cfg->isr3 = GPIO0_ISR3L;
	} else if (num < 32) {
		*bit = PPC_BIT(num);
		*mask = PPC_BIT(2*num - 32) | PPC_BIT((2*num - 32) + 1);
		*shift = 31 - ((2*num - 32) + 1);
		cfg->tcr = GPIO0_TCR;
		cfg->odr = GPIO0_ODR;
		cfg->tsr = GPIO0_TSRH;
		cfg->osr = GPIO0_OSRH;
		cfg->isr1 = GPIO0_ISR1H;
		cfg->isr2 = GPIO0_ISR2H;
		cfg->isr3 = GPIO0_ISR3H;
	} else if (num < 48) {
		*bit = PPC_BIT(num-32);
		*mask = PPC_BIT(2*num - 64) | PPC_BIT((2*num - 64) + 1);
		*shift = 31 - ((2*num - 64) + 1);
		cfg->tcr = GPIO1_TCR;
		cfg->odr = GPIO1_ODR;
		cfg->tsr = GPIO1_TSRL,
		cfg->osr = GPIO1_OSRL;
		cfg->isr1 = GPIO1_ISR1L;
		cfg->isr2 = GPIO1_ISR2L;
		cfg->isr3 = GPIO1_ISR3L;
	} else {
		*bit = PPC_BIT(num-32);
		*mask = PPC_BIT(2*num - 96) | PPC_BIT((2*num - 96) + 1);
		*shift = 31 - ((2*num - 96) + 1);
		cfg->tcr = GPIO1_TCR;
		cfg->odr = GPIO1_ODR;
		cfg->tsr = GPIO1_TSRH;
		cfg->osr = GPIO1_OSRH;
		cfg->isr1 = GPIO1_ISR1H;
		cfg->isr2 = GPIO1_ISR2H;
		cfg->isr3 = GPIO1_ISR3H;
	}
}

static void config_gpio_input_normal(int num)
{
	uint shift;
	u32 bit, mask;
	struct config_gpio_regs cfg;

	config_gpio_init(num, &bit, &mask, &shift, &cfg);
	GPIO_CLR_BITS(cfg.tcr, bit);
	GPIO_CLR_BITS(cfg.odr, bit);
	GPIO_CLR_BITS(cfg.tsr, mask);
	GPIO_CLR_BITS(cfg.isr1, mask);
	GPIO_CLR_BITS(cfg.isr2, mask);
	GPIO_CLR_BITS(cfg.isr3, mask);
}

static void config_gpio_input_alternate(int num, int altnum, int altval)
{
	uint shift;
	u32 bit, mask;
	struct config_gpio_regs cfg;

	if (altval > 3)
		printf("config_gpio_input_alternate: altval %d out of range!\n", altval);
	config_gpio_init(num, &bit, &mask, &shift, &cfg);
	GPIO_CLR_BITS(cfg.tcr, bit);
	GPIO_CLR_BITS(cfg.tsr, mask);
	switch (altnum) {
	case 1:
		GPIO_SET_MASK(cfg.isr1, mask, altval << shift);
		GPIO_CLR_BITS(cfg.isr2, mask);
		GPIO_CLR_BITS(cfg.isr3, mask);
		break;
	case 2:
		GPIO_CLR_BITS(cfg.isr1, mask);
		GPIO_SET_MASK(cfg.isr2, mask, altval << shift);
		GPIO_CLR_BITS(cfg.isr3, mask);
		break;
	case 3:
		GPIO_CLR_BITS(cfg.isr1, mask);
		GPIO_CLR_BITS(cfg.isr2, mask);
		GPIO_SET_MASK(cfg.isr3, mask, altval << shift);
		break;
	}
}

static void config_gpio_output_normal(int num)
{
	uint shift;
	u32 bit, mask;
	struct config_gpio_regs cfg;

	config_gpio_init(num, &bit, &mask, &shift, &cfg);
	GPIO_SET_BITS(cfg.tcr, bit);
	GPIO_CLR_BITS(cfg.odr, bit);
	GPIO_CLR_BITS(cfg.tsr, mask);
	GPIO_CLR_BITS(cfg.osr, mask);
}

static void config_gpio_output_alternate(int num, int altnum, int altval)
{
	uint shift;
	u32 bit, mask;
	struct config_gpio_regs cfg;

	if (altval > 3)
		printf("config_gpio_output_alternate: altval %d out of range!\n", altval);
	config_gpio_init(num, &bit, &mask, &shift, &cfg);
	GPIO_SET_BITS(cfg.tcr, bit);
	GPIO_CLR_BITS(cfg.tsr, mask);
	GPIO_SET_MASK(cfg.osr, mask, altval << shift);
}
/* pw end adds for HW flow control*/


int board_early_init_f(void)
{
	unsigned long sdr0_cust0;
	unsigned long sdr0_pfc1, sdr0_pfc2;
	register uint reg;

	mtdcr(ebccfga, xbcfg);
	mtdcr(ebccfgd, 0xb8400000);

	/*--------------------------------------------------------------------
	 * Setup the GPIO pins
	 *-------------------------------------------------------------------*/

/* 
*** GPIO 0 ***
GPIO BIT/DIR    TCR     OSRL       TSRL       ISR1L      ISR2L       ISR3L 
   DIR: NAME:   [BIT/S] VAL
0  I CMCF_CMCL_ [0]  0  [0:1]   00 [0:1]   00 [0:1]   00 [0:1]   00  [0:1]   00  
     HEARTBEAT
1  I CMCF_CMCL_ [1]  0  [2:3]   00 [2:3]   00 [2:3]   00 [2:3]   00  [2:3]   00
     ACTIVE
2  I PPC_GPIO_  [2]  0  [4:5]   00 [4:5]   00 [4:5]   00 [4:5]   00  [4:5]   00
     INIT_B 
3  O LED_AMBER  [3]  1  [6:7]   00 [6:7]   00 [6:7]   00 [6:7]   00  [6:7]   00
4  O LED_BLUE   [4]  1  [8:9]   00 [8:9]   00 [8:9]   00 [8:9]   00  [8:9]   00
5  O LED_POWER  [5]  1  [10:11] 00 [10:11] 00 [10:11] 00 [10:11] 00  [10:11] 00
6  x ALT1/cs1   [6]  0  [12:13] 01 [12:13] 01 [12:13] 00 [12:13] 00  [12:13] 00
7  X ALT1/cs2   [7]  0  [14:15] 01 [14:15] 01 [14:15] 00 [14:15] 00  [14:15] 00
8  0 CPLD_TMS   [8]  1  [16:17] 00 [16:17] 00 [16:17] 00 [16:17] 00  [16:17] 00
9  O CPLD_THERM [9]  1  [18:19] 00 [18:19] 00 [18:19] 00 [18:19] 00  [18:19] 00
     TRIP_CLEAR_N
10 I PPC_GPIO_  [10] 0  [20:21] 00 [20:21] 00 [20:21] 00 [20:21] 00  [20:21] 00
     BUSY 
11 O CMCL_CMCF_ [11] 1  [22:23] 00 [22:23] 00 [22:23] 00 [22:23] 00  [22:23] 00
     HEARTBEAT
12 X N/A        [12] 0  [24:25] 00 [24:25] 00 [24:25] 00 [24:25] 00  [24:25] 00
13 X N/A        [13] 0  [26:27] 00 [26:27] 00 [26:27] 00 [26:27] 00  [26:27] 00
14 X N/A        [14] 0  [28:29] 00 [28:29] 00 [28:29] 00 [28:29] 00  [28:29] 00
15 X N/A        [15] 0  [30:31] 00 [30:31] 00 [30:31] 00 [30:31] 00  [30:31] 00

                TCR     OSRH       TSRH       ISR1H       ISR2H      ISR3H
16 I CMC_REV0   [16] 0  [0:1]   00 [0:1]   00 [0:1]   00 [0:1]   00  [0:1]   00 
17 0 DEBUG_LED  [17] 1  [2:3]   00 [2:3]   00 [2:3]   00 [2:3]   00  [2:3]   00
18 I UART_DCD   [18] 0  [4:5]   00 [4:5]   00 [4:5]   00 [4:5]   00  [4:5]   00
19 I CMC_REV2   [19] 0  [6:7]   00 [6:7]   00 [6:7]   00 [6:7]   00  [6:7]   00
20 I PPC_GPIO_  [20] 0  [8:9]   00 [8:9]   00 [8:9]   00 [8:9]   00  [8:9]   00
     DONE 
21 I UART_RI    [21] 0  [10:11] 00 [10:11] 00 [10:11] 00 [10:11] 00  [10:11] 00
22 X N/A        [22] 0  [12:13] 00 [12:13] 00 [12:13] 00 [12:13] 00  [12:13] 00
23 0 FPGA_LOGIC_[23] 1  [14:15] 00 [14:15] 00 [14:15] 00 [14:15] 00  [14:15] 00
     RESET_N
24 I PWORD_N    [24] 0  [16:17] 00 [16:17] 00 [16:17] 00 [16:17] 00  [16:17] 00
25 I CMC_REV1   [25] 0  [18:19] 00 [18:19] 00 [18:19] 00 [18:19] 00  [18:19] 00 
26 X N/A        [26] 0  [20:21] 00 [20:21] 00 [20:21] 00 [20:21] 00  [20:21] 00
27 O CMC_BUF_EN [27] 1  [22:23] 00 [22:23] 00 [22:23] 00 [22:23] 00  [22:23] 00
28 O CMCL_CMCF_ [28] 1  [24:25] 00 [24:25] 00 [24:25] 00 [24:25] 00  [24:25] 00
     ACTIVE
29 O PPC_UART0_ [29] 1  [26:27] 00 [26:27] 00 [26:27] 00 [26:27] 00  [26:27] 00
     DTR
30 I PPC_UART0_ [30] 0  [28:29] 00 [28:29] 00 [28:29] 00 [28:29] 00  [28:29] 00
     DSR
31 O UART_CMC_  [31] 1  [30:31] 00 [30:31] 00 [30:31] 00 [30:31] 00  [30:31] 00
     IOM_DTR_B


***** GPIO 1 *****
GPIO BIT/DIR    TCR     OSRL       TSRL       ISR1L      ISR2L       ISR3L 
   DIR: NAME:   [BIT/S] VAL
32 I UART_IOM_  [0]  0  [0:1]   00 [0:1]   00 [0:1]   00 [0:1]   00  [0:1]   00  
     CMC_DSR_N
33 O CMC_INPUT_ [1]  1  [2:3]   00 [2:3]   00 [2:3]   00 [2:3]   00  [2:3]   00 
     EN_N
34 I UART_IOM_  [2]  0  [4:5]   00 [4:5]   00 [4:5]   00 [4:5]   01  [4:5]   00
     CMC_CTS_B
35 O UART_CMC_  [3]  1  [6:7]   10 [6:7]   00 [6:7]   00 [6:7]   00  [6:7]   00
     IOM_RTS
36 X N/A        [4]  0  [8:9]   00 [8:9]   00 [8:9]   00 [8:9]   00  [8:9]   00
37 X N/A        [5]  0  [10:11] 00 [10:11] 00 [10:11] 00 [10:11] 00  [10:11] 00
38 O UART_CMC_  [6]  1  [12:13] 10 [12:13] 00 [12:13] 00 [12:13] 00  [12:13] 00
     IOM_TX
39 I UART_IOM_  [7]  0  [14:15] 00 [14:15] 00 [14:15] 00 [14:15] 01  [14:15] 00
     CMC_RX
40 I IRQ0 FAN   [8]  0  [16:17] 00 [16:17] 00 [16:17] 01 [16:17] 00  [16:17] 00
41 I IRQ1 RTC   [9]  0  [18:19] 00 [18:19] 00 [18:19] 01 [18:19] 00  [18:19] 00
42 I IRQ2 TEMP  [10] 0  [20:21] 00 [20:21] 00 [20:21] 01 [20:21] 00  [20:21] 00
43 I IRQ3 SHFTE [11] 0  [22:23] 00 [22:23] 00 [22:23] 01 [22:23] 00  [22:23] 00
44 I IRQ4 KVM   [12] 0  [24:25] 00 [24:25] 00 [24:25] 01 [24:25] 00  [24:25] 00
45 I IRQ6 IOM   [13] 0  [26:27] 00 [26:27] 00 [26:27] 01 [26:27] 00  [26:27] 00
46 I IRQ7 PSU   [14] 0  [28:29] 00 [28:29] 00 [28:29] 01 [28:29] 00  [28:29] 00
47 I IRQ8 CP    [15] 0  [30:31] 00 [30:31] 00 [30:31] 01 [30:31] 00  [30:31] 00

                TCR     OSRH       TSRH       ISR1H       ISR2H      ISR3H
48 O KVM_RST_N  [16] 1  [0:1]   00 [0:1]   00 [0:1]   00 [0:1]   00  [0:1]   00
49 I FLASH_     [17] 0  [2:3]   00 [2:3]   00 [2:3]   00 [2:3]   00  [2:3]   00
     RECOVER_N
50 I MFG_MODE_N [18] 0  [4:5]   00 [4:5]   00 [4:5]   00 [4:5]   00  [4:5]   00
51 O SWITCH_    [19] 1  [6:7]   00 [6:7]   00 [6:7]   00 [6:7]   00  [6:7]   00
     RESET_N
52 0 CPLD_TCK   [20] 1  [8:9]   00 [8:9]   00 [8:9]   00 [8:9]   00  [8:9]   00
53 O CPLD_TDI   [21] 1  [10:11] 00 [10:11] 00 [10:11] 00 [10:11] 00  [10:11] 00
54 I CPLD_TDO   [22] 0  [12:13] 00 [12:13] 00 [12:13] 00 [12:13] 00  [12:13] 00
55 O CP_POWER_  [23] 1  [14:15] 00 [14:15] 00 [14:15] 00 [14:15] 00  [14:15] 00
     LED
56 0 PPC_GPIO_  [24] 1  [16:17] 00 [16:17] 00 [16:17] 00 [16:17] 00  [16:17] 00
     PROG_B  
57 O I2C_PSU123 [25] 1  [18:19] 00 [18:19] 00 [18:19] 00 [18:19] 00  [18:19] 00
     _RST_N_B
58 O I2C_PSU456 [26] 1  [20:21] 00 [20:21] 00 [20:21] 00 [20:21] 00  [20:21] 00
     _RST_N_B
59 O I2C_IOMSET1[27] 1  [22:23] 00 [22:23] 00 [22:23] 00 [22:23] 00  [22:23] 00
     _RST_N_B
60 O I2C_IOMSET2[28] 1  [24:25] 00 [24:25] 00 [24:25] 00 [24:25] 00  [24:25] 00
     _RST_N_B
61 O I2C_KVM_   [29] 1  [26:27] 00 [26:27] 00 [26:27] 00 [26:27] 00  [26:27] 00
     RST_N_B
62 O I2C_CP_RST [30] 1  [28:29] 00 [28:29] 00 [28:29] 00 [28:29] 00  [28:29] 00
     _N_B
63 O PPC_MICREL [31] 1  [30:31] 00 [30:31] 00 [30:31] 00 [30:31] 00  [30:31] 00
     _EN

*/

	out32(GPIO0_OR,    0x04404100);
	out32(GPIO0_TCR,   0x1cd0411d); /*was 0x1c50411d, CPLD TMS as output*/
	out32(GPIO0_OSRL,  0x00050000);
	out32(GPIO0_OSRH,  0x00000000);
	out32(GPIO0_TSRL,  0x00050000);
	out32(GPIO0_TSRH,  0x00000000);
	out32(GPIO0_ISR1L, 0x00000000);
	out32(GPIO0_ISR1H, 0x00000000); 
	out32(GPIO0_ISR2L, 0x00000000); 
	out32(GPIO0_ISR2H, 0x00000000);
	out32(GPIO0_ISR3L, 0x00000000);
	out32(GPIO0_ISR3H, 0x00000000);

	out32(GPIO1_ODR,   0x0000007e);	/*open drain I2C *_RST_N_B signals*/
	out32(GPIO1_OR,    0x000090fe); /*MUST set PROG_B high or FPGA resets*/
	out32(GPIO1_TCR,   0x52009dff); /*was 0x520091ff, CPLD TCK, TDI as outputs*/
	out32(GPIO1_OSRL,  0x02080000);  
	out32(GPIO1_OSRH,  0x00000000);
	out32(GPIO1_TSRL,  0x00000000);
	out32(GPIO1_TSRH,  0x00000000);
	out32(GPIO1_ISR1L, 0x00005555);  /*enable IRQ 0,1,2,3,4,6,7,8*/
	out32(GPIO1_ISR1H, 0x00000000);
	out32(GPIO1_ISR2L, 0x04010000);  /*UART1 to IOM support*/
	out32(GPIO1_ISR2H, 0x00000000);
	out32(GPIO1_ISR3L, 0x00000000);
	out32(GPIO1_ISR3H, 0x00000000);

/* pw start adds for HW flow control*/
		/* UART configuration */
	config_gpio_input_normal(GPIO_BITNUM(0, GPIO_UART_DCD));
	config_gpio_input_normal(GPIO_BITNUM(0, GPIO_UART_RI));
	config_gpio_output_normal(GPIO_BITNUM(0, GPIO_PPC_UART0_DTR));
	config_gpio_input_normal(GPIO_BITNUM(0, GPIO_PPC_UART0_DSR));
	config_gpio_output_normal(GPIO_BITNUM(0, GPIO_UART_CMC_IOM_DTR_B));
	config_gpio_input_normal(GPIO_BITNUM(1, GPIO_UART_IOM_CMC_DSR_N));

	/* Alternate 1 configuration */
	config_gpio_input_alternate(GPIO_BITNUM(1, GPIO_PPC_UART0_CTS), 1, 1);
	config_gpio_output_alternate(GPIO_BITNUM(1, GPIO_PPC_UART0_RTS), 1, 1);

	/* Alternate 2 configuration */
	config_gpio_input_alternate(GPIO_BITNUM(1, GPIO_UART_IOM_CMC_CTS_B), 2, 1);
	config_gpio_input_alternate(GPIO_BITNUM(1, GPIO_UART_IOM_CMC_RX), 2, 1);
	config_gpio_output_alternate(GPIO_BITNUM(1, GPIO_UART_CMC_IOM_RTS), 2, 2);
	config_gpio_output_alternate(GPIO_BITNUM(1, GPIO_UART_CMC_IOM_TX), 2, 2);
/* pw end adds for HW flow control*/


	/*--------------------------------------------------------------------
	 * Setup the interrupt controller polarities, triggers, etc.
	 *-------------------------------------------------------------------*/
#if 1 /*origional*/	/*10-31-07 change back to level*/
	mtdcr(uic0sr, 0xffffffff);	/* clear all */
	mtdcr(uic0er, 0x00000000);	/* disable all */
	mtdcr(uic0cr, 0x00000005);	/* ATI & UIC1 crit are critical */
	mtdcr(uic0pr, 0xfffff7ff);	/* per ref-board manual */
	mtdcr(uic0tr, 0x00000000);	/* per ref-board manual */
	mtdcr(uic0vr, 0x00000000);	/* int31 highest, base=0x000 */
	mtdcr(uic0sr, 0xffffffff);	/* clear all */

	mtdcr(uic1sr, 0xffffffff);	/* clear all */
	mtdcr(uic1er, 0x00000000);	/* disable all */
	mtdcr(uic1cr, 0x00000000);	/* all non-critical */
	mtdcr(uic1pr, 0xffffffff);	/* per ref-board manual */
	mtdcr(uic1tr, 0x00000000);	/* per ref-board manual */
	mtdcr(uic1vr, 0x00000000);	/* int31 highest, base=0x000 */
	mtdcr(uic1sr, 0xffffffff);	/* clear all */

	mtdcr(uic2sr, 0xffffffff);	/* clear all */
	mtdcr(uic2er, 0x00000000);	/* disable all */
	mtdcr(uic2cr, 0x00000000);	/* all non-critical */
	mtdcr(uic2pr, 0xffffffff);	/* per ref-board manual */
	mtdcr(uic2tr, 0x00000000);	/* per ref-board manual */
	mtdcr(uic2vr, 0x00000000);	/* int31 highest, base=0x000 */
	mtdcr(uic2sr, 0xffffffff);	/* clear all */
#else 
	/*
	*	DF150835 blb  change to "falling edge" triggered: PR=0=falling  TR=1=edge
	*	UIC0-bit27-int27-IRQ4, 
	*	UIC1-bit18-int50-IRQ7, UIC1-bit19-int51-IRQ8, UIC1-bit28-int60-IRQ0
	*	UIC2-bit00-int64-IRQ5, UIC2-bit01-int65-IRQ6, UIC2-bit04-int68-IRQ4
	*/
	mtdcr(uic0sr, 0xffffffff);	/* clear all */
	mtdcr(uic0er, 0x00000000);	/* disable all */
	mtdcr(uic0cr, 0x00000005);	/* ATI & UIC1 crit are critical */
	mtdcr(uic0pr, 0xfffff7ef);	/* per ref-board manual, (blb  BIT27-IRQ0) */
	mtdcr(uic0tr, 0x00000010);	/* per ref-board manual  (blb  BIT27-IRQ0) */
	mtdcr(uic0vr, 0x00000000);	/* int31 highest, base=0x000 */
	mtdcr(uic0sr, 0xffffffff);	/* clear all */

	mtdcr(uic1sr, 0xffffffff);	/* clear all */
	mtdcr(uic1er, 0x00000000);	/* disable all */
	mtdcr(uic1cr, 0x00000000);	/* all non-critical */
	mtdcr(uic1pr, 0xffffcff7);	/* per ref-board manual(blb 50-IRQ7,51-IRQ8,60-IRQ0)*/
	mtdcr(uic1tr, 0x00003008);	/* per ref-board manual(blb 50-IRQ7,51-IRQ8,60-IRQ0)*/
	mtdcr(uic1vr, 0x00000000);	/* int31 highest, base=0x000 */
	mtdcr(uic1sr, 0xffffffff);	/* clear all */

	mtdcr(uic2sr, 0xffffffff);	/* clear all */
	mtdcr(uic2er, 0x00000000);	/* disable all */
	mtdcr(uic2cr, 0x00000000);	/* all non-critical */
	mtdcr(uic2pr, 0x37ffffff);	/* per ref-board manual(blb 64-IRQ5,65-IRQ6,68-IRQ4)*/
	mtdcr(uic2tr, 0xc8000000);	/* per ref-board manual(blb 64-IRQ5,65-IRQ6,68-IRQ4)*/
	mtdcr(uic2vr, 0x00000000);	/* int31 highest, base=0x000 */
	mtdcr(uic2sr, 0xffffffff);	/* clear all */
#endif


	/* select Ethernet pins */
	mfsdr(SDR0_PFC1, sdr0_pfc1);
	sdr0_pfc1 = (sdr0_pfc1 & ~SDR0_PFC1_SELECT_MASK) | SDR0_PFC1_SELECT_CONFIG_6;
	sdr0_pfc1 |= SDR0_PFC1_SIS_IIC1_SEL;	/*CMC never uses SPI, select I2C*/
	sdr0_pfc1 |= SDR0_PFC1_UES_EBCHR_SEL;	/*CMC disable device controller*/

	sdr0_pfc1 |= SDR0_PFC1_DIS_UICIRQ5_SEL;	/*CMC Enable UIC and NOT DMA*/

/* pw start adds for HW flow control*/
	/* UART setup */
	sdr0_pfc1 &= ~SDR0_PFC1_U1ME_MASK;
	sdr0_pfc1 |= SDR0_PFC1_U1ME_CTS_RTS;
	sdr0_pfc1 &= ~SDR0_PFC1_U0ME_MASK;
	sdr0_pfc1 |= SDR0_PFC1_U0ME_CTS_RTS;
	sdr0_pfc1 &= ~SDR0_PFC1_U0IM_MASK;
	sdr0_pfc1 |= SDR0_PFC1_U0IM_4PINS;
/* pw end adds for HW flow control*/

	mfsdr(SDR0_PFC2, sdr0_pfc2);
	sdr0_pfc2 = (sdr0_pfc2 & ~SDR0_PFC2_SELECT_MASK) | SDR0_PFC2_SELECT_CONFIG_6;
	mtsdr(SDR0_PFC2, sdr0_pfc2);
	mtsdr(SDR0_PFC1, sdr0_pfc1);

	/* PCI arbiter enabled */

 // HACK we don't need this
	mfsdr(sdr_pci0, reg);
	mtsdr(sdr_pci0, 0x80000000 | reg);


	/* setup NAND FLASH */
	mfsdr(SDR0_CUST0, sdr0_cust0);
	sdr0_cust0 = SDR0_CUST0_MUX_NDFC_SEL	|
		SDR0_CUST0_NDFC_ENABLE		|
		SDR0_CUST0_NDFC_BW_8_BIT	|
		SDR0_CUST0_NDFC_ARE_MASK	|
		(0x80000000 >> (28 + CFG_NAND_CS));
	mtsdr(SDR0_CUST0, sdr0_cust0);

	return 0;
}

/*---------------------------------------------------------------------------+
  | misc_init_r.
  +---------------------------------------------------------------------------*/
int misc_init_r(void)
{
	uint pbcr;
	int size_val = 0;
	u32 reg;
#ifdef CONFIG_440EPX
	unsigned long usb2d0cr = 0;
	unsigned long usb2phy0cr, usb2h0cr = 0;
	unsigned long sdr0_pfc1;
	char *act = getenv("usbact");
#endif

	u32 srst0, srst1;

	/*
	 * FLASH stuff...
	 */

	/* Re-do sizing to get full correct info */
#if defined(CONFIG_NAND_U_BOOT) || defined(CONFIG_NAND_SPL)
	mtdcr(ebccfga, pb3cr);
#else
	mtdcr(ebccfga, pb0cr);
#endif
	pbcr = mfdcr(ebccfgd);
	switch (gd->bd->bi_flashsize) {
	case 1 << 20:
		size_val = 0;
		break;
	case 2 << 20:
		size_val = 1;
		break;
	case 4 << 20:
		size_val = 2;
		break;
	case 8 << 20:
		size_val = 3;
		break;
	case 16 << 20:
		size_val = 4;
		break;
	case 32 << 20:
		size_val = 5;
		break;
	case 64 << 20:
		size_val = 6;
		break;
	case 128 << 20:
		size_val = 7;
		break;
	}
	pbcr = (pbcr & 0x0001ffff) | gd->bd->bi_flashstart | (size_val << 17);
#if defined(CONFIG_NAND_U_BOOT) || defined(CONFIG_NAND_SPL)
	mtdcr(ebccfga, pb3cr);
#else
	mtdcr(ebccfga, pb0cr);
#endif
	mtdcr(ebccfgd, pbcr);

	/* adjust flash start and offset */
	gd->bd->bi_flashstart = 0 - gd->bd->bi_flashsize;
	gd->bd->bi_flashoffset = 0;

#ifdef CFG_ENV_IS_IN_FLASH
	/* Monitor protection ON by default */
	(void)flash_protect(FLAG_PROTECT_SET,
			    -CFG_MONITOR_LEN,
			    0xffffffff,
			    &flash_info[0]);

	/* Env protection ON by default */
	(void)flash_protect(FLAG_PROTECT_SET,
			    CFG_ENV_ADDR_REDUND,
			    CFG_ENV_ADDR_REDUND + 2*CFG_ENV_SECT_SIZE - 1,
			    &flash_info[0]);
#endif

	/*
	 * USB suff...
	 */

#ifdef CONFIG_440EPX
	/*CMC is a HOST*/
	if (act == NULL || strcmp(act, "hostdev") == 0)	{		
		/* SDR Setting */

#define SDR0_USB0                    0x0320     /* USB Control Register */

		/*blb  CMC ADD info msg  (sequence the release of the various resets)*/
		mfsdr(SDR0_SRST1, srst1);
		udelay (1000);
		mfsdr(SDR0_SRST0, srst0);
		printf("USB:   INIT: srst0=0x%x  srst1=0x%x\n", srst0, srst1);
		/*END blb  CMC ADD info msg*/

		mfsdr(SDR0_PFC1, sdr0_pfc1);
		mfsdr(SDR0_USB0, usb2d0cr);
		mfsdr(SDR0_USB2PHY0CR, usb2phy0cr);
		mfsdr(SDR0_USB2H0CR, usb2h0cr);

		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_XOCLK_MASK;
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_XOCLK_EXTERNAL;	/*0 CMC add*/
		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_WDINT_MASK;
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_WDINT_16BIT_30MHZ;	/*1*/

		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_DVBUS_MASK;
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_DVBUS_PURDIS;		/*0*/
		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_DWNSTR_MASK;
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_DWNSTR_HOST;		/*1*/
		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_UTMICN_MASK;
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_UTMICN_HOST;		/*1*/


		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_CLKSEL_MASK;		/*CMC add*/
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_CLKSEL_48MHZ;		/*1*/
		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_XOREF_MASK;		/*CMC add*/
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_XOREF_XO;			/*1*/

		/* An 8-bit/60MHz interface is the only possible alternative
		   when connecting the Device to the PHY */
		usb2h0cr   = usb2h0cr &~SDR0_USB2H0CR_WDINT_MASK;
		usb2h0cr   = usb2h0cr | SDR0_USB2H0CR_WDINT_16BIT_30MHZ;	/*1*/

		/* To enable the USB 2.0 Device function through the UTMI interface */
		usb2d0cr = usb2d0cr &~SDR0_USB2D0CR_USB2DEV_EBC_SEL_MASK;
		usb2d0cr = usb2d0cr | SDR0_USB2D0CR_EBC_SELECTION;		/*CMC add 0*/

		sdr0_pfc1 = sdr0_pfc1 &~SDR0_PFC1_UES_MASK;
		//sdr0_pfc1 = sdr0_pfc1 | SDR0_PFC1_UES_USB2D_SEL;		/*0*/
		sdr0_pfc1 = sdr0_pfc1 | SDR0_PFC1_UES_EBCHR_SEL;		/*CMC add 1*/

		mtsdr(SDR0_PFC1, sdr0_pfc1);
		mtsdr(SDR0_USB0, usb2d0cr);
		mtsdr(SDR0_USB2PHY0CR, usb2phy0cr);
#if 0
		/*for debug hard code the values*/
		mtsdr(SDR0_USB2PHY0CR, 0x51780000);		/*hack SDR addr 0x4103 48M*/
#endif
		mtsdr(SDR0_USB2H0CR, usb2h0cr);

		/*blb  CMC ADD  (AMCC errata, sequence the release of the various resets)*/
		/*at this point in time     srst0=0x01000080  srst1=0x673c6000*/
		udelay (1000);
		mtsdr(SDR0_SRST1, 0x603c6000);			/*clr OPB2, OPB2PLB40, PLB42OPB1*/
		udelay (1000);
		mtsdr(SDR0_SRST1, 0x602c6000);			/*clr USB20PHY*/
		udelay (1000);
		mtsdr(SDR0_SRST0, 0x00000080);			/*clr USB2H*/
		udelay (10000);							/*10msec !!!*/
		mtsdr(SDR0_SRST1, 0x60246000);			/*clr USB2HUTMI*/
		udelay (10000);							/*10msec !!!*/
		/*blb END CMC ADD*/

		/*clear resets*/
		udelay (1000);
		mtsdr(SDR0_SRST1, 0x00000000);
		udelay (1000);
		mtsdr(SDR0_SRST0, 0x00000000);

		printf("USB:   Host(int phy) Device(ext phy)\n");

	} else if (strcmp(act, "dev") == 0) {
		/*-------------------PATCH-------------------------------*/
		mfsdr(SDR0_USB2PHY0CR, usb2phy0cr);

		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_XOCLK_MASK;
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_XOCLK_EXTERNAL;	/*0*/
		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_DVBUS_MASK;
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_DVBUS_PURDIS;		/*0*/
		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_DWNSTR_MASK;
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_DWNSTR_HOST;		/*1*/
		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_UTMICN_MASK;
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_UTMICN_HOST;		/*1*/
		mtsdr(SDR0_USB2PHY0CR, usb2phy0cr);

		udelay (1000);
		mtsdr(SDR0_SRST1, 0x672c6000);

		udelay (1000);
		mtsdr(SDR0_SRST0, 0x00000080);

		udelay (1000);
		mtsdr(SDR0_SRST1, 0x60206000);

		*(unsigned int *)(0xe0000350) = 0x00000001;

		udelay (1000);
		mtsdr(SDR0_SRST1, 0x60306000);
		/*-------------------PATCH-------------------------------*/

		/* SDR Setting */
		mfsdr(SDR0_USB2PHY0CR, usb2phy0cr);
		mfsdr(SDR0_USB2H0CR, usb2h0cr);
		mfsdr(SDR0_USB0, usb2d0cr);
		mfsdr(SDR0_PFC1, sdr0_pfc1);

		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_XOCLK_MASK;
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_XOCLK_EXTERNAL;	/*0*/
		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_WDINT_MASK;
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_WDINT_8BIT_60MHZ;	/*0*/
		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_DVBUS_MASK;
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_DVBUS_PUREN;		/*1*/
		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_DWNSTR_MASK;
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_DWNSTR_DEV;		/*0*/
		usb2phy0cr = usb2phy0cr &~SDR0_USB2PHY0CR_UTMICN_MASK;
		usb2phy0cr = usb2phy0cr | SDR0_USB2PHY0CR_UTMICN_DEV;		/*0*/

		usb2h0cr   = usb2h0cr &~SDR0_USB2H0CR_WDINT_MASK;
		usb2h0cr   = usb2h0cr | SDR0_USB2H0CR_WDINT_8BIT_60MHZ;		/*0*/

		usb2d0cr = usb2d0cr &~SDR0_USB2D0CR_USB2DEV_EBC_SEL_MASK;
		usb2d0cr = usb2d0cr | SDR0_USB2D0CR_EBC_SELECTION;		/*0*/

		sdr0_pfc1 = sdr0_pfc1 &~SDR0_PFC1_UES_MASK;
		sdr0_pfc1 = sdr0_pfc1 | SDR0_PFC1_UES_EBCHR_SEL;		/*1*/

		mtsdr(SDR0_USB2H0CR, usb2h0cr);
		mtsdr(SDR0_USB2PHY0CR, usb2phy0cr);
		mtsdr(SDR0_USB0, usb2d0cr);
		mtsdr(SDR0_PFC1, sdr0_pfc1);

		/*clear resets*/
		udelay (1000);
		mtsdr(SDR0_SRST1, 0x00000000);
		udelay (1000);
		mtsdr(SDR0_SRST0, 0x00000000);

		printf("USB:   Device(int phy)\n");
	}
#endif /* CONFIG_440EPX */

	/*
	 * Clear PLB4A0_ACR[WRP]
	 * This fix will make the MAL burst disabling patch for the Linux
	 * EMAC driver obsolete.
	 *
	 * CR128638 This fix allows ECC to work with tftp operations.
	 */
	reg = mfdcr(plb4_acr) & ~PLB4_ACR_WRP;
	mtdcr(plb4_acr, reg);
	return 0;
}

int checkboard(void)
{
	char *s = getenv("serial#");

#ifdef CONFIG_440EPX
	printf("Board: CMC_10G - DELL PPC440EPx Chassis Managment Controller");
#else
	printf("Board: Rainier - AMCC PPC440GRx Evaluation Board");
#endif
	if (s != NULL) {
		puts(", serial# ");
		puts(s);
	}
	putc('\n');

	return (0);
}

#if defined(CFG_DRAM_TEST)
/*****************************************************************************
*	Core memory tests run under On Chip Memory, so the entire array is 
*	available to test.
*	1)  membus()
*	2)  memaddr()
*	3)  The origional "testdram()" is a pattern test of A's and 5's to the 
*		first 1024 bytes, of each Meg boundary.  This test is tied to all 
*		three jumpers  1) vendor mode 2) force recover recover  3) password
*****************************************************************************/
int testdram(void)
{
char msg[80];

	mtmsr(0);

	if( membus(msg) )
		core_death(msg);

	if( memaddr(msg) )
		core_death(msg);


	if( is_vendor_mode() && is_recover_mode() &&  is_hard_password_mode() )
		{
		if( mempat(msg) )
			core_death(msg);
		}

	return 0;
}

/*****************************************************************************
*	FUNCTION:	membus()
*	DATE:		11-6-06
*	Core memory test, walk a data bit at address 0.  After each write, write a 
*	0 to addr+4 to force the bus to zero, then read back and verify. 
*	Console message indicating pass/fail.
*
*	RETURN:	0  - test passed
*			-1 - test failed
*****************************************************************************/
static u32 membus(char *msg)
{
volatile u32 exp = 0x1;
volatile u32 rec;
volatile u32 *addr		= (u32*)CFG_MEMTEST_START_FLASH_EXE;
volatile u32 *nulladdr	= (u32*)CFG_MEMTEST_START_FLASH_EXE + 4;

do
	{
	*addr = exp;
	*nulladdr = 0;	/*drive the bus low*/
	rec = *addr;

	if(exp != rec)
		{
		sprintf(msg, "membus():ERROR: At 0x%x exp=0x%x rec=0x%x\n", addr, exp, rec);
		return(-1);
		}
	exp <<= 1;
	}
while(exp);

printf("SDRAM: memory bus test:PASS\n");
return(0);
}/*end FUNCTION: membus()*/



/*****************************************************************************
*	FUNCTION:	memaddr()
*	DATE:		11-6-06
*	Core memory test, starting at address 0, write a pattern of address=data,
*	shifting the address by a power of two.  Then starting at address 0 read 
*	and verify address=data.   Console message indicating pass/fail.
*
*	RETURN:	0  - test passed
*			-1 - test failed
*****************************************************************************/
static u32 memaddr(char *msg)
{
volatile u32 addr		= CFG_MEMTEST_START_FLASH_EXE;
volatile u32 addr_end	= CFG_MEMTEST_END_FLASH_EXE;
volatile u32 rec;

/*	From 0 write addr = data*/
do
	{
	*(u32*)addr = addr;
	if(addr == 0)
		addr = 4;
	else
		addr <<= 1;
	}
while(addr < addr_end);

/*	From 0 read back and verify, addr = data*/
addr = 0;
do
	{
	rec = *(u32*)addr;
	if(rec != addr)
		{
		sprintf(msg, "memaddr():ERROR: At 0x%x exp=0x%x rec=0x%x\n", addr,addr,rec);
		return(-1);
		}

	if(addr == 0)
		addr = 4;
	else
		addr <<= 1;
	}
while(addr < addr_end);

printf("SDRAM: memory address test:PASS\n");
return(0);
}/*end FUNCTION: memaddr()*/


/*****************************************************************************
*	FUNCTION:	mempat()
*	DATE:		11-18-06
*	Run a simple memory pattern test of A's, 5's, leaving memory cleared. 
*	Console message indicating pass/fail.
*
*	RETURN:	0  - test passed
*			-1 - test failed
*****************************************************************************/
#define MEM_PATTERN_CNT	3

static u32 mempat(char *msg)
{
/*	In the pattern arrray 0 MUST be the last pattern*/
u32 pats[MEM_PATTERN_CNT] = {0xaaaaaaaa, 0x55555555, 0x00000000 };
u32 saddr	= CFG_MEMTEST_START_FLASH_EXE;
u32 eaddr	= CFG_MEMTEST_END_FLASH_EXE;
u32 bcnt	= (CFG_MEMTEST_END_FLASH_EXE - CFG_MEMTEST_START_FLASH_EXE);
u32 pindex	= 0;
u32 *s;
u32 exp, rec, nextpat;

/*	Write first pattern*/
printf("mempat(): fill 0x%x -to- 0x%x with 0x%x\r", 
		CFG_MEMTEST_START_FLASH_EXE, CFG_MEMTEST_END_FLASH_EXE, pats[pindex]);
memset(CFG_MEMTEST_START_FLASH_EXE, pats[pindex], bcnt);


while(pats[pindex] != 0x0)
	{
	s = (u32*)saddr;
	exp = pats[pindex];
	nextpat = pats[pindex + 1];

	printf("mempat(): Read/verify 0x%x, write 0x%x                       \r", 
											pats[pindex], pats[pindex + 1]);
	while( (u32)s < eaddr)
		{	
		rec = *s;
		if(rec != exp)
			{
			sprintf(msg, "mempat():ERROR: At 0x%x exp=0x%x rec=0x%x\n", s, exp, rec);
			return(-1);
			}
		
		*s++ = nextpat;
		}

	++pindex;
	}/*end outer while*/

/*	Verify 0's*/
printf("mempat(): Verify 0's                                             \r");
s	= (u32*)saddr;
exp = 0;
while( (u32)s < eaddr)
	{	
	rec = *s;
	if(rec != exp)
		{
		sprintf(msg, "mempat():ERROR: At 0x%x exp=0x%x rec=0x%x\n", s, exp, rec);
		return(-1);
		}
	++s;
	}

printf("SDRAM: memory pattern test:PASS                                  \n");
return(0);
}/*end FUNCTION: mempat()*/


/*****************************************************************************
*	FUNCTION:	core_death()
*	DATE:		6-15-07
*	
*	If a core memory test or other core function fails were we cannot proceed
*	with CMC boot. Throw a message out and togle the debug LED every X seconds.
*
*	RETURN:	None
*****************************************************************************/
static u32 core_death(char *msg)
{
u32 tmp; 

while(1)
	{
	/*	Toggle LED*/
	tmp = in32(GPIO0_OR);
	if(tmp & GPIO_DEBUG_LED)
		out32(GPIO0_OR, tmp & ~GPIO_DEBUG_LED);
	else
		out32(GPIO0_OR, tmp | GPIO_DEBUG_LED);

	/*	Print msg and delay*/
	printf("CMC core failure, cannot proceed with boot.\n");
	printf("%s\n\n", msg);
	udelay(5000000);
	}


return(0);
}/*end FUNCTION: core_death() */

#endif


/*****************************************************************************
*	FUNCTION:	last_stage_init()
*	DATE:		11-20-06
*	Compiled in for CONFIG_CMC_10G.  Calls the CMC selftest just before main()
*	is run. 
*
*****************************************************************************/

#define MAX_OPTARG_LEN	80

int last_stage_init(void)
{
u16  mode;	
char *con_p;			/*pointers to env strings*/
char *ip_p;		
char *gate_p;
char *mask_p;
char *serv_p;
char *host_p;
char *netdev_p;
char *baud_p;
char *nfsargs_p;
char *rw_attribute_p;
s32  rc;
u32  offset;
u32  bcnt;
u32  exp_crc;
#ifdef DYNAMIC_IMAGE_START	
u32  rootfs_offset;
u32  rootfs_bcnt;
u32  rootfs_exp_crc;
u32  varfs_offset;
u32  varfs_bcnt;
u32  varfs_exp_crc;
u32  varfs_size;
#endif
u32  rootfs_size;
char msg[MAX_MSG_LEN];
char buf[CFG_CBSIZE];	
char vmodes_str[80];		
char *vmode_p;
char *rw_attribute;
char *rw_attribute1;
char *mem_p;
char *pwjumper_p;
char mem_size_str[40];
char *optarg_p;
char optarg_str[MAX_OPTARG_LEN];
u32	 optarg_len;
u32	 ksize	= FLASH_MAX_SIZE_DIAG;
u32	 kstart	= SECTOR_TO_PHYSADDR(FLASH_SECT_START_DIAG);
u32  kend	= SECTOR_TO_PHYSADDR(FLASH_SECT_END_DIAG);



/*	
*	To help in debug, IF hard vendor mode AND hard recover mode then skip 
*	util_cmc_main(). If I put in a bug that crashed u-boot we can 
*	still boot up and reflash u-boot without the jtag.
*/
if( !(is_hard_vendor_mode() && is_hard_recover_mode() && !is_hard_password_mode()) )
	util_cmc_main();

/*	
*	Get kernel parameters needed to build the bootcmd below.  Do it here so 
*	if it fails we can force a recover mode.  If we already have an POST error
*	no need to waste time and messup the origional error code and message.
*/
if(!g_post_rc)
	{
	if( (rc = parse_env_descriptor(ENV_FWUPD_KERNEL, &offset, &bcnt, &exp_crc, 
																	msg)) != 0)
		{
		g_post_rc  = R_RECOVER_BAD_ENV_FWUPD_KERNEL;	
		sprintf(g_post_msg,  "ENV_FWUPD_KERNEL read failure, RC=0x%x.", rc);
		offset	= 0;
		bcnt	= 0;
		exp_crc	= 0;
		}	

#ifdef DYNAMIC_IMAGE_START
	/*
	*	We need env's for rootfs and varfs later, but make sure they exists
	*	now so we can mark recover in one place.  
	*/
	if( (rc = parse_env_descriptor(ENV_FWUPD_ROOTFS, &rootfs_offset, 
									&rootfs_bcnt, &rootfs_exp_crc, msg)) != 0)
		{
		g_post_rc  = R_RECOVER_BAD_ENV_FWUPD_ROOTFS;	
		sprintf(g_post_msg,  "ENV_FWUPD_ROOTFS read failure, RC=0x%x.", rc);
		offset	= 0;
		bcnt	= 0;
		exp_crc	= 0;
		}	

	if( (rc = parse_env_descriptor(ENV_FWUPD_VARFS, &varfs_offset, 
									&varfs_bcnt, &varfs_exp_crc, msg)) != 0)
		{
		g_post_rc  = R_RECOVER_BAD_ENV_FWUPD_VARFS;	
		sprintf(g_post_msg,  "ENV_FWUPD_VARFS read failure, RC=0x%x.", rc);
		offset	= 0;
		bcnt	= 0;
		exp_crc	= 0;
		}	
#endif /*DYNAMIC_IMAGE_START*/
	}

	
/*	
*	Set self test failure or recover condition.  We cannot do this in 
*	util_cmc_main() we might enable console prematurly.  For a recover 
*	codition enable the console in main_loop() in main.c.
*/
if( g_post_rc || is_recover_mode() )
	{
	if( ABS_read_fpga(FPGA_REG_ID) == FPGA_ID)
		{
		mode = ABS_read_fpga(FPGA_REG_MODE0);
		if(g_post_rc & LOG_ERROR)
			mode |= MODE_SELF_TEST_ERROR;

		if(g_post_rc & LOG_RECOVER)
			mode |= MODE_IMAGE_RECOVER_CONDITION;

		if( is_recover_mode() )
			{
			mode |= MODE_RECOVER_HARD;
			g_post_rc = R_RECOVER_FORCED;
			strcpy(g_post_msg,  "Hard/Soft forced recover");
			}

		ABS_write_fpga(FPGA_REG_MODE0, mode);
		}
	}

/*	Set amber error LED on CMC front panel upon any error*/
if(g_post_rc)
	{
	clearlow_gpio_or(GPIO_BANK0, GPIO_LED_BLUE);
	sethigh_gpio_or(GPIO_BANK0, GPIO_LED_AMBER);
	}


/*
*	At this point we are finished with all the selftest and are ready to go 
*	into main_loop().  Set up any env variables that contain compile or 
*	run time dependent arguments.  We do not want to hard code then all in 
*	cmc_10g.h under the CONFIG_EXTRA_ENV_SETTINGS area, as then are not 
*	dynamic.  
*	
*	At the same time we MUST always be able to boot.  Some env variables will
*	be constructed at run time thus you will NEVER be able to change them
*	to be persistant accross reset.  As always, in vendor mode...your choice.
*/
/*	
*	However, even though we are building these args im memory u-boot seems 
*	to take the the env flash area before the memory area.  So set them to 
*	NULL to force the memory env that we are about to create.  
*	Do NOT do a saveenv.
*/
setenv("bootargs", 0);
setenv("bootcmd",  0);
setenv("kozio",	   0);

/****************************************************************************
*	Set up Kozio diags
*****************************************************************************/
sprintf(buf, "bootm %x", SECTOR_TO_PHYSADDR(FLASH_SECT_START_DIAG) );
setenv("kozio", buf);

/*	
*	Removing Kozio from the firmimg.cmc file for A rev.  Provide a env cmd so
*	the lab can "run kozioinstall" and not have to remember the details.
*/
sprintf(buf,	"mw.b 200000 ff %x;  tftp 200000 %s;  protect off all;  "
				"erase %x %x;  cp.b 200000 %x %x",
				ksize, IMAGE_FILENAME_DIAG, kstart, (kend - 1), kstart, ksize);
setenv("kozioinstall", buf);
	
/****************************************************************************
*	Set up bootargs and bootcmd.
*	NOTE: bootcmd is defined in cmc_10g.h  CONFIG_BOOTCOMMAND
*****************************************************************************/
/*	
*	Take existing args to build a bootargs line.  If it does not exists
*	(should not happen) use the default.
*/
if( (serv_p = getenv("serverip")) == 0)
	serv_p = DEFAULT_SERVERIP;	

if( (ip_p = getenv("ipaddr")) == 0)
	ip_p = DEFAULT_IPADDR;	

if( (gate_p = getenv("gatewayip")) == 0)
	gate_p = DEFAULT_GATEWAYIP;	

if( (mask_p = getenv("netmask")) == 0)
	mask_p = DEFAULT_NETMASK;	

if( (host_p = getenv("hostname")) == 0)
	host_p = DEFAULT_HOSTNAME;	

if( (netdev_p = getenv("netdev")) == 0)
	netdev_p = DEFAULT_NETDEV;	

if( (con_p = getenv("console")) == 0)
		con_p = DEFAULT_CONSOLE;

if( (baud_p = getenv("baudrate")) == 0)
		baud_p = DEFAULT_BAUDRATE;


#ifdef DYNAMIC_IMAGE_START	/*ROOTFS/VARFS error checking was done above*/
	parse_env_descriptor(ENV_FWUPD_ROOTFS, &rootfs_offset, 
									&rootfs_bcnt, &rootfs_exp_crc, msg);	
	rootfs_size = rootfs_bcnt;

	/*rootfs and var, seperated by a pad at fwupd, IF vmode allow play area*/
	if( is_vendor_mode() )	
		rootfs_size += ROOTFS_VMODE_PAD;


	parse_env_descriptor(ENV_FWUPD_VARFS, &varfs_offset, 
									&varfs_bcnt, &varfs_exp_crc, msg);
	varfs_size = varfs_bcnt;
#else
		rootfs_size = FLASH_MAX_SIZE_ROOT;
#endif /*DYNAMIC_IMAGE_START*/


/*	
*	CR154059, We need to create a Linux boot argument as a way to inform the 
*	OS which 0 mode jumpers are installed.  We do NOT want Linux polling 
*	the various jumpers.  Leave that to u-boot.  
*/
vmodes_str[0] = '\0';
strcat(vmodes_str, "vendor_mode=");	

if( is_hard_vendor_mode() )
	{
	vmode_p = "header:";			/*hard vendor mode via 2 pin header*/
	strcat(vmodes_str, vmode_p);
	}

if( is_soft_vendor_mode() )
	{
	vmode_p = "soft:";				/*soft vendor mode via MODE reg*/
	strcat(vmodes_str, vmode_p);
	}

if( is_fpga_vendor_mode() )
	{
	vmode_p = "fpga:";				/*fpga vendor mode via NEB/carrier*/
	strcat(vmodes_str, vmode_p);
	}
	
if(vmodes_str[strlen(vmodes_str)-1] == 0x3a)
	vmodes_str[strlen(vmodes_str)-1] = '\0';	/*NULL the ":" end character*/

/*3-24-09 if no vendor mode Phil does NOT want the text=*/
if(!strcmp(vmodes_str, "vendor_mode="))	
	vmodes_str[0] = '\0';

/*	Phil needed a flag for the password jumper*/
if ( is_hard_password_mode() )
	pwjumper_p = " password_mode ";
else
	pwjumper_p = " ";


/*	
*	Add the ability to include kernel boot arguments on the fly via an env 
*	variable. Be careful as it is NOT qualified with vendor mode.  The feature
*	is for NFS and normal boot.
*/
if((optarg_p = getenv("optarg")) != 0)
	{
	sprintf(optarg_str, " %s", optarg_p);	

	/*	Strip off any leading or trailing " or ' characters*/
	optarg_len = strlen(optarg_str);
	/*NOTE: offset is not 0 but 1, (leading space in sprintf)*/
	if( (optarg_str[1] == 0x22) || (optarg_str[1] == 0x27) )
		optarg_str[1] = 0x20;
	if( (optarg_str[optarg_len-1] == 0x22) || (optarg_str[optarg_len-1] == 0x27) )
		optarg_str[optarg_len-1] = 0x20;
	}
else
	strcpy(optarg_str, " ");	

/*	
*	Add the memory size argument to tell the kernel how much memory it has.
*	In board.c we are reserving CONFIG_PRAM bytes at the top of memory for 
*	u-boot->kernel messages.
*	blb mod 10-27-08, move mem_size_str calc BEFORE nfsargs creation
*/
if((mem_p = getenv("mem")) != 0)
	sprintf(mem_size_str, " mem=%s ", mem_p);		
else
	strcpy(mem_size_str, " ");	


/*	
*	Now we have to patch up "nfsargs".  Copy in env string.  If "vendor_mode" 
*	text exists, this function put it there, then someone did a saveenv under
*	u-boot.  Do not keep appending another copy on the end.  IF SOMEONE DOES A 
*	saveenv ON NFSARGS, THEN ADDS ANOTHER ARG AFTER vendor_mode= IT WILL NOT 
*	BE SEEN.
*/
if( (nfsargs_p = getenv("nfsargs")) != 0)
	{
	strcpy(buf, nfsargs_p);						
	if( (vmode_p = strstr(buf, "vendor_mode")) != 0 )
		*(--vmode_p) = '\0';

	//strcat(buf, " vendor_mode=");	/*shows up twice in nfsargs*/
	/*do not let vendor_mode= string bump up to ROOTPATH, add a space*/
	strcat(buf, " ");		
	strcat(buf, vmodes_str);

	strcat(buf, pwjumper_p);
	
	strcat(buf, optarg_str);

	setenv("nfsargs", 0);
	setenv("nfsargs", buf);
	}

/*	Determine if we are read only or read/write*/
if( is_vendor_mode() )
	{
	rw_attribute = "rw";	
	rw_attribute1 = "";		/*PORT 2.6.24 must have empty string*/
	/*	For kernel debug they want to be able to be "ro" in vendor mode*/
	if((rw_attribute_p = getenv("rw_attribute")) != 0)
		{
		if(!strcmp(rw_attribute_p, "ro"))
			{
			rw_attribute  = "ro";
			rw_attribute1 = "ro";
			printf("\nEnv \'rw_attribute\' overrides to \'ro\'\n");
			}
		}
	}
else
	{
	rw_attribute  = "ro";
	rw_attribute1 = "ro";
	}


#if 0 /*DF144098*/
sprintf(buf,	"root=/dev/mtdblock1 %s rootfstype=jffs2 "
				"ip=%s:%s:%s:%s:%s:%s:off panic=1 %s "
				"console=%s,%s %s %s", rw_attribute, ip_p, serv_p, gate_p, mask_p, host_p,
				netdev_p, vmodes_str, con_p, baud_p, mem_size_str, optarg_str);
#else

#ifdef DYNAMIC_IMAGE_START	
sprintf(buf,	"root=mtd:rootfs %s rootfstype=jffs2 "
				" panic=1 %s %s console=%s,%s %s %s "
				"mtdparts=ndfc:0x%x@0x%x(rootfs)%s,0x%x@0x%x(var),"
                "0x3a00000@0x4600000(fw_image) ", 
				rw_attribute, vmodes_str, pwjumper_p, con_p, baud_p, mem_size_str, optarg_str,
				rootfs_size, rootfs_offset, rw_attribute1, varfs_size, varfs_offset);
#else	
sprintf(buf,	"root=mtd:rootfs  %s rootfstype=jffs2 "
				" panic=1 %s %s console=%s,%s %s %s "
				"mtdparts=ndfc:0x%x@0x%x(rootfs)%s,0xc00000@0x3a00000(var),"
                "0x3a00000@0x4600000(fw_image) ", 
				rw_attribute, vmodes_str, pwjumper_p, con_p, baud_p, mem_size_str, optarg_str,
				FLASH_MAX_SIZE_ROOT, FLASH_OFF_START_ROOT, rw_attribute1);
#endif
#endif

setenv("bootargs", buf);

/*
*	The "bootcmd" was 
*		#define CONFIG_BOOTCOMMAND	"nboot 200000 uImage; bootm"
*		setenv("bootcmd", CONFIG_BOOTCOMMAND);
*	However the "nboot" cmd does not appear to be "nand aware" as a bad block 
*	in the kernel image causes us to fail.  We simply need to do a jffs2 read.
*	We have all the data in a env var so parse it out and build a cmd line to 
*	boot.   NOTE: The kernel image has already been verified correct by 
*	util_cmc_main().  If the descriptor fails....what to do??
*/
sprintf(buf, "nand read.jffs2 0x%x 0x%x 0x%x; bootm", CFG_LOAD_ADDR, offset, bcnt);
setenv("bootcmd", buf);

/*	NON-vendor mode boot delay is set in main_loop()*/


return(0);
}


