
#ifndef CPLD_IN_UBOOT
#define GPIO0_ODR		0xEF600B18	//Open Drain Register, TMS
#define GPIO1_ODR		0xEF600C18	//Open Drain Register, TCK, TDI, TDO
#define GPIO0_TCR		0xEF600B04	//Three-State Control Register, TMS
#define GPIO1_TCR		0xEF600C04	//Three-State Control Register, TCK, TDI, TDO
#define GPIO0_OR		0xEF600B00  //Output Register, TMS
#define GPIO1_OR		0xEF600C00	//Output Register, TCK, TDI
#define GPIO1_OSRH		0xEF600C0C	//Output Select Register High
#define GPIO1_TSRH		0xEF600C14	//Three-State Select Register High
#define GPIO1_IR		0xEF600C1C	//Input Register, TDO
//TMS is bit 8 in GPIO0
//TCK is bit 20 in GPIO1
//TDI is bit 21 in GPIO1
//TDO is bit 22 in GPIO1
#define GPI_PSU_MSK		0xA0000442	//Power supply present interrupt mask
#define GPI_IOM_MSK		0xA0000462	//IOM present interrupt mask
#define GPI_CP_MSK		0xA00004B2	//Control Panel present interrupt mask


#include <util_defs.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <net/if.h>
#include <linux/ioctl.h>
#include <linux/i2c-dev.h>
#include <linux/sockios.h>
#include <errno.h>
#include <fcntl.h>
#include <gpio_struct.h>
#include <gpioiocts.h>

#else
#include <common.h>		/*for u-boot*/
//#include <util_defs.h>
#include <util_fpga.h>

/*	Map to util_fpga.h*/
#define GPI_PSU_MSK		FPGA_REG_GPI_PSU_MASK
#define GPI_IOM_MSK		FPGA_REG_GPI_IOM_MASK
#define GPI_CP_MSK		FPGA_REG_GPI_CP_MASK
#endif


int conio_setup_once = 1;

int conio_setup(void);
int conio_outp(unsigned short g_siIspPins);
int conio_inp(void);


#ifndef CPLD_IN_UBOOT
//u32 read_phys32(u32 address);
//void write_phys32(u32 address, u32 val);

u32 read_phys32(u32 address)
{
    int fd,status;
    GpioCmdArg GpioPacket;
    u32 result;

    fd = open("/dev/GpioDev",O_RDWR);
    if(fd == -1) {
        printf("Unable to open /dev/GpioDev\n");
        exit(-1);
    }

    GpioPacket.CmdBuf.ReqType = IOCTL_GPIO_READ_CMD_32;
    GpioPacket.CmdBuf.BaseAddress = address;
    GpioPacket.CmdBuf.Offset = 0;

    status = ioctl(fd, IOCTL_GPIO_READ_CMD_32, &GpioPacket);
    if(status == -1){
        printf("Memory Access Failure - Results will be invalid.\n");
        //print_errno(errno);
        exit(-1);
    }

    result = (u32) (GpioPacket.RespBuf.Result & 0xFFFFFFFF);
    close(fd);
    return result;
}

void write_phys32(u32 address, u32 val)
{
    int fd,status;
    GpioCmdArg GpioPacket;
    u16 result;

    fd = open("/dev/GpioDev",O_RDWR);
    if(fd == -1) {
        printf("Unable to open /dev/GpioDev\n");
        return;
    }

    GpioPacket.CmdBuf.ReqType = IOCTL_GPIO_WRITE_CMD_32;
    GpioPacket.CmdBuf.BaseAddress = address;
    GpioPacket.CmdBuf.Offset = 0;
    GpioPacket.CmdBuf.Value = val;

    status = ioctl(fd, IOCTL_GPIO_WRITE_CMD_32, &GpioPacket);
    if(status == -1){
        printf("Memory Access Failure - Results will be invalid.\n");
        //print_errno(errno);
        return;
    }

    result = (u16) (GpioPacket.RespBuf.Result & 0xFFFF);
    close(fd);
    return;
}

#else
/*****************************************************************************
*	FUNCTION: read_phys32(), write_phys32()
*	DATE:	9-18-07
*	BUILD:	u-boot
*
*	Abstraction API.  when porting the CPLD prgramming code from Linux they 
*	use the HAPI calls.  Add a look alike for u-boot.   They did not implimit
*	an error return.
*****************************************************************************/
u32 read_phys32(u32 address)
{
return( *(u32*)address );
}/*end FUNCTION: read_phys32()*/


u32 write_phys32(u32 address, u32 val)
{
*(u32*)address = val;
return(0);
}/*end FUNCTION: write_phys32()*/
#endif 



int conio_setup(void)
{
	//TMS is bit 8 in GPIO0
	//TCK is bit 20 in GPIO1
	//TDI is bit 21 in GPIO1
	//TDO is bit 22 in GPIO1

	u32 read_phys32_return = 0;
	conio_setup_once = 0;

	//There is a serial bus that runs between the CPLD and the FPGA.  During the programming
	//process, the CPLD will stop responding to the FPGA which causes all sorts of errors to be reported.
	//To handle this, the Power Supply present, IOMs present, and I/O panel presents are disabled so we wont
	//get interrupts during programming.

	//mask the power supply present interrupts

	read_phys32_return = read_phys32(GPI_PSU_MSK); //Power supply present interrupt mask
	read_phys32_return = read_phys32_return | 0xFFFF0000; //set the 16 LSB bits, LSB on left
	write_phys32(GPI_PSU_MSK, read_phys32_return);

	//mask the IOM present interrupts

	read_phys32_return = read_phys32(GPI_IOM_MSK); //IOM present interrupt mask
	read_phys32_return = read_phys32_return | 0xFFFF0000; //set the 16 LSB bits, LSB on left
	write_phys32(GPI_IOM_MSK, read_phys32_return);

	//mask the control panel present interrupts

	read_phys32_return = read_phys32(GPI_CP_MSK); //Control Panel present interrupt mask
	read_phys32_return = read_phys32_return | 0xFFFF0000; //set the 16 LSB bits, LSB on left
	write_phys32(GPI_CP_MSK, read_phys32_return);



	//Open Drain Register bit 20 - 22 needs to be 0, TCK, TDI, and TDO
	
	read_phys32_return = read_phys32(GPIO1_ODR); //Open Drain Register
	read_phys32_return = read_phys32_return & 0xFFFFF1FF; //make sure bit 20 - 22 is off, LSB on left
	write_phys32(GPIO1_ODR, read_phys32_return);

	//Open Drain Register bit 8 needs to be 0, TMS

	read_phys32_return = read_phys32(GPIO0_ODR); //Open Drain Register
	read_phys32_return = read_phys32_return & 0xFF7FFFFF; //make sure bit 8 is off, LSB on left
	write_phys32(GPIO0_ODR, read_phys32_return);

	//Output Register, TCK and TDI low.  Did this before enabling the output to make sure they are in the default state.
	
	read_phys32_return = read_phys32(GPIO1_OR); //Output Register
	read_phys32_return = read_phys32_return & 0xFFFFF3FF; //make sure bit 20 and 21 is off, LSB on left
	write_phys32(GPIO1_OR, read_phys32_return);

	//Output Register, TMS low.  Did this before enabling the output to make sure they are in the default state.
	
	read_phys32_return = read_phys32(GPIO0_OR); //Output Register
	read_phys32_return = read_phys32_return & 0xFF7FFFFF; //make sure bit 8 is off, LSB on left
	write_phys32(GPIO0_OR, read_phys32_return);

	//Set the Three-State Control Register, TCK and TDI output, TDO input

	read_phys32_return = read_phys32(GPIO1_TCR); //Three-State Control Register
	read_phys32_return = read_phys32_return | 0x00000C00; //make sure bit 20 and 21 on, LSB on left
	read_phys32_return = read_phys32_return & 0xFFFFFDFF; //make sure bit 22 is off, LSB on left
	write_phys32(GPIO1_TCR, read_phys32_return);

	//Set the Three-State Control Register, TMS output

	read_phys32_return = read_phys32(GPIO0_TCR); //Three-State Control Register
	read_phys32_return = read_phys32_return | 0x00800000; //make sure bit 8 is on, LSB on left
	write_phys32(GPIO0_TCR, read_phys32_return);

	return 0;
}

int conio_outp(unsigned short g_siIspPins)
{
	//below is what hardware.c is sending me
	//const unsigned char g_ucPinTDI		= 0x01;    /* Bit address of TDI signal */
	//const unsigned char g_ucPinTCK		= 0x02;    /* Bit address of TCK signal */
	//const unsigned char g_ucPinTMS		= 0x04;    /* Bit address of TMS signal */
	//const unsigned char g_ucPinENABLE		= 0x08;    /* Bit address of chip enable */
	//const unsigned char g_ucPinTRST		= 0x10;    /* Bit address of TRST signal */
	//const unsigned char g_ucPinCE			= 0x20;    /* Bit address of buffer chip enable */
	//const unsigned char g_ucPinTDO		= 0x40;    /* Bit address of TDO signal */

	u32 read_phys32_return;

	if(conio_setup_once)
		conio_setup();

	//This variable is used to store the state of the IO pins.  This way we are not toggling pins that dont need to be toggled and hopefully save some time.
	static unsigned short IO_pin_state = 0;
	unsigned short TDI = g_siIspPins & 0x0001;
	unsigned short TMS = g_siIspPins & 0x0004;
	unsigned short TCK = g_siIspPins & 0x0002;
	unsigned short TDI_pin_state = IO_pin_state & 0x0001;
	unsigned short TMS_pin_state = IO_pin_state & 0x0004;
	unsigned short TCK_pin_state = IO_pin_state & 0x0002;

	//I know I can switch the TDI and the TCK pins at the same time and save an IO transaction,
	//  but I broke it out because I didnt know if switching the data at the same time as the clock would cause problems.

	if (TDI != TDI_pin_state)
		{
		if (TDI)
			{
			//Output Register, TDI high
			read_phys32_return = read_phys32(GPIO1_OR); //Output Register
			read_phys32_return = read_phys32_return | 0x00000400; //make sure bit 21 is on, LSB on left
			write_phys32(GPIO1_OR, read_phys32_return);
			}
		else
			{
			//Output Register, TDI low
			read_phys32_return = read_phys32(GPIO1_OR); //Output Register
			read_phys32_return = read_phys32_return & 0xFFFFFBFF; //make sure bit 21 is off, LSB on left
			write_phys32(GPIO1_OR, read_phys32_return);
			}
		}

	if (TMS != TMS_pin_state)
		{
		if (TMS)
			{
			//Output Register, TMS high
			read_phys32_return = read_phys32(GPIO0_OR); //Output Register
			read_phys32_return = read_phys32_return | 0x00800000; //make sure bit 8 is on, LSB on left
			write_phys32(GPIO0_OR, read_phys32_return);
			}
		else
			{
			//Output Register, TMS low
			read_phys32_return = read_phys32(GPIO0_OR); //Output Register
			read_phys32_return = read_phys32_return & 0xFF7FFFFF; //make sure bit 8 is off, LSB on left
			write_phys32(GPIO0_OR, read_phys32_return);
			}
		}

	if (TCK != TCK_pin_state)
		{
		if (TCK)
			{
			//Output Register, TCK high
			read_phys32_return = read_phys32(GPIO1_OR); //Output Register
			read_phys32_return = read_phys32_return | 0x00000800; //make sure bit 20 is on, LSB on left
			write_phys32(GPIO1_OR, read_phys32_return);
			}
		else
			{
			//Output Register, TCK low
			read_phys32_return = read_phys32(GPIO1_OR); //Output Register
			read_phys32_return = read_phys32_return & 0xFFFFF7FF; //make sure bit 20 is off, LSB on left
			write_phys32(GPIO1_OR, read_phys32_return);
			}
		}
	IO_pin_state = g_siIspPins;
	return 0;
}

int conio_inp(void)
{
	u32 read_phys32_return = 0;

	if(conio_setup_once)
		conio_setup();

	read_phys32_return = read_phys32(GPIO1_IR);
	read_phys32_return = read_phys32_return & 0x00000200; //turn off all bits except 22
	if (read_phys32_return)
		return 0x01;
	else
		return 0x00;
}

//int main(void)
//{
//	conio_outp(0x01); //TDI high
//	printf ("TDO1 is %.2X\n",conio_inp());
//	conio_outp(0x05); //TDI and TMS high
//	printf ("TDO2 is %.2X\n",conio_inp());
//	conio_outp(0x07); //TDI, TMS, and TCK high
//	printf ("TDO3 is %.2X\n",conio_inp());
//	conio_outp(0x06); //TMS and TCK high, TDI low
//	printf ("TDO4 is %.2X\n",conio_inp());
//	conio_outp(0x02); //TCK high, TDI and TMS low
//	printf ("TDO5 is %.2X\n",conio_inp());
//	conio_outp(0x00); //all low
//	printf ("TDO6 is %.2X\n",conio_inp());
//
//	return 0;
//}
