/*
 * (C) Copyright 2006
 * Stefan Roese, DENX Software Engineering, sr@denx.de.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <asm/processor.h>
#include <asm/io.h>
#include <ppc440.h>

#include <sdram.h>

static void ecc_enable(void);

/*-----------------------------------------------------------------------------+
 * wait_for_dlllock.
 +----------------------------------------------------------------------------*/
static int wait_for_dlllock(void)
{
	unsigned long val;
	int wait = 0;

	/* -----------------------------------------------------------+
	 * Wait for the DCC master delay line to finish calibration
	 * ----------------------------------------------------------*/
	mtdcr(ddrcfga, DDR0_17);
	val = DDR0_17_DLLLOCKREG_UNLOCKED;

	while (wait != 0xffff) {
		val = mfdcr(ddrcfgd);
		if ((val & DDR0_17_DLLLOCKREG_MASK) == DDR0_17_DLLLOCKREG_LOCKED)
			/* dlllockreg bit on */
			return 0;
		else
			wait++;
	}
	debug("0x%04x: DDR0_17 Value (dlllockreg bit): 0x%08x\n", wait, val);
	debug("Waiting for dlllockreg bit to raise\n");

	return -1;
}

/*-----------------------------------------------------------------------------+
 * wait_for_dram_init_complete.
 +----------------------------------------------------------------------------*/
int wait_for_dram_init_complete(void)
{
	unsigned long val;
	int wait = 0;

	/* --------------------------------------------------------------+
	 * Wait for 'DRAM initialization complete' bit in status register
	 * -------------------------------------------------------------*/
	mtdcr(ddrcfga, DDR0_00);

	while (wait != 0xffff) {
		val = mfdcr(ddrcfgd);
		if ((val & DDR0_00_INT_STATUS_BIT6) == DDR0_00_INT_STATUS_BIT6)
			/* 'DRAM initialization complete' bit */
			return 0;
		else
			wait++;
	}

	debug("DRAM initialization complete bit in status register did not rise\n");

	return -1;
}

#define NUM_TRIES 64
#define NUM_READS 10

/*-----------------------------------------------------------------------------+
 * denali_core_search_data_eye.
 +----------------------------------------------------------------------------*/
void denali_core_search_data_eye(unsigned long memory_size)
{
	int k, j;
	u32 val;
	u32 wr_dqs_shift, dqs_out_shift, dll_dqs_delay_X;
	u32 max_passing_cases = 0, wr_dqs_shift_with_max_passing_cases = 0;
	u32 passing_cases = 0, dll_dqs_delay_X_sw_val = 0;
	u32 dll_dqs_delay_X_start_window = 0, dll_dqs_delay_X_end_window = 0;
	volatile u32 *ram_pointer;
	u32 test[NUM_TRIES] = {
		0x00000000, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF,
		0x00000000, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF,
		0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0x00000000,
		0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0x00000000,
		0xAAAAAAAA, 0xAAAAAAAA, 0x55555555, 0x55555555,
		0xAAAAAAAA, 0xAAAAAAAA, 0x55555555, 0x55555555,
		0x55555555, 0x55555555, 0xAAAAAAAA, 0xAAAAAAAA,
		0x55555555, 0x55555555, 0xAAAAAAAA, 0xAAAAAAAA,
		0xA5A5A5A5, 0xA5A5A5A5, 0x5A5A5A5A, 0x5A5A5A5A,
		0xA5A5A5A5, 0xA5A5A5A5, 0x5A5A5A5A, 0x5A5A5A5A,
		0x5A5A5A5A, 0x5A5A5A5A, 0xA5A5A5A5, 0xA5A5A5A5,
		0x5A5A5A5A, 0x5A5A5A5A, 0xA5A5A5A5, 0xA5A5A5A5,
		0xAA55AA55, 0xAA55AA55, 0x55AA55AA, 0x55AA55AA,
		0xAA55AA55, 0xAA55AA55, 0x55AA55AA, 0x55AA55AA,
		0x55AA55AA, 0x55AA55AA, 0xAA55AA55, 0xAA55AA55,
		0x55AA55AA, 0x55AA55AA, 0xAA55AA55, 0xAA55AA55 };

	ram_pointer = (volatile u32 *)(CFG_SDRAM_BASE);

	for (wr_dqs_shift = 64; wr_dqs_shift < 96; wr_dqs_shift++) {
		/*for (wr_dqs_shift=1; wr_dqs_shift<96; wr_dqs_shift++) {*/

		/* -----------------------------------------------------------+
		 * De-assert 'start' parameter.
		 * ----------------------------------------------------------*/
		mtdcr(ddrcfga, DDR0_02);
		val = (mfdcr(ddrcfgd) & ~DDR0_02_START_MASK) | DDR0_02_START_OFF;
		mtdcr(ddrcfgd, val);

		/* -----------------------------------------------------------+
		 * Set 'wr_dqs_shift'
		 * ----------------------------------------------------------*/
		mtdcr(ddrcfga, DDR0_09);
		val = (mfdcr(ddrcfgd) & ~DDR0_09_WR_DQS_SHIFT_MASK)
			| DDR0_09_WR_DQS_SHIFT_ENCODE(wr_dqs_shift);
		mtdcr(ddrcfgd, val);

		/* -----------------------------------------------------------+
		 * Set 'dqs_out_shift' = wr_dqs_shift + 32
		 * ----------------------------------------------------------*/
		dqs_out_shift = wr_dqs_shift + 32;
		mtdcr(ddrcfga, DDR0_22);
		val = (mfdcr(ddrcfgd) & ~DDR0_22_DQS_OUT_SHIFT_MASK)
			| DDR0_22_DQS_OUT_SHIFT_ENCODE(dqs_out_shift);
		mtdcr(ddrcfgd, val);

		passing_cases = 0;

		for (dll_dqs_delay_X = 1; dll_dqs_delay_X < 64; dll_dqs_delay_X++) {
			/*for (dll_dqs_delay_X=1; dll_dqs_delay_X<128; dll_dqs_delay_X++) {*/
			/* -----------------------------------------------------------+
			 * Set 'dll_dqs_delay_X'.
			 * ----------------------------------------------------------*/
			/* dll_dqs_delay_0 */
			mtdcr(ddrcfga, DDR0_17);
			val = (mfdcr(ddrcfgd) & ~DDR0_17_DLL_DQS_DELAY_0_MASK)
				| DDR0_17_DLL_DQS_DELAY_0_ENCODE(dll_dqs_delay_X);
			mtdcr(ddrcfgd, val);
			/* dll_dqs_delay_1 to dll_dqs_delay_4 */
			mtdcr(ddrcfga, DDR0_18);
			val = (mfdcr(ddrcfgd) & ~DDR0_18_DLL_DQS_DELAY_X_MASK)
				| DDR0_18_DLL_DQS_DELAY_4_ENCODE(dll_dqs_delay_X)
				| DDR0_18_DLL_DQS_DELAY_3_ENCODE(dll_dqs_delay_X)
				| DDR0_18_DLL_DQS_DELAY_2_ENCODE(dll_dqs_delay_X)
				| DDR0_18_DLL_DQS_DELAY_1_ENCODE(dll_dqs_delay_X);
			mtdcr(ddrcfgd, val);
			/* dll_dqs_delay_5 to dll_dqs_delay_8 */
			mtdcr(ddrcfga, DDR0_19);
			val = (mfdcr(ddrcfgd) & ~DDR0_19_DLL_DQS_DELAY_X_MASK)
				| DDR0_19_DLL_DQS_DELAY_8_ENCODE(dll_dqs_delay_X)
				| DDR0_19_DLL_DQS_DELAY_7_ENCODE(dll_dqs_delay_X)
				| DDR0_19_DLL_DQS_DELAY_6_ENCODE(dll_dqs_delay_X)
				| DDR0_19_DLL_DQS_DELAY_5_ENCODE(dll_dqs_delay_X);
			mtdcr(ddrcfgd, val);

			ppcMsync();
			ppcMbar();

			/* -----------------------------------------------------------+
			 * Assert 'start' parameter.
			 * ----------------------------------------------------------*/
			mtdcr(ddrcfga, DDR0_02);
			val = (mfdcr(ddrcfgd) & ~DDR0_02_START_MASK) | DDR0_02_START_ON;
			mtdcr(ddrcfgd, val);

			ppcMsync();
			ppcMbar();

			/* -----------------------------------------------------------+
			 * Wait for the DCC master delay line to finish calibration
			 * ----------------------------------------------------------*/
			if (wait_for_dlllock() != 0) {
				printf("dlllock did not occur !!!\n");
				printf("denali_core_search_data_eye!!!\n");
				printf("wr_dqs_shift = %d - dll_dqs_delay_X = %d\n",
				       wr_dqs_shift, dll_dqs_delay_X);
				hang();
			}
			ppcMsync();
			ppcMbar();

			if (wait_for_dram_init_complete() != 0) {
				printf("dram init complete did not occur !!!\n");
				printf("denali_core_search_data_eye!!!\n");
				printf("wr_dqs_shift = %d - dll_dqs_delay_X = %d\n",
				       wr_dqs_shift, dll_dqs_delay_X);
				hang();
			}
			udelay(100);  /* wait 100us to ensure init is really completed !!! */

			/* write values */
			for (j=0; j<NUM_TRIES; j++) {
				ram_pointer[j] = test[j];

				/* clear any cache at ram location */
				__asm__("dcbf 0,%0": :"r" (&ram_pointer[j]));
			}

			/* read values back */
			for (j=0; j<NUM_TRIES; j++) {
				for (k=0; k<NUM_READS; k++) {
					/* clear any cache at ram location */
					__asm__("dcbf 0,%0": :"r" (&ram_pointer[j]));

					if (ram_pointer[j] != test[j])
						break;
				}

				/* read error */
				if (k != NUM_READS)
					break;
			}

			/* See if the dll_dqs_delay_X value passed.*/
			if (j < NUM_TRIES) {
				/* Failed */
				passing_cases = 0;
				debug("Failure - wr_dqs_shift %d\n",wr_dqs_shift);
				/* break; */
			} else {
				/* Passed */
				if (passing_cases == 0)
					dll_dqs_delay_X_sw_val = dll_dqs_delay_X;
				passing_cases++;
                debug("Pass - wr_dqs_shift = %d, pass count = %d\n",wr_dqs_shift, passing_cases);
				if (passing_cases >= max_passing_cases) {
					max_passing_cases = passing_cases;
					wr_dqs_shift_with_max_passing_cases = wr_dqs_shift;
					dll_dqs_delay_X_start_window = dll_dqs_delay_X_sw_val;
					dll_dqs_delay_X_end_window = dll_dqs_delay_X;
				}
			}

			/* -----------------------------------------------------------+
			 * De-assert 'start' parameter.
			 * ----------------------------------------------------------*/
			mtdcr(ddrcfga, DDR0_02);
			val = (mfdcr(ddrcfgd) & ~DDR0_02_START_MASK) | DDR0_02_START_OFF;
			mtdcr(ddrcfgd, val);

		} /* for (dll_dqs_delay_X=0; dll_dqs_delay_X<128; dll_dqs_delay_X++) */

	} /* for (wr_dqs_shift=0; wr_dqs_shift<96; wr_dqs_shift++) */

	/* -----------------------------------------------------------+
	 * Largest passing window is now detected.
	 * ----------------------------------------------------------*/

	/* Compute dll_dqs_delay_X value */
	dll_dqs_delay_X = (dll_dqs_delay_X_end_window + dll_dqs_delay_X_start_window) / 2;
	wr_dqs_shift = wr_dqs_shift_with_max_passing_cases;

	debug("DQS calibration - Window detected:\n");
	debug("max_passing_cases = %d\n", max_passing_cases);
	debug("wr_dqs_shift      = %d\n", wr_dqs_shift);
	debug("dll_dqs_delay_X   = %d\n", dll_dqs_delay_X);
	debug("dll_dqs_delay_X window = %d - %d\n",
	       dll_dqs_delay_X_start_window, dll_dqs_delay_X_end_window);

	/* -----------------------------------------------------------+
	 * De-assert 'start' parameter.
	 * ----------------------------------------------------------*/
	mtdcr(ddrcfga, DDR0_02);
	val = (mfdcr(ddrcfgd) & ~DDR0_02_START_MASK) | DDR0_02_START_OFF;
	mtdcr(ddrcfgd, val);

	/* -----------------------------------------------------------+
	 * Set 'wr_dqs_shift'
	 * ----------------------------------------------------------*/
	mtdcr(ddrcfga, DDR0_09);
	val = (mfdcr(ddrcfgd) & ~DDR0_09_WR_DQS_SHIFT_MASK)
		| DDR0_09_WR_DQS_SHIFT_ENCODE(wr_dqs_shift);
	mtdcr(ddrcfgd, val);
	debug("DDR0_09=0x%08lx\n", val);

	/* -----------------------------------------------------------+
	 * Set 'dqs_out_shift' = wr_dqs_shift + 32
	 * ----------------------------------------------------------*/
	dqs_out_shift = wr_dqs_shift + 32;
	mtdcr(ddrcfga, DDR0_22);
	val = (mfdcr(ddrcfgd) & ~DDR0_22_DQS_OUT_SHIFT_MASK)
		| DDR0_22_DQS_OUT_SHIFT_ENCODE(dqs_out_shift);
	mtdcr(ddrcfgd, val);
	debug("DDR0_22=0x%08lx\n", val);

	/* -----------------------------------------------------------+
	 * Set 'dll_dqs_delay_X'.
	 * ----------------------------------------------------------*/
	/* dll_dqs_delay_0 */
	mtdcr(ddrcfga, DDR0_17);
	val = (mfdcr(ddrcfgd) & ~DDR0_17_DLL_DQS_DELAY_0_MASK)
		| DDR0_17_DLL_DQS_DELAY_0_ENCODE(dll_dqs_delay_X);
	mtdcr(ddrcfgd, val);
	debug("DDR0_17=0x%08lx\n", val);

	/* dll_dqs_delay_1 to dll_dqs_delay_4 */
	mtdcr(ddrcfga, DDR0_18);
	val = (mfdcr(ddrcfgd) & ~DDR0_18_DLL_DQS_DELAY_X_MASK)
		| DDR0_18_DLL_DQS_DELAY_4_ENCODE(dll_dqs_delay_X)
		| DDR0_18_DLL_DQS_DELAY_3_ENCODE(dll_dqs_delay_X)
		| DDR0_18_DLL_DQS_DELAY_2_ENCODE(dll_dqs_delay_X)
		| DDR0_18_DLL_DQS_DELAY_1_ENCODE(dll_dqs_delay_X);
	mtdcr(ddrcfgd, val);
	debug("DDR0_18=0x%08lx\n", val);

	/* dll_dqs_delay_5 to dll_dqs_delay_8 */
	mtdcr(ddrcfga, DDR0_19);
	val = (mfdcr(ddrcfgd) & ~DDR0_19_DLL_DQS_DELAY_X_MASK)
		| DDR0_19_DLL_DQS_DELAY_8_ENCODE(dll_dqs_delay_X)
		| DDR0_19_DLL_DQS_DELAY_7_ENCODE(dll_dqs_delay_X)
		| DDR0_19_DLL_DQS_DELAY_6_ENCODE(dll_dqs_delay_X)
		| DDR0_19_DLL_DQS_DELAY_5_ENCODE(dll_dqs_delay_X);
	mtdcr(ddrcfgd, val);
	debug("DDR0_19=0x%08lx\n", val);

	/* -----------------------------------------------------------+
	 * Assert 'start' parameter.
	 * ----------------------------------------------------------*/
	mtdcr(ddrcfga, DDR0_02);
	val = (mfdcr(ddrcfgd) & ~DDR0_02_START_MASK) | DDR0_02_START_ON;
	mtdcr(ddrcfgd, val);

	ppcMsync();
	ppcMbar();

	/* -----------------------------------------------------------+
	 * Wait for the DCC master delay line to finish calibration
	 * ----------------------------------------------------------*/
	if (wait_for_dlllock() != 0) {
		printf("dlllock did not occur !!!\n");
		hang();
	}
	ppcMsync();
	ppcMbar();

	if (wait_for_dram_init_complete() != 0) {
		printf("dram init complete did not occur !!!\n");
		hang();
	}
	udelay(100);  /* wait 100us to ensure init is really completed !!! */
}

/*************************************************************************
 *
 * ecc_enable -- enable ECC dection and correction
 *
 ************************************************************************/
static void ecc_enable()
{
ulong val;
	/*	blb, Flow taken from spd_sdram_denali.c from AMCC*/

	/*	Enable ECC detection and correction*/
	mtdcr(ddrcfga, DDR0_22);
	val = (mfdcr(ddrcfgd) | DDR0_22_CTRL_RAW_ECC_ENABLE);
	mtdcr(ddrcfgd, val);

	/*	Clear memory*/
	printf("Enable ECC, Clear Memory...");
	memset(CFG_MEMTEST_START_FLASH_EXE, 0,
	(CFG_MEMTEST_END_FLASH_EXE - CFG_MEMTEST_START_FLASH_EXE) );
	printf("PASS:");

	/*	Clear any memory error interrupts we caused*/
	mtdcr(ddrcfga, DDR0_00);
	val = (mfdcr(ddrcfgd) | DDR0_00_INT_ACK_ALL);
	mtdcr(ddrcfgd, val);

	/*	Set 'int_mask' parameter to functionnal value */
	mtdcr(ddrcfga, DDR0_01);
	val=((mfdcr(ddrcfgd) & ~DDR0_01_INT_MASK_MASK) | DDR0_01_INT_MASK_ALL_OFF);
	mtdcr(ddrcfgd, val);
}


/*************************************************************************
 *
 * initdram -- 440EPx's DDR controller is a DENALI Core
 *
 ************************************************************************/
long int initdram (int board_type)
{
#if !defined(CONFIG_NAND_U_BOOT) || defined(CONFIG_NAND_SPL)
	volatile ulong val;

	mtsdram(DDR0_02, 0x00000000);  /* Clear START bit*/

	mtsdram(DDR0_00, 0x0000190A);  /* Set DLL_INCREMENT to 0x19, DLL_START_POINT to 0xA*/
	mtsdram(DDR0_01, 0x01000000);  /*PLB_DB_CS_LOWER = 0x1 (req'd), PLB_DB_CS_UPPER = 0x0 (req'd)*/

	/* BLR - I have modified Sequoia's timings for the CMC supporting the following two devices
             on our AVL:
             Micron MT47H64M16BT-5E-A
             Hynix HY5PS1G1631AFP-E3

             If new devices get added to the AVL, we must verify the timings are still satisfied
    */
	mtsdram(DDR0_03, 0x02030602); /* BSTLEN = 2 (req'd for DDR2), CSLAT = 3, INAREF= 2*/
	mtsdram(DDR0_04, 0x0A020200); /* TRC  - Active to Active (same bank) = 10 cycles,
                                     TRRD - Active to active (dfft bank) =  2 cycles,
                                     TRTP - Precharge time = 2 cycles */
                                  /* NOTE - MICRON - BLR - Min TRC = 10, TRRD = 2, TRTP = 2*/
                                  /* NOTE - HYNIX - BLR - Min TRC = 10, TRRD = 2, TRTP = 2*/

	mtsdram(DDR0_05, 0x02020307); /* TMRD - Mode register set time - 2,
                                     TEMRS - Ext. mode register set - 2,
                                     TRP - Pre-charge command time - 3,
                                     TRAS, Min row activate time -  7*/
                                  /* NOTE - MICRON - BLR - Min TMRD =2, TEMRS = 2, TRP = 3, TRAS = 7*/
                                  /* NOTE - HYNIX  - BLR - Min TMRD =2, TEMRS = 2, TRP = 3, TRAS = 7*/
	mtsdram(DDR0_06, 0x0102C816); /* WINTRP = 1 - Allows read's to interrupt write bursts,
                                     TWTR - Write to Read turnaround - 2,
                                     TDLL - DLL Lock Time - 200,
                                     TRFC - DRAM Refresh Time - 22*/
                                  /* NOTE - MICRON - BLR - min TWTR = 2, TDLL = 200, TRFC = 22*/
                                  /* NOTE - HYNIX  - BLR - min TWTR = 2, TDLL = 200(?), TRFC = 22*/
	mtsdram(DDR0_07, 0x00070100); /* NCI = 0 (req'd),
                                     TFAW - Four active window = 7,
                                     ARM = 1 (req'd),
                                     AREFR = 0 = (req'd) */
                                  /* NOTE - MICRON - BLR - min TFAW = 7*/
                                  /* NOTE - HYNIX  - BLR - min TFAW = 7*/
	mtsdram(DDR0_08, 0x02360001); /* WRLAT - Write Latency  = 2, (For DDR2 always CL-1)
                                     TCPD - Clock en to precharge = 54,
                                     DQSNEN = 0 (req'd),
                                     DDR2 1 (req'd, DDR2) */
                                  /* NOTE - MICRON - BLR - min TCPD = NOT GIVEN*/
                                  /* NOTE - HYNIX  - BLR - min TCPD = NOT GIVEN*/
	mtsdram(DDR0_09, 0x00011D5F); /* OAPDC0 = 0, (No idea what this does),
                                     RTT0 = 1 - 75Ohm,
                                     WR_DQS_SHIFT_BYPASS = 29,
                                     WR_DQS_SHIFT = 95 ~ 3/4 cycle*/
	mtsdram(DDR0_10, 0x00000300); /* WMDREG = 0 (No special EMRS activity)
                                     CSMAP = 3 (I think this COULD be 1, but why bother)
                                     OAPUC0 = 0 - Same as above? No clue, leaving it
                                     */
	mtsdram(DDR0_11, 0x0017C800); /* SREFR = 0, (Not needed here),
                                     TXSNR = 23, (Self refresh exit to non-Read)
                                     TXSR = 200, (Self refresh exit time)*/
                                  /* NOTE - MICRON - BLR - min TXSNR = 23, TXSR = 200 */
                                  /* NOTE - HYNIX  - BLR - min TXSNR = 23, TXSR = 200 */
	mtsdram(DDR0_12, 0x00000003); /* TCKE - CKE Pulse Width = 3 - Micron Agrees*/
	mtsdram(DDR0_14, 0x00010000); /* DLL_BYPASS_MODE = 0 (Req'd, basically)
                                     REDUC - 32/40 bit memory bus/ecc
                                     RDIM = 0 (not using registered DRAM)*/

	/* These are all for DQS control - Ok for defaults, but where is calibration?*/
	mtsdram(DDR0_17, 0x19000000); /* All delays set to "25" or ~20% of a cycle*/
	mtsdram(DDR0_18, 0x19191919);
	mtsdram(DDR0_19, 0x19191919);
	mtsdram(DDR0_20, 0x0B0B0B0B); /* DLL_DQS_DELAY_BYPASSes = 11*/
	mtsdram(DDR0_21, 0x0B0B0B0B);

	mtsdram(DDR0_22, 0x00267F0B); /* CTRL_RAW = No ECC (This is set elsewhere),
                                     DQSOSB - Doesn't matter, Not dll bypassed,
                                     DQSOSH - Doesn't matter, not dll bypassed,
                                     DDDB8 - Doesn't matter, not dll bypassed*/
	mtsdram(DDR0_23, 0x00000000); /*ORMC0 - CS0 gets ODT during ODT read*/
	mtsdram(DDR0_24, 0x01010002); /*RTTPT - 01 - 75Ohm,
                                    OWMC1 - 1 - Does not matter, CS1 unused,
                                    ORMC1 - 1 - Does not matter, CS1 unused,
                                    OWMC0 - 2 - Does not matter, CS1 unused */
	mtsdram(DDR0_26, 0x2D64050F); /* TRAS - Max row active time - 11620 cycles,
                                     TREF - Cycles between refresh - 1295 */
                                  /* NOTE - MICRON - BLR - max TRAS = 11620 cycles, TREF = 1295*/
                                  /* NOTE - HYNIX  - BLR - max TRAS = 11620 cycles, TREF = 1295*/
	mtsdram(DDR0_27, 0x0000682B); /* EMRS_DATA = 0, Extended mode register data,
                                     TINIT - DRAM init time = 26667 cycles*/
                                  /* NOTE - MICRON - BLR - TINIT = 33200 cycles)*/
                                  /* NOTE - HYNIX  - BLR - TINIT = 33200 cycles)*/
                                  /* NOTE - This can technically be 0, there's a 50ms delay in hardware for this*/
	mtsdram(DDR0_28, 0x00000000); /* EMRSn_DATA - Not used*/
	mtsdram(DDR0_31, 0x00000000); /* Not used, ECC Disabled*/
	mtsdram(DDR0_42, 0x01000006); /* APIN = 1 (Correct for CMC, we do not use A13),
                                     CLGATE = 6 (CL=3)*/
	mtsdram(DDR0_43, 0x030A0201); /* TWR - Dram write recovery time = 3 cycles,
                                     APREBIT - Bit 10 - Doesn't seem to matter for the EPx,
                                     COLSIZ  = 2, CMC uses 64Mx16 memories - or 10 Column bits
                                     BNK8 = 1, CMC uses 8-bank memories*/
                                  /*NOTE - MICRON - BLR - TWR - 3 cycles */
                                  /*NOTE - HYNIX  - BLR - TWR - 3 cycles */
	mtsdram(DDR0_44, 0x00000003); /*TRCD - Active to read/write = 3 cycles*/
                                  /*NOTE - MICRON - BLR - TRCD - 3 cycles */
                                  /*NOTE - HYNIX  - BLR - TRCD - 3 cycles */
	mtsdram(DDR0_02, 0x00000001); /* Set start bit */

#if 1
	wait_for_dlllock();
	/* -----------------------------------------------------------+
	 * Perform data eye search if requested.
	 * ----------------------------------------------------------*/
	denali_core_search_data_eye(CFG_MBYTES_SDRAM << 20);

#else
	/*
	 * Wait for DCC master delay line to finish calibration
	 */
	mfsdram(DDR0_17, val);
	while (((val >> 8) & 0x000007f) == 0) {
		mfsdram(DDR0_17, val);
	}
#endif

	ecc_enable(); // dhn - enable ECC


#endif /* #ifndef CONFIG_NAND_U_BOOT */

	return (CFG_MBYTES_SDRAM << 20);
}
