/****************************************************************************

  Copyright (c) 2006, Dell Inc
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

  - Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
    
  - Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution. 

  - Neither the name of Dell Inc nor the names of its contributors
    may be used to endorse or promote products derived from this software 
    without specific prior written permission. 

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. 

*****************************************************************************/


/*
 * Main line code for Dell CMC
 */
#include <common.h>		/*must be first in file*/
#include <command.h>

#ifdef CONFIG_CMC_10G	/*only compile in for a CMC_10G build*/
/*****************************************************************************
*	FILE: util_main.c
*
*****************************************************************************/

/*****************************************************************************
*  Include files
*****************************************************************************/
#include <util_defs.h>				/*includes fwupdate.h*/
#include <util_fpga.h>
#include <util_ret_codes.h>
#include <linux/mtd/nand.h>
#include <usb.h>
#include <nand.h>
#include "xiic_l.h"
#include <i2c.h>
#include <miiphy.h>


/*****************************************************************************
*  Local Defines
*****************************************************************************/
#define BLINK_RATE_MASK_RECOVER		0x000001ff		/*~1/2 second rate*/
#define BLINK_RATE_MASK_SELFTEST	0x0000007f		/*~1/8 second rate*/

#define PPID_LEN					20				/*Dell barcode*/


/*****************************************************************************
*  Globals
*****************************************************************************/
static u32 tstamp_post_start;
static u32 tstamp_post_last;
static s32 post_mode_display(char *msg);
static s32 get_cmc_fru(u32 fru_field,  char *field_string, char * msg);
static s32 post_verify_rtc(u32 rtc_ticks[], char *msg);
static s32 log_preboot_fwupd_info(u32 state, s32 rc, char *status_msg, 
															char *raclog_msg);

/* See ../drivers/nand/nand_util.c   DF249959*/
int new_bad_blocks_idx;						
int new_bad_blocks[CFG_FLASH_TOTAL_BLOCKS+1];	/*save a place for an end marker*/
				

u32	 g_post_rc;
char g_post_msg[MAX_POST_MSG_SIZE];
u32	 g_warn_rc;			
char g_warn_msg[MAX_POST_MSG_SIZE];

/*	Required for 2.6.24 kernel ECC fixes, see comment in code*/
u32 g_ecc_hack = 0;


/*****************************************************************************
*  Prototypes/Externs
*****************************************************************************/
extern char version_string[];

extern flash_info_t flash_info[];			/*nor flash into struct*/
extern int flash_autoselect_seq(flash_info_t * info, u16 * Manuf_ID, u16 * Dev_ID);

extern s32 get_micrel_en(void);
extern s32 set_micrel_en(u32 dir);
extern int miiphy_read(char *devname, unsigned char addr, unsigned char reg,
		unsigned short *value);
extern int miiphy_write(char *devname, unsigned char addr, unsigned char reg,
		unsigned short value);
extern  char *miiphy_get_current_dev(void);
extern  s32 parse_force_ecc_err_addrs(u32 *forced_addrs, u32 clear_env_flag);

extern u32 g_ecc_index;
extern u32 g_total_ecc_ints;	
extern u32 g_force_ecc_addr;

/*****************************************************************************
*	List all structures defines
*****************************************************************************/


/*****************************************************************************
*	List all "HW inventory" struct here as the central location.
*****************************************************************************/

/*	Manufacturer ID list for NAND	*/
t_hw_inventory_nand		hw_inv_nand[] = {
	{NAND_MFR_SAMSUNG,	"Samsung"},
	{NAND_MFR_HYNIX,	"Hynix"},  
	{NAND_MFR_STMICRO,	"ST Micro"},
	{0x0, "Unknown"} };

t_hw_inventory_nor		hw_inv_nor[] = {
	{NOR_MFR_SPANSION,	"Spansion", NOR_DEV_SPANSION, "S29AL032D"},
	{NOR_MFR_MACRONIX,	"Macronix", NOR_DEV_MACRONIX, "MX29LV320C"  },
	{NOR_MFR_ATMEL,		"Atmel",    NOR_DEV_ATMEL,    "AT49BV322DT"  },
	{0x0,               "Unknown",	0x0,              "Unknown"} };


t_hw_inventory_fcr		hw_inv_fcr[] = {
	{FCR_MFR_SMSC,		FCR_PRODUCT1_SMSC,	"SMSC"},
	{FCR_MFR_SMSC,		FCR_PRODUCT2_SMSC,	"SMSC"},
	{0x0, 0x0, "Unknown"} };


/*	Manufacture IDs for Lan devices*/
t_hw_inventory_phy micrel_phy_def = {
        .phy_id         = 0x00221400,
        .phy_id_mask    = 0x0ffffff0,
        .vendor_name    = "Micrel KSZ8041TL phy",
        };

t_hw_inventory_phy bcm5481_phy_def = {
        .phy_id         = 0x0143bca0,
        .phy_id_mask    = 0x0ffffff0,
        .vendor_name    = "BCM5481 Gigabit phy",
        };

t_hw_inventory_phy bcm5324_phy_def = {
        .phy_id         = 0x0103bc20,
        .phy_id_mask    = 0x0f0ffff0,
        .vendor_name    = "BCM5324 24Port Switch",
};



/*****************************************************************************
*
*	FUNCTION: util_dellcmc_main()
*	DATE:	  9-20-06
*
*****************************************************************************/
s32 util_cmc_main(void)
{
s32		rc = 0;
s32		rc_tmp = 0;
s32		saved_silent;
s32     tftp_retry;
s32     fpga_rc = 0;
s32		fwupd_image_bad = 0;
s32		tftp_image_bad  = 0;
s32     need_fwupd = 0; 
u32		fpga_flags = 0;
u32     total_remapped_blocks = 0;
u32		therm_count;
u32     rtc_ticks[2];
u32		firmimg_size;
u32		fecc_flag;
u32		faddrs[MAX_FORCE_ECC_BUF];	/*NAND addresses to force ECC read error*/
char	*env_p;
char	*env_single_p;
char	*ppid_p;
char    ppid_str[PPID_LEN+1];						/* + null*/
char	msg[MAX_MSG_LEN];	
char	fwupd_msg[MAX_MSG_LEN];	
char	temp_msg[MAX_MSG_LEN + BDINFO_MAX_LEN_NAME];	
char	tstamp_str[MAX_TSTAMP_STR_LEN];


/*****************************************************************************
*	Initializations
*****************************************************************************/
tstamp_post_start = ABS_get_timer(0);	/*get time stamp to calc elapsed time*/
tstamp_post_last  = tstamp_post_start;

g_post_rc		= 0;		/*default POST status indications*/
g_post_msg[0]	= '\0';

g_warn_rc		= 0;		/*default POST warning status indications*/
g_warn_msg[0]	= '\0';

fwupd_msg[0]	= '\0';

log_preboot_init();					/*init preboot log area*/
log_preboot(LOG2TRACE, LOG_INFO, "Start-up -to- util_cmc_main()" );

/*Init array used to hold newly found erase block failures*/
new_bad_blocks_idx = 0;
new_bad_blocks[new_bad_blocks_idx] = BAD_BLOCK_ARRAY_END_MARKER;


/*****************************************************************************
*	Real Time Clock, Time-of-Day(TOD) Read.  We cannot go into an extensive 
*	test for the RTC.  We are simply going to make sure that it is ticking.
*	Do a read here and at the end of POST and make sure it is different.
*****************************************************************************/
if( (rc = get_rtc_tod(&rtc_ticks[0])) != 0)
	rtc_ticks[0] = -1;


/*****************************************************************************
*	Load the FPGA 
*****************************************************************************/
rc = post_fpga_load(fpga_flags, msg);
fpga_rc = rc;

/*
*	HW has added a vendor mode bit on the CMC backplane connector.  This was
*	to be CMCF_BLANK_N.  This way PCBA test can control vendor mode from the 
*	connector and NOT rely on the 2 pin header. The FPGA is now loaded and
*	the mode reg reflects 2pin header, AND BP connector pin.  We need to 
*	re-evaluate the silent console so we can enable messages.
*/
if( (0 == fpga_rc) && (is_fpga_vendor_mode()) )
	clear_save_silent_mode();		/*do not care about saving current mode*/

log_preboot(LOG2TRACE, rc, msg);	/*log msg now that vendor mode is known*/

/****************************************************************************
*	See if we are forcing double bit NAND ECC errors. This first error offset
*	encountered will trigger a nand_ecc_scan() so it MUST be in the kernel or
*	rootfs area. NOTE: in vendor mode rootfs is NOT CRCed.  The forced 
*	addresses are setup by "util flash -fecc <offset>:<offset>:  ..."
****************************************************************************/
g_force_ecc_addr = 0;	/*default to not force ECC errors*/
if(is_vendor_mode())
	{
	memset((void*)faddrs, 0,  MAX_FORCE_ECC_BUF );
	fecc_flag = parse_force_ecc_err_addrs(faddrs, 0);
	if(fecc_flag)
		g_force_ecc_addr = faddrs[0];
	}

/*****************************************************************************
*	Display CMC hard strapped revision bits and monitor version.  
*	We do NOT want to embed a CMC FW Version string in u-boot as it is 
*	possible to spin FW and not u-boot (if no change is required).
*	The CMC FW is taken from the image file, displayed, and put into a env 
*	later on in this function.
*****************************************************************************/
sprintf(msg, "%s", version_string);
log_preboot(LOG2TRACE, LOG_INFO, msg);

/*	
*	Add the 20 char Dell barcode as useful Debug info that will show up in the
*	trace log.	Protect ourselves, do NOT take over the expected 20.
*/
memset(ppid_str, 0, PPID_LEN+1);
if((ppid_p = getenv("PPID")) != 0)		/*20 digit Dell barcode*/
	strncpy(ppid_str, ppid_p, PPID_LEN);

sprintf(msg, "CMC %s Rev Strappings=%d", ppid_str, get_cmc_rev() );
log_preboot(LOG2TRACE, LOG_INFO, msg);

sprintf(msg, "CMC Mid-plane rev=%d  CPLD rev=%d", 
					(ABS_read_fpga(FPGA_REG_MP_REV) & FPGA_MP_REV_MASK),  
					(ABS_read_fpga(FPGA_REG_CPLD_REV) & FPGA_CPLD_REV_MASK) );
log_preboot(LOG2TRACE, LOG_INFO, msg);

/*****************************************************************************
*	11-2-07 CR179202  The "Board Revision" was simply pin strappings from 3 
*	GPIO's.  Previous RAC products never had a FRU so they used the production
*	sector for this.  The pin strappings and the FRU Rev fields are NOT the 
*	same thing.   Linux cfgmgr is keying on the "Board Revision" field output
*	from the "uutil post -dump" command.  So replace the pin straps rev with
*	the FRU rev and make a new text line for "CMC Rev strappings". 
*****************************************************************************/
rc = get_cmc_rev_from_fru(msg);
log_preboot(LOG2TRACE, rc, msg);

/******************************************************************************
*	After AC power up it takes time for the 7460 to complete its initial poll
*	cycle.  The temp readings would always fail after AC (128 deg), but work
*	after reset. Move the init here and use the time it takes to run the bulk 
*	of the POST test as a delay. This is better than adding a brain dead delay. 
******************************************************************************/
rc = post_adt7460_init(msg);
log_preboot(LOG2TRACE, rc, msg);

/*****************************************************************************
*	Chassis component control
*****************************************************************************/
/*	
*	If the FPGA load fails THIS CMC cannot drive external controls, so no 
*	reason to be involved in chassis setup as we will soon be in recover.
*/
if(fpga_rc == 0)
	{
	rc = post_chas_setup(msg);
	log_preboot(LOG2TRACE, rc, msg);

	if(!is_local_active() )
		post_passive_setup(msg);		/*DF125823 init components on passive CMC*/
	
	rc = post_cmc_status(msg);
	log_preboot(LOG2TRACE, rc, msg);

	/*	Early HW bugs require a different setup*/
	env_p = getenv("post_chas");
	if(env_p)
		{
		if( !strcmp(env_p, "psubug") )
			{
			rc = post_chas_setup_psubug(msg);
			log_preboot(LOG2TRACE, rc, msg);
			}
		}
	}

/*****************************************************************************
*	Test coin cell battery
*****************************************************************************/
if( ABS_read_fpga(FPGA_REG_ID) == FPGA_ID)
	{
	if( ABS_read_fpga(FPGA_REG_VBAT_PWR_GOOD) & FPGA_VBAT_PWR_GOOD)
		{
		sprintf(msg, "Coin Cell VBAT PWR Good");
		rc = 0;
		}
	else
		{
		sprintf(msg, "Failure in Coin Cell VBAT PWR Good");
		rc = W_POST_VBAT_PWR_FAIL;
		}
	log_preboot(LOG2TRACE, rc, msg);
	}


/*****************************************************************************
*	BLR - FPGA I2C Initialization, these cannot fail
*****************************************************************************/
xiic_init(CMC_I2C_PSU123);
xiic_init(CMC_I2C_PSU456);
xiic_init(CMC_I2C_IOMSET1);
xiic_init(CMC_I2C_IOMSET2);
xiic_init(CMC_I2C_LKVM);


/*****************************************************************************
*	ID flash
*****************************************************************************/
rc = post_id_nor(msg);
log_preboot(LOG2TRACE, rc, msg);

rc = post_id_nand(msg);
log_preboot(LOG2TRACE, rc, msg);

rc = post_badblock_nand(&total_remapped_blocks, msg);
log_preboot(LOG2TRACE, rc, msg);

/****************************************************************************
*	If the current number of bad/remapped NAND blocks returned by
*	post_badblock_nand() exceeds a threshold then create a raclog message/OS 
*	alert for S99final to inform the customer they need to replace the CMC.
****************************************************************************/
#if 0 /*architect/system team did not want to put this in yet. 1-14-09 blb*/
if(total_remapped_blocks >= MAX_BAD_BLOCKS_ALLOWED)
	{
	sprintf(msg, "CMC%d: Replace CMC due to NAND device degrading.", 
				(ABS_read_fpga(FPGA_REG_CMC_SLOT_ID) & FPGA_CMC_SLOT_ID) + 1);
	log_preboot((LOG2TRACE | LOG2RAC | LOG2OSEVENT), LOG_INFO, msg );
	}
#endif

/*****************************************************************************
*	If MAC ethaddr equals the default then fail, manuf has not written it.
*****************************************************************************/
rc = post_test_mac(FRU_GET_MAC0, msg);
log_preboot(LOG2TRACE, rc, msg);

rc = post_test_mac(FRU_GET_MAC1, msg);
log_preboot(LOG2TRACE, rc, msg);

/*****************************************************************************
*	ID Networking components
*****************************************************************************/
rc = post_id_phy(PHY_SWITCH, msg);
log_preboot(LOG2TRACE, rc, msg);

rc = post_id_phy(PHY_GIG0, msg);
log_preboot(LOG2TRACE, rc, msg);

rc = post_id_phy(PHY_GIG1, msg);
log_preboot(LOG2TRACE, rc, msg);

if(get_cmc_rev() >= 0x2 )
	{
	rc = post_id_phy(PHY_MICREL, msg);
	log_preboot(LOG2TRACE, rc, msg);
	}


/*****************************************************************************
*	ID USB.  Cannot run this at post time.  I faulty clock locks up the 
*	CMC.  PCBA test can run "util post -usb" if desired.
*****************************************************************************/
#if 0
rc = post_id_usb(msg);
log_preboot(LOG2TRACE, rc, msg);
#endif

/*****************************************************************************
*	ID I2C components
*****************************************************************************/
rc = post_i2c_bus0_scan(msg);
log_preboot(LOG2TRACE, rc, msg);

rc = post_id_rtc(msg);
log_preboot(LOG2TRACE, rc, msg);

rc = post_id_temp(msg);
log_preboot(LOG2TRACE, rc, msg);

rc = post_id_fru(msg);
log_preboot(LOG2TRACE, rc, msg);

rc = post_id_adt7460(msg);
log_preboot(LOG2TRACE, rc, msg);


/*****************************************************************************
*	See if we are forcing a self test error.
*****************************************************************************/
rc = post_force_error(msg);
log_preboot(LOG2TRACE, rc, msg);

/*****************************************************************************
*	Mem ECC
*****************************************************************************/
rc = post_ecc_init(msg);
g_ecc_index		 = 0;
g_total_ecc_ints = 0;
log_preboot(LOG2TRACE, rc, msg);


/*****************************************************************************
*	Display the Thermal trip count.
*****************************************************************************/
if( (therm_count = get_thermtrip_count()) != 0)
	rc = W_POST_THERM_TRIP_NON_ZERO;

/*	Do not scare the customer, make a non-threating/criptic/pleasing msg*/
sprintf(msg, "CMC HW TTCNT=%d ", therm_count);
log_preboot(LOG2TRACE, rc, msg);

/*****************************************************************************
*	Display password jumper msg
*****************************************************************************/
if( is_hard_password_mode() )
	sprintf(msg, "Password jump IS detected");
else
	sprintf(msg, "Password jump is NOT detected");
	
log_preboot(LOG2TRACE, LOG_INFO, msg);


/*****************************************************************************
*	Display usefull mode information
*****************************************************************************/
rc = post_mode_display(msg);
log_preboot(LOG2TRACE, LOG_INFO, msg);


/*****************************************************************************
*	See if we have a trigger for a firmware update. 
*
*	After u-boot/NOR update the CMC resets in order to run with ther latest 
*	code and does not return from fwupd().
*	The function is_fwupd_needed() will flag the fact that we are not yet 
*	complete and recall fwupd().    After the fwupd is marked complete the 
*	boot will continue. Improper AC loss will cause the update to reoccur as 
*	is_fwupd_needed() detects this. 
*
*	At the completion of the post_verify_fwupd_image() the firmimg.cmc image
*	will be left in memory for a pending update or repair to use. 
*
*	On a boot the firmimg.cmc binary will be verified in S99final.sh, a OS 
*	event will be generated if it is bad.
*****************************************************************************/
need_fwupd = is_fwupd_needed(msg);		/*PORT_2.6.25  set global g_ecc_hack*/
log_preboot(LOG2TRACE, need_fwupd, msg );	
/*DO NOT CORRUPT msg JUST YET, NEEDED AFTER "if(need_fwupd)"*/

if(need_fwupd)
	{
	/*	
	*	Enable video to fwupdate messages.
	*
	*	Just because we need an update don't mean we can.  Verify the 
	*	redundant back up. If it fails it is INFO.  We do not want a selftest
	*	failure IF the original OS image is good and we can boot.
	*/
	saved_silent = clear_save_silent_mode();
	/*If we need a fwupd, display the reason now that it is visable*/
	if(!is_vendor_mode() )
		log_preboot(LOG2TRACE, need_fwupd, msg );

	/*	
	*	If I_FWUPD_TRIGGER_FROM_PEER (soft recover FPGA bit) then the image 
	*	must be loaded from the peer, verified, and copied to NAND. We do 
	*	not need to waste time verifing the existing image file as we are 
	*	going to replace it.
	*/
	if(I_FWUPD_TRIGGER_FROM_PEER == need_fwupd)
		{
		log_preboot(LOG2TRACE, LOG_INFO, "Loading image from PEER CMC...");
		for(tftp_retry = 0; tftp_retry < MAX_TFTP_RETRY; ++tftp_retry)
			{
			if((tftp_image_bad = tftp_from_peer(msg)) != 0)	
				{
				sprintf(temp_msg, "FWupdate from PEER failure retry-%d %s", 
															tftp_retry, msg);
				log_preboot(LOG2TRACE, LOG_INFO, temp_msg);
				}
			else				
				break;	/*tftp from peer, verify and commit to NAND worked*/
			}

		clear_fwupd_req_from_peer();
		if(tftp_retry >= MAX_TFTP_RETRY)
			{
			sprintf(temp_msg, "FWupdate from PEER %s failed", "test");
			log_preboot(LOG2TRACE, LOG_INFO, temp_msg);
			tftp_image_bad = 1;
			}
		}
	else
		{
		/*
		*	We are NOT a load from PEER so get existing image and verify.
		*	Image is so large it takes many seconds to read from NAND and verify.
		*	Inform the user what is occuring.
		*/
		log_preboot(LOG2TRACE, LOG_INFO, "Verifying update image please wait...");
		fwupd_image_bad = post_verify_fwupd_image(CFG_LOAD_ADDR, msg); 
		log_preboot(LOG2TRACE, LOG_INFO, msg );
		}

	if(fwupd_image_bad)
		{
		/*	
		*	Requesting a fwupd, BUT the back up is bad.  So retest the OS 
		*	images.  If they are good, boot from the old images and let 
		*	S99final.sh re-verify the backup and log an OS event. 
		*/
		if( (rc = post_verify_os_images(msg)) != 0)
			{
			/*	We are one step away from death. 1)bad OS images 2) no backup
			*	image 3)update from peer is not set. Can you say recover...
			*	As a last ditch effort try a peer update. If it works then we
			*	will have the back up flashed, so reset.  Then on the next go 
			*	round I_FWUPD_TRIGGER_BAD_OS_IMAGE will force an update, we 
			*	will NOT be in an infinite loop.
			*/
			log_preboot(LOG2TRACE, LOG_INFO, "Recover image from PEER CMC...");
			for(tftp_retry = 0; tftp_retry < MAX_TFTP_RETRY; ++tftp_retry)
				{
				if((tftp_image_bad = tftp_from_peer(msg)) != 0)	
					{
					sprintf(temp_msg, "FWupdate from PEER failure retry-%d %s", 
															tftp_retry, msg);
					log_preboot(LOG2TRACE, LOG_INFO, temp_msg);
					}
				else				
					break;	/*tftp from peer, verify and commit to NAND worked*/
				}

			if(tftp_retry >= MAX_TFTP_RETRY)
				{
				g_post_rc  = R_RECOVER_BAD_BACKUP_AND_OS;		
				strcpy(g_post_msg,
				"FWupdate is required, no valid backup/peer image was located");
				tftp_image_bad = 1;
				}
			else
				reset_after_fwupd();
			}
		else
			{
			/*	TODO: need a OS event ???*/
			strcpy(msg, "FWupdate requested but backup image is bad, boot with original");
			log_preboot(LOG2TRACE, rc, msg);
			log_preboot_fwupd_info(FWUPD_STATE_FAIL, fwupd_image_bad, 
													FWUPD_MSG_STATE_FAIL, msg);
			}		
		}
	else if(need_fwupd && !tftp_image_bad)
		{
		/*
		*	PORT_2.6.25     Going DOWN from 2.1 or >,  to 1.x
		*	A firmimg.cmc that has been written using >=2.1 ECC algorithm. 
		*	This u-boot >=2.1 code tests the flag indicating a down grade to a 
		*	u-boot that has no knowledge of different ECC algorithms. 
		*	1) verify the image 2) determine a downgrade boundary change
		*	3) re-write the firmimg.cmc that is in memory using the OLD ECC
		*	algorithm so the next reboot (old u-boot) will work.
		*/
		/*IF down-rev'ing, flag will be set by previous is_fwupd_needed() call*/
		if(-1 == g_ecc_hack)	
			{
			/*	call verify image to simply to return the size*/
			if( (rc_tmp = fwupd_verify_image(CFG_LOAD_ADDR, &firmimg_size, msg)) != 0)
				{
				sprintf(msg, "Write-back update image verify failed RC=0x%x", rc_tmp);
				log_preboot(LOG2TRACE, LOG_INFO, msg ); /*HACK what to do???*/
				}	
			else
				{
				printf("\nRewriting image file using old ECC algorithm.");
				if(fwupd_ram2flash(CFG_LOAD_ADDR, firmimg_size, msg) != 0)
					log_preboot(LOG2TRACE, LOG_INFO, msg ); /*HACK what to do ???*/
				}
			}

		/*	
		*	This env will protect us upon an abnormal termination (EST clowns 
		*	pulling AC at every whim). If we detect this env flag, fwupd() will
		*	be called again to repair any incomplete or corrupt image, and will 
		*	only be cleared by fwupd() when complete.
		*/
		set_fwupd_state_env(FWUPD_ENV_STATE_PENDING);
		rc = fwupd(CFG_LOAD_ADDR, fwupd_msg);	/*resets after NOR u-boot update*/
		log_preboot(LOG2TRACE, rc, msg );

		/*
		*	PORT_2.6.24     Going UP from 1.x to >= 2.0.
		*	This code does not exist in u-boot 1.0 so we  1) verify the image
		*	2) update NOR images 3) reset.    NOW we are running a new u-boot 
		*	but a firmimg.cmc that has been written using 1.0 ECC algorithm.  
		*	This new u-boot must now re-write the firmimg.cmc that is in memory
		*	using the new ECC algorithm or future resets will die.
		*/
		/*IF up-rev'ing, flag will be set by previous is_fwupd_needed() call*/
		if(1 == g_ecc_hack)	
			{
			/*	call verify image to simply to return the size*/
			if( (rc_tmp = fwupd_verify_image(CFG_LOAD_ADDR, &firmimg_size, msg)) != 0)
				{
				log_preboot(LOG2TRACE, LOG_INFO, msg );
				sprintf(msg, "Write-back update image verify failed RC=0x%x", rc_tmp);
				log_preboot(LOG2TRACE, LOG_INFO, msg ); /*HACK what to do???*/
				}	
			else
				{
				printf("\nRewriting image file using new ECC algorithm.");
				if(fwupd_ram2flash(CFG_LOAD_ADDR, firmimg_size, msg) != 0)
					log_preboot(LOG2TRACE, LOG_INFO, msg ); /*HACK what to do ???*/
				}
			}

		/*
		*	To get here the NOR and NAND images have been attempted.
		*	A failure will put us into recover. At the end of fwupd() the 
		*	fwupd_state will be "pass" or "fail".
		*	
		*	fwupdate is done, restore the silent/vendor mode
		*/	
		if(0 == rc)
			{
			log_preboot_fwupd_info(FWUPD_STATE_PASS, 0, 
													FWUPD_MSG_STATE_PASS, fwupd_msg);

			printf("\n\n** FWupdate complete, booting OS **\n\n");
			}

		restore_silent_mode(saved_silent);
		}
	}/*end if need_fwupd*/

g_ecc_hack = 0; 
/*****************************************************************************
*	Verify Real Time Clock Operation.
*****************************************************************************/
if( (rc = get_rtc_tod(&rtc_ticks[1])) != 0)
	rtc_ticks[1] = -1;

rc = post_verify_rtc(rtc_ticks, msg);
log_preboot(LOG2TRACE, rc, msg);

/*****************************************************************************
*	See if we have any NAND single bit ECC errors.  If so create an obscure 
*	message so we can at least see it in the trace log.
*****************************************************************************/
if((env_single_p = (char*)getenv(SINGLE_BLOCKS_ENV_NAME)) != 0)
	{
	sprintf(msg, "CMC HW  NDSBE_TCNT_TBLK=%s", env_single_p);
	log_preboot(LOG2TRACE, LOG_INFO, msg );
	}

/*****************************************************************************	
*	To reprogram the CPLD under u-boot an OS app may choose to load a new 
*	bit stream/vme file into a dedicated sector.  Then upon reset program the 
*	CPLD and erase the sector so we only do it once.  It may be deamed safer 
*	to program in u-boot and not in Linux.  The CMC will AC cycle after 
*	programming.
*
*	NOTE: DO THIS CALL AT THE "END" OF util_cmc_main*() "AFTER" FWUPDATE.
*	IN THE FUTURE FWUPDATE MAY WRITE THE SECTOR WITH THE NEW VME FILE.  
*	AFTER CPLD PROGRAMMING WE AC CYCLE AND LOSE ANY FPGA MODE BITS, OR UPDATE
*	TRIGGER BY OS.  THIS WILL INSURE FWUPDATE HAS COMPLETED "BOTH" PASSES, 
*	THEN ANY CPLD UPDATE BEFORE OS BOOT.
*****************************************************************************/
saved_silent = clear_save_silent_mode();	/*make visable*/
cpld_check(msg);
restore_silent_mode(saved_silent);



/************************************************************************
*	Collect misc status indications for usefull info for the log.	
************************************************************************/
/*	Save self test error cocde history*/
if(g_post_rc)
	post_errorocde_history(g_post_rc);

/*	If no failures then log a pass message.*/
if(g_post_rc == 0)
	strcpy(g_post_msg, "PASS");


/*	If no warnings then log a pass message.*/
if(g_warn_rc == 0)
	strcpy(g_warn_msg, "PASS");
	

//	Log POST info to be passed up to S99final.
sprintf(temp_msg, "%s=0x%x", BDPROP_POST_RC, g_post_rc);
log_preboot(LOG2PROP, LOG_OS_TARGET, temp_msg);

sprintf(temp_msg, "%s=%s", BDPROP_POST_MSG, g_post_msg);
log_preboot(LOG2PROP, LOG_OS_TARGET, temp_msg);

calc_tstamp_str(tstamp_post_start, tstamp_str);
sprintf(msg, "Start-up -to- exit of util_cmc_main()  %s", tstamp_str);
log_preboot(LOG2TRACE, LOG_INFO, msg );

/*	We are done logging, cacl crc*/
log_preboot_crc_gen();

return(0);
}/*end FUNCTION: util_main() */



/*****************************************************************************
*	Support functions for util_cmc_main()
*****************************************************************************/

/************************************************************************
*	Function: log_preboot_init()
*	Ported:	  9-14-06
*	Initialize the pre-boot log located in the paramater area.
************************************************************************/
s32 log_preboot_init(void)
{
DECLARE_GLOBAL_DATA_PTR;	/*necessary to get info from gd_t*/
bd_t *bd = gd->bd;
t_pram_preboot *pram  = (t_pram_preboot*) (bd->bi_memsize-(CONFIG_PRAM*1024));
t_preboot_log  *log   = (t_preboot_log*) &pram->log;
t_fwupd_status *fwupd = (t_fwupd_status*)&pram->fwupd_info;

#if 0	/*debug display*/
printf("\nlog_preboot_init()\n");
printf("pram addr=0x%x log addr=0x%x\n", pram, log);
#endif

/*	Clear all msg logs, init log pointers*/
memset(log, 0, PREBOOT_LOG_MAX_SIZE );
memset(fwupd, 0, sizeof(t_fwupd_info) );

log->bytes_remaining	= PREBOOT_LOG_BUFFER_MAX_SIZE - 
						  strlen(PREBOOT_LOG_FULL_MSG) + 1;	/*plus null*/
log->next_offset		= (char*)((s32)log+ sizeof(t_preboot_log));  
log->total_bytes		= sizeof(t_preboot_log);
return(0);
}//end FUNCTION: preboot_log_init()


/*****************************************************************************
*	Function:	log_preboot()
*	PORTED:		9-14-06
*
*	Add entry to pre-boot log.
*
*	INPUT:	flags  - Used by the OS to determine the target OS log.
*			rc     - Used to determine LOG_ERROR | LOG_INFO, and save
*					 into POST results
*			logstr - Text string to log
*
*	RETURN: 0		- entry added to log ok
*			non-0	- log is full (write a defect)
*****************************************************************************/
s32 log_preboot(u32 flags, s32 rc, char *logstr)
{
DECLARE_GLOBAL_DATA_PTR;	/*necessary to get info from gd_t*/
s32		log_rc  = 0;
s32		len		= 0;
char	*lp;
char	logmsg[MAX_MSG_LEN + 20];
char	tstamp_str[MAX_TSTAMP_STR_LEN];

bd_t *bd = gd->bd;
t_pram_preboot *pram = (t_pram_preboot*) (bd->bi_memsize-(CONFIG_PRAM*1024));
t_preboot_log  *log  = (t_preboot_log*)&pram->log;

/*
*	As noted in the CMC_Pre-boot_and_POST.doc WARN, or warnings are a special 
*	case. In vendor mode they are warnings and PCBA test can treat them as 
*	errors.  In NON-vendor mode they are INFO messages.  This catagory is for 
*	the HW inventory class.   Change the message type here.
*/
if( (rc & LOG_WARN) && !is_vendor_mode() )
	rc = ((rc & LOG_MASK) | LOG_INFO);


tstamp_post_last = calc_tstamp_str(tstamp_post_last, tstamp_str);

if(rc & LOG_ERROR)				/*error message*/
	{
	sprintf(logmsg, "FAIL: %s %s\n", tstamp_str, logstr);

	if(g_post_rc == 0)			/*1st selftest error?*/
		{
		g_post_rc = rc;			/*save the 1st error code and msg only*/
		strcpy(g_post_msg, logstr); /*do not want timestamp*/
		if('\n' == g_post_msg[strlen(g_post_msg)-1])
			g_post_msg[strlen(g_post_msg)-1] = '\0'; /*remove the \n*/
		}
	}
else if(rc & LOG_WARN)			/*warning message*/
	{
	sprintf(logmsg, "WARN: %s %s\n", tstamp_str, logstr);	

	if(g_warn_rc == 0)		/*1st warning error?*/
		{
		g_warn_rc = rc;			/*save the 1st error code and msg only*/
		strcpy(g_warn_msg, logstr); /*do not want timestamp*/
		if('\n' == g_warn_msg[strlen(g_warn_msg)-1])
			g_warn_msg[strlen(g_warn_msg)-1] = '\0'; /*remove the \n*/
		}
	}
else if(rc & LOG_INFO)			/*informational message*/
	sprintf(logmsg, "INFO: %s %s\n", tstamp_str, logstr);	
else if(rc & LOG_OS_TARGET)	/*OS msg does not need INFO/PASS/timestamp*/
	sprintf(logmsg, "%s\n", logstr);	
else							/*Pass message*/
	sprintf(logmsg, "PASS: %s %s\n", tstamp_str, logstr);


/*if log full, (write a defect), increase buf size, or entries*/
if( log->full || (log->index >= PREBOOT_LOG_MAX_ENTRIES) )
	return(E_LOG_ALREADY_FULL);							


len = strlen(logmsg);								
/*save last idx for full msg*/
if((log->index == PREBOOT_LOG_MAX_ENTRIES-1) || (log->bytes_remaining < len)) 
	{
	strcpy(log->next_offset, PREBOOT_LOG_FULL_MSG);
	log->full = 1;
	log_rc = E_LOG_IS_NOW_FULL;
	len = strlen(PREBOOT_LOG_FULL_MSG);
	}
else
	strcpy(log->next_offset, logmsg);

lp = log->next_offset;				/*save for printf*/
log->msg_p[log->index]	= log->next_offset;
log->flags[log->index]	= flags;
++log->index;
log->msg_p[log->index]	=  0;		/*NULL last entry+1*/
log->bytes_remaining	-= len;
log->total_bytes		+= (len+1);	/*null must be part of total byte count*/
log->next_offset		+= (len + 1);

if( !(rc & LOG_OS_TARGET) )			/*do not display msgs targeted for OS only*/
	printf("%s", lp);				/*Write message to u-boot stdio*/
return(log_rc); 
}// end FUNCTION: log_preboot()

/*****************************************************************************
*	FUNCTION:	log_preboot_crc_gen()
*	DATE:		7-16-07
*
*	Call this function when main is done writing any pre-boot log messages.
*	We need to crc this area so that Linux can verify the Integrity in 
*	S99final.sh before moving any messags into Linux logs.
*****************************************************************************/
s32 log_preboot_crc_gen(void)
{
DECLARE_GLOBAL_DATA_PTR;	/*necessary to get info from gd_t*/
bd_t *bd = gd->bd;
t_pram_preboot *pram = (t_pram_preboot*) (bd->bi_memsize-(CONFIG_PRAM*1024));
t_preboot_log  *log  = (t_preboot_log*)&pram->log;

/*1st crc element is NOT part of crc, inc the addr, dec the count*/
log->crc = crc32(0, (u8*)&log->total_bytes, log->total_bytes - 4);

return(0);
}/*end FUNCTION:log_preboot_crc_gen() */ 


/*****************************************************************************
*	Function:	log_preboot_fwupd_info()
*	DATE:		8-13-07
*
*	Add entry to pre-boot fwupd info area.  this data will be used by
*	syncd to update fwupdate state files.
*****************************************************************************/
static s32 log_preboot_fwupd_info(u32 state, s32 rc, char *status_msg, 
															char *raclog_msg)
{
DECLARE_GLOBAL_DATA_PTR;	/*necessary to get info from gd_t*/
bd_t *bd = gd->bd;
t_pram_preboot *pram = (t_pram_preboot*) (bd->bi_memsize-(CONFIG_PRAM*1024));
t_fwupd_info   *fwupd_info  = (t_fwupd_info*)&pram->fwupd_info;


fwupd_info->fwupd_status.state	= state;
fwupd_info->fwupd_status.rc		= rc;
strcpy(fwupd_info->fwupd_status.msg,  status_msg);
strcpy(fwupd_info->fwupd_raclog_msg,  raclog_msg);

/*
*	Recalc the crc. 
*	1st crc element is NOT part of crc, inc the addr, dec the count
*/
fwupd_info->crc = crc32(0, (u8*)(u32)fwupd_info + 4, sizeof(t_fwupd_info) - 4);
return(0);
}/*end FUNCTION: log_preboot_fwupd_info()*/


/************************************************************************
*	Function: post_test_mac()
*	DATE:	9-20-06
*
*	Get the MAC/ethaddr from the env area and make sure it has been 
*	programmed.  If it is the default value we know the vendor messed up.
*	Else, we can only make sure it has 6 ":" seperated fields, and that 
*	they contain vaild decimal numbers between 0-255.
************************************************************************/
s32 post_test_mac(u32 mac, char *msg)
{
s32		rc = 0;
s32		error = 0;
s32		new_mac = 0;
char	*dev_str;
char	*env_str;
char	mac_str[MAX_MAC_STR_LEN];
char	*s;

if(mac == FRU_GET_MAC0)
	{
	env_str = "ethaddr";
	dev_str = "MAC 0";
	s = getenv("ethaddr");
	}
else
	{
	env_str = "eth1addr";
	dev_str = "MAC 1";
	s = getenv("eth1addr");
	}

if(s == 0)
	{
       if( (rc = get_cmc_fru(mac, mac_str, msg)) != 0)
		{
          return(E_POST_MAC_FRU_READ_FAILURE);
		}

	++new_mac;
	s = mac_str;
	}

if( (rc = mac_verify_colon(s, msg)) != 0)
{   
    ++error;
    //sprintf(msg, "Invalid %s: %s", dev_str, s);		//setup error
    return(rc);
}

if(!((strcmp(DEFAULT_ETHADDR_0, s) && 
	  strcmp(DEFAULT_ETHADDR_1, s) &&
	  strcmp(DEFAULT_ETHADDR_2, s))) )
	{
	++error;
	sprintf(msg, "Unprogrammed default %s", s); 
	rc = E_POST_MAC_ETHADDR_IS_DEFAULT_VAL;
	}

if(!error)
	{
	if(new_mac)
		{
		setenv(env_str, mac_str);
		saveenv();
		}

	sprintf(msg, "%s programmed to: %s", dev_str, s); 
	}

return(rc);
}//end FUNCTION: post_test_mac()


/************************************************************************
*	Function: post_force_error()
*	DATE:	3-26-07
*
*	Allow the user to force a self test error, warning, info, or 
*	recover return code.  If none are present in the user entered data, 
*	force it to be an error.
************************************************************************/
s32 post_force_error(char *msg)
{
s32  error_rc = 0;
char *env_p;
char *cp;

env_p = getenv(ENV_FORCE_POST_ERROR);
if(env_p)
	{
	error_rc = (s32)simple_strtol(env_p, &cp, 16);
	if( (s32)cp < (strlen(env_p) + (s32)env_p) )
		error_rc = 0xdeadc0de | LOG_ERROR;
	else
		{
		if(!(error_rc & LOG_MASK))	/*error, warning, info, or recover?*/
			error_rc |= LOG_ERROR;	/*make it an error*/
		}
	}

sprintf(msg, "Test \"%s\" 0x%x", ENV_FORCE_POST_ERROR, error_rc);
return(error_rc);					/*no force self test error*/
}/*end FUNCTION: post_force_error()*/


/************************************************************************
*	Function: post_errorocde_history()
*	DATE:	3-26-07
*
*	Allow the user to force a self test error, and warning, info, or 
*	recover message.  If none are present in the user entered data force 
*	it to be an error.
************************************************************************/
s32 post_errorocde_history(s32 error_rc)
{
#define	MAX_HEX_STRING	15			/*<space> + 0x + 8 chars + null = 12*/

char *env_p;
char buf[MAX_ERRORCODE_HISTORY];
s32  saved_silent;				/*save quiet flag*/


env_p = getenv(ENV_ERRORCODE_HISTORY);
if(env_p)
	{
	if( strlen(env_p) > (MAX_ERRORCODE_HISTORY - MAX_HEX_STRING) )
		return(0);								/*new error will not fit*/

	sprintf(buf, "%s 0x%x", env_p, error_rc);	/*existing + new code*/
	}
else
	{
	sprintf(buf, " 0x%x", error_rc);			/*1st error code saved*/
	}


setenv(ENV_ERRORCODE_HISTORY, buf);
saved_silent = set_save_silent_mode();
saveenv();
restore_silent_mode(saved_silent);
return(0);					
}/*end FUNCTION: post_errorocde_history()*/


/****************************************************************************
*	Function: post_mode_display()
*	DATE:	4-27-07
*
*	Build a display line with various pieces of information from the FPGA 
*	MODE register.   This is useful as a trace log message.
*	This call MUST ALWAYS occur AFTER the FPGA is loaded.
*	No need to test for self test or recover error bits as this will prevent 
*	the boot anyway.
*
*	INPUT:	msg - pointer for put mode information to display
*	RETURN: 0
*****************************************************************************/
static s32 post_mode_display(char *msg)
{
char *reset_cause_p;
char *recover_mode_p;
char vendor_mode[50];
char *fwupd_p		= "";
u16 mode0 = ABS_read_fpga(FPGA_REG_MODE0);


/*	Reset cause*/
if(mode0 & MODE_RESET_THERM_TRIP)
	reset_cause_p = "HW Thermal Trip";
else if(mode0 & MODE_RESET_PON)
	reset_cause_p = "AC Power On";
else if(mode0 & MODE_RESET_BOARD)
	reset_cause_p = "Board Reset";
else if(mode0 & MODE_PEER_RESET)
	reset_cause_p = "Peer CMC";
else if(mode0 & MODE_FPGA_WDOG)
	reset_cause_p = "FPGA Watch Dog Reset";
else
	reset_cause_p = "PPC core/chip/push button";

/*	Vendor Mode possibility*/ 
vendor_mode[0] = '\0';
if(mode0 & MODE_VENDOR_HARD)
	{
	strcat(vendor_mode, "Hard");
	if(is_hard_vendor_mode())
		strcat(vendor_mode, "-header");
	if(is_fpga_vendor_mode())
		strcat(vendor_mode, "-FPGA");
	}
if(mode0 & MODE_VENDOR_SOFT)
	strcat(vendor_mode, " Soft");

if(!strlen(vendor_mode))
	strcpy(vendor_mode, "None");	

/*	Recover Mode possibility*/ 
if( (mode0 & MODE_RECOVER_HARD) && (mode0 & MODE_RECOVER_SOFT) )
	recover_mode_p = "Hard and Soft";
else if(mode0 & MODE_RECOVER_HARD)
	recover_mode_p = "Hard";
else if(mode0 & MODE_RECOVER_SOFT)
	recover_mode_p = "Soft";
else
	recover_mode_p = "None";


if(mode0 & MODE_FWUPD_REQUEST)
	fwupd_p = "FWupdate request";

sprintf(msg,	"Reset cause=%s   %s\n"
				"             Vendor mode=%s   Recover mode=%s", 
				reset_cause_p, fwupd_p, vendor_mode, recover_mode_p);

return(0);
}/*end FUNCTION: post_mode_display() */

/*****************FWUPDATE functions*****************************************/


/*****************************************************************************
*	Function: get_current_fwver()
*	DATE:	4-7-07
*
*	Get the current CMC FW verson in the env var. Cannot get it from the u-boot 
*	image as it is possible that it will not change every version.
*
*	INPUT:	ver - return version string of MAX_VERSION_STRING_SIZE sixe
*	RETURN:	non-zero - env var does not exist
*			0        - version returned in arg ver.
******************************************************************************/
s32 get_current_fwver(char *ver)
{
char *s;

if( (s = getenv(FWUPD_ENV_FW_VERSION)) == 0)
	{
	strcpy(ver, "");
	return(I_FWUPD_TRIGGER_FW_VERSION_MISSING);
	}	
else
	{
	strcpy(ver, s);
	return(0);
	}
}/*end FUNCTION: get_current_fwver() */


/*****************************************************************************
*	Function: get_image_fwver()
*	DATE:	4-7-07
*
*	Get the CMC FW verson string found in the mhdr->version_string element.
*
*	INPUT:	ver - return version string of MAX_VERSION_STRING_SIZE sixe
*	RETURN:	non-zero - Image file is bad or could not locate the version
*			0        - version returned in arg ver.
******************************************************************************/
s32 get_image_fwver(char *ver, char *msg)
{
s32 rc = 0;
s32 saved_silent = set_save_silent_mode();	/*kill any stdout*/
nand_read_options_t opts;

strcpy(ver, "");				/*error, overwrite if good*/

/*global structs containing nand info, cmc will always have 1 NAND*/
nand_info_t *nand			= &nand_info[0];
t_master_image_header *mhdr = (t_master_image_header*)CFG_LOAD_ADDR;

/*	Read master header to get version string*/
memset(&opts, 0, sizeof(opts));
opts.buffer	= (u_char*) CFG_LOAD_ADDR;
opts.length	= MASTER_IMAGE_HEADER_SIZE;
opts.offset	= FLASH_OFF_START_FWUPD_IMAGE;
opts.quiet	= 1;
if((rc = nand_read_opts(nand, &opts)) != 0)
	{
	sprintf(msg, "ERROR: nand read master header failure RC=0x%x\n", rc);
	restore_silent_mode(saved_silent);
	return(E_FLASH_GET_IMAGE_FWVER_HDR_READ);
	}

/*
*	Look for valid token and name string, (build tools do not swizzle token)
*/
restore_silent_mode(saved_silent);

if(swizzle(mhdr->token) != IMAGE_TOKEN_MASTER)
	return(E_FWUPD_NO_MASTER_TOKEN);

if( strncmp(mhdr->name, IMAGE_TEXTNAME_MASTER, MAX_IMAGE_NAME_SIZE) )
	return(E_FWUPD_NO_MASTER_TEXTNAME);

strcpy(ver, mhdr->version_string);
return(0);		
}/*end FUNCTION: get_image_fwver() */



/*****************************************************************************
*	Function: is_fwupd_needed()
*	DATE:	4-7-07
*
*	This function will test the various ways that a fwupdate can be 
*	triggered and log a informational message.
*	
*	1. An FPGA mode bit set by the OS. (normal force)
*	2. The fwupd_state=fwupd_pending (2nd time through after NOR reset or an
*	   interruption and fwupd() did not complete).
*	3. An unknown or missing current version (this would be an env corruption)
*	4. A version mismatch (same as #1 but may an AC loss at OS reboot)
*	5. A corrupt OS image
*
*	INPUT:	msg - pointer for any message to be returned
*	RETURN:	0        - FWupdate is NOT requested
*			non-zero - FWupdate is requested
*			
******************************************************************************/
s32 is_fwupd_needed(char *msg)
{
u32  need_fwupd = 0;
u32  bad_image = 0;
char cur_ver[MAX_VERSION_STRING_SIZE];
char new_ver[MAX_VERSION_STRING_SIZE];
char buf[MAX_MSG_LEN];	
char cv[MAX_VERSION_STRING_SIZE];
char nv[MAX_VERSION_STRING_SIZE];


memset(cur_ver, 0, MAX_VERSION_STRING_SIZE);
memset(new_ver, 0, MAX_VERSION_STRING_SIZE);
memset(cv, 0, MAX_VERSION_STRING_SIZE);
memset(nv, 0, MAX_VERSION_STRING_SIZE);

get_current_fwver(cur_ver);					/*read from env*/
bad_image = get_image_fwver(new_ver, msg);	/*read firmimg.cmc from NAND*/

/*
*	When moving from the 2.6.14 to the 2.6.25 kernel, the NAND driver changed 
*	the ECC algorithm.  1) Fwupd writes firmimg.cmc using old ecc   
*	2) U-boot is updated to new image and reset   3) new u-boot uses new ecc on
*	firmimg.cmc which was written with OLD ecc algorthm.  Need to make a flag 
*	for the flash driver.  Then instead of making 1.x -> 2.0 a clean updated.  
*	They moved the new kernel to 2.1 !!!
*
*	PRIOR to uboot flash
*	FW downgrade = read new alg, write old alg
*	FW upgrade   = read old alg, write new alg
*
*	g_ecc_hack = 0  -   (1.x <-> 2.0) OR  (2.1 <-> (>= 2.1))  Normal
*	g_ecc_hack = 1  -   (1.x <-> 2.0) -to- (>= 2.1)			  Going UP to new alg
*	g_ecc_hack = -1 -	(>= 2.1) -to- (2.0 <-> 1.x)			  Going DOWN to old alg
*/

g_ecc_hack = 0;		/*default: fwupd is NOT crossing the 2.1 barrier, normal operation*/
if(bad_image == 0)
	{
	/*	
	*	A bug made 1.0 into A02, and 1.01 into A05!!!  Cannot simply cmp the 
	*	first byte.  Going forward this numbering CANNOT be allowed.  
	*	So if 'A' then it MUST be 1.x
	*/
	strncpy(cv, cur_ver, 3);
	strncpy(nv, new_ver, 3);

	if( (cv[0] < 0x30) || (cv[0] > 0x39) )	/*convert A0X to 1.0*/
		strcpy(cv, "1.0");
	if( (nv[0] < 0x30) || (nv[0] > 0x39) )	
		strcpy(nv, "1.0");

	if( (strcmp(cv, "2.0") > 0) && (strcmp(nv, "2.1") < 0) )
		g_ecc_hack = -1;	/*moving from >= 2.1 to < 2.1*/
	}
else
	{
	g_ecc_hack = 1;			/*nand read error always force a bad_image flag*/
	}


if( is_fwupd_mode_set() )
	{
	need_fwupd = I_FWUPD_TRIGGER_MODE_DETECTED;
	sprintf(msg, "FWupdate request initiated by OS.");
	}
else if( is_fwupd_env_pending() )
	{
	/*fwupd has not yet run to completion, pass or fail*/
	need_fwupd = I_FWUPD_TRIGGER_STATE_PENDING;
	sprintf(msg, "FWupdate request from previous pending update.");
	}
else if( strlen(cur_ver) == 0 )
	{
	need_fwupd = I_FWUPD_TRIGGER_FW_VERSION_MISSING;
	sprintf(msg, "FWupdate triggered by missing fwupd env.");
	}
	/*	Verify the CRC of the current OS images*/
else if( post_verify_os_images(msg) )
	{
	need_fwupd = I_FWUPD_TRIGGER_BAD_OS_IMAGE;
	/*	post_verify_os_images() has loaded msg*/
	}
else if( strcmp(cur_ver, new_ver) )
	{
	need_fwupd = I_FWUPD_TRIGGER_FW_VERSION_DIFFERENT;
	sprintf(msg, "FWupdate triggered by version mismatch.");
	}
/*
*	This contidion should remain last as the previous triggers require the 
*	firmimg.cmc file to already be in the NAND.  This condition indicates that
*	the image will need to be tftp loaded from the peer.
*/
else if( is_fwupd_req_from_peer() )
	{
	need_fwupd = I_FWUPD_TRIGGER_FROM_PEER;
	sprintf(msg, "FWupdate request, load image from peer.");
	}


if(need_fwupd)
	{
	sprintf(buf,"\n             CMC FW version current=\"%s\" new=\"%s\"",
															cur_ver, new_ver);
	strcat(msg, buf);


	if(1 == g_ecc_hack)
		strcat(msg, "\n             Appling new ECC algorithm");
	if(-1 == g_ecc_hack)
		strcat(msg, "\n             Appling old ECC algorithm");
	}
else
	{
	sprintf(msg,"CMC FW version=\"%s\" ", new_ver);
	}

if(need_fwupd == 0)
	g_ecc_hack = 0;

return(need_fwupd);
}/*end FUNCTION: is_fwupd_needed() */



/*****************************************************************************/


/************************************************************************
*	Function: post_id_usb()
*	DATE:	2-14-07
*
*	Initilize the USB the verify that the "Flash Card Reader" is found.
************************************************************************/
s32 post_id_usb(char *msg)
{
u32 i;
s32	rc;
s32 saved_silent;				/*save quiet flag*/
struct usb_device *dev;

/*	
*	If the USB crystal is bad the CMC locks up.  Put out a message
*	indicating this will help in diags, it will be overwritten when it passes.
*
*	In short we are running the "usb reset" command to make sure we can find 
*	the Flash Card Reader device.  This will cause various messages to be
*	displayed, and messes up the POSt lessages.  Put us in quiet mode.
*/
printf("**USB scan hangs on a crystal fault**\r");

saved_silent = set_save_silent_mode();

/*	Stop the USB and initialize it*/
usb_stop();
if ( (rc = usb_init() ) < 0)
	{
	restore_silent_mode(saved_silent);
	sprintf(msg, "USB failed to init, rc=0x%08lx         ", rc);
	return(E_POST_USB_INIT);
	}

/*	usb_init() passed, now scan storage devices*/
usb_stor_scan(1);
dev = usb_get_dev_index(1);	/*PPC host is 0, FCR is 1 or 2nd dev on list*/
if(dev == NULL)
	{
	restore_silent_mode(saved_silent);
	sprintf(msg, "USB Flash Card Reader cannot be found  ");
	return(E_POST_USB_FLASHCARDREADER_NOT_FOUND);
	}

/*	See if we have a descriptor that we understand*/
if (dev->descriptor.bDescriptorType != USB_DT_DEVICE) 
	{
	restore_silent_mode(saved_silent);
	sprintf(msg, "USB Flash Card Reader invalid descriptor");
	return(E_POST_USB_FLASHCARDREADER_DESCRIPTOR);
	}

/*	See if the FCR is a manufacturer and product that we know about*/
for(i = 0; hw_inv_fcr[i].idvendor != 0; ++i)
	{
	if( (dev->descriptor.idVendor  == hw_inv_fcr[i].idvendor) &&
		(dev->descriptor.idProduct == hw_inv_fcr[i].idproduct) )
		{
		restore_silent_mode(saved_silent);
		sprintf(msg, "ID \"%s\" Flash Card Reader, product ID 0x%04lx",
						hw_inv_fcr[i].manuf, hw_inv_fcr[i].idproduct);
		return(0);	/*found it get out*/
		}
	}

/*	Unknown Manufacturer*/
restore_silent_mode(saved_silent);
sprintf(msg, "Unknown Flash Card Reader, Vendor-0x%04lx product-0x%04lx",
						dev->descriptor.idVendor, dev->descriptor.idProduct);
return(W_POST_USB_UNKNOWN_FCR_VENDOR);
}/*end FUNCTION: post_id_usb()*/



s32 post_fpga_load(u32 flag, char *msg)
{
return( fpga_preload(flag, msg) );
}/*end FUNCTION: post_fpga_load()*/



s32 post_id_bmc5221(char *msg)
{
strcpy(msg, "Invalid BCM5221 PHY ID");
return(0x80000000);		/*hack*/
}/*end FUNCTION: post_id_bmc5221()*/


s32 post_id_bmc5324(char *msg)
{
strcpy(msg, "Invalid BCM5324 switch ID");
return(0x80000000);		/*hack*/
}/*end FUNCTION: post_id_bmc5324()*/



/************************************************************************
*	Function: get_cmc_fru()
*	DATE:	5-11-07
*
*	At the board vendor the two CMC MAC addresses are programmed into the FRU
*	at ICT.  Upon the first power up post_test_mac() will detect the ethadd 
*	and eth1addr do not exist and call this function to get the MAC's from 
*	the FRU.
*
*	INPUT:	fru_field - 0 for MAC0 FRU data, 1 for MAC1 FRU data
*			field_string - pointer to place MAC ":" seperated string
*
*	RETURN:	0 - MAC data was read from FRU
*			non-0 - Failure reading FRU
************************************************************************/
static s32 get_cmc_fru(u32 fru_field, char * field_string, char *msg)
{ 
  char fru[CMCFRUSIZE-CMC_HEADER_LEN];
  char *fru_ptr;
  char *eeprom_ptr;
  char str[6];
  char *p;
  u32   addr;
  u8 internal_usr_area_offset;
  u16 mac_offset;
  s32 rc, i;
  u8  check_sum = 0;

  fru_ptr = &fru[0];
  eeprom_ptr = fru_ptr;
  p = str;
  
  /*Read 16 bytes at a time instead of the whole 256 bytes*/
  for(addr = CMC_HEADER_LEN; addr < CMCFRUSIZE; addr +=16,fru_ptr +=16)
  {
    rc = i2c_read(CHIP_ADDR_FRU, addr, 1, fru_ptr, 16);
    if (rc == 1)
    {
      sprintf(msg,"Cannot read FRU: I2C read failed");
      return(1);
    }
  }

  for(i=0;i<8;i++)
  {
    check_sum += (u8)fru[i];
  }
  if (check_sum != 0) 
  {
    sprintf(msg,"Common Header Checksum fail Byte7=0x%02X",fru[7]);
    return(1);
  }
    
    internal_usr_area_offset = 8 * (*(eeprom_ptr+1));
    eeprom_ptr = eeprom_ptr + internal_usr_area_offset + BMC_MAC_OFFSET;
    mac_offset = *(eeprom_ptr + 1) << 8;
    mac_offset = mac_offset + *eeprom_ptr;
    eeprom_ptr = &fru[mac_offset];
    check_sum = 0;
    for(i = mac_offset; i< (mac_offset+16); i++)    // 16 Bytes in element 3
    {
      check_sum += fru[i];
    }
    if(check_sum != 0)
    {
      sprintf(msg,"MAC area Checksum failed");
      return(1);
    }
    
    if(fru_field == FRU_GET_MAC0)
    {
      eeprom_ptr = eeprom_ptr + 4;
      for(i = 0; i< 6; i++)
      {
       *p++ = *eeprom_ptr++;
      }
    }
    else if(fru_field == FRU_GET_MAC1)
    {
      eeprom_ptr = eeprom_ptr + 10;
      for(i = 0; i< 6; i++)
      {
        *p++ = *eeprom_ptr++;
      }
    }
    else
    {
      sprintf(msg,"Invalid MAC request");
      return(1);
    }
    sprintf(field_string, "%02X:%02X:%02X:%02X:%02X:%02X", 
							str[0], str[1], str[2], str[3],str[4],str[5]);
          
  return(0);
}/*end FUNCTION: get_cmc_fru()*/






/***************************************************************************
*	Function: blink_debug_led()
*	DATE:	5-09-07
*
*	When in recover mode we want to flash the debug LED at ~1/2 second rate
*	and a self test error at ~1/8 second rate.
*	NOTE: this function is called from the timer_interrupt() thread in 
*	../lib_ppc/interrupt.c, be quick about it.
*
*	INPUT:	current timestamp in milliseconds
*	RETURN:	0 
*
****************************************************************************/
s32 blink_debug_led(u32 timestamp)
{
u32 blink_mask;
u32 tmp;

if(g_post_rc)
	{
    if(LOG_RECOVER & g_post_rc)
		blink_mask = BLINK_RATE_MASK_RECOVER;
	else if(LOG_ERROR & g_post_rc)
		blink_mask = BLINK_RATE_MASK_SELFTEST;
	else
		return(0);				/*only blink on recover, selftest*/

	if( !(timestamp & blink_mask) )	
		{
		tmp = in32(GPIO0_OR);
		if(tmp & GPIO_DEBUG_LED)
			out32(GPIO0_OR, tmp & ~GPIO_DEBUG_LED);
		else
			out32(GPIO0_OR, tmp | GPIO_DEBUG_LED);
		}
	}

return(0);
}/*end FUNCTION: blink_debug_led() */



/***************************************************************************
*	Function: post_verify_rtc()
*	DATE:	5-12-07
*
*	We cannot go into an extensive test for the RTC.  We are simply going 
*	to make sure that it is ticking.  Do a read here and at the end of POST 
*	and make sure it is different.
*   
*	INPUT:	rtc_ticks - A 2 element array with 2 RTC readings.
*			msg - pointer to hold returned message
*	RETURN:	0 - rtc passed
*			non-zero - rtc failed
****************************************************************************/
static s32 post_verify_rtc(u32 rtc_ticks[], char *msg)
{

if(rtc_ticks[0] == rtc_ticks[1])
	{
	udelay(1000000);
	if(get_rtc_tod(&rtc_ticks[1]) != 0)
		rtc_ticks[1] = -1;
	}


if( (rtc_ticks[0] == -1) || (rtc_ticks[1] == -1) )
	{
	sprintf(msg,"Cannot read RTC Time-of-day counter");
	return(E_POST_RTC_READ_FAILURE);	
	}


if(rtc_ticks[0] == rtc_ticks[1])
	{
	sprintf(msg,"RTC is not incrementing, tod=0x%08lx", rtc_ticks[0]);
	return(E_POST_RTC_NOT_COUNTING);	
	}

sprintf(msg,"RTC counter Ticking...0x%08lx:0x%08lx", rtc_ticks[0], rtc_ticks[1]);
return(0);
}/*end FUNCTION: post_verify_rtc*/

/*********************************************************************
*  FUNCTION: post_id_phy()
*  DATE:     05-24-07
*
*  Read the ID of the PHY chips on the CMC.
*  INPUT:	phy - which LAN device to ID.
*           msg - pointer to place error or success message.
*  RETURN:  0 - The device was ID'ed correctly
*           non-0 - Failed to reasd device OR, invalid ID
*********************************************************************/
s32 post_id_phy(u32 phy, char * msg)
{
    unsigned short val = 0;
    u32 micrel;
    char *name;
    unsigned int tmp;
    int addr;
        
    switch(phy)
    {
        case 1:
        {
            micrel = get_micrel_en();
            set_micrel_en(1);
            name = miiphy_get_current_dev();
            if (miiphy_read(name, 3, PHY_PHYIDR1, &val) != 0) //3 is the address
            {
                sprintf(msg,"Cannot ID the PHY device");
                set_micrel_en(micrel);
                return(W_POST_MII_PHY_READ_FAIL);
            }
            tmp = val << 16;
            if (miiphy_read(name, 3, PHY_PHYIDR2, &val) != 0)
            {
                sprintf(msg,"Cannot ID the PHY device");
                set_micrel_en(micrel);
                return(W_POST_MII_PHY_READ_FAIL);
            }
            tmp |= val;
            set_micrel_en(micrel);
            /* Do not print a WARN or FAIL message in case of a mismatch in ID,
             This is to cope with Rev changes in the PHY chips. */
            
            if((bcm5324_phy_def.phy_id_mask & tmp) == bcm5324_phy_def.phy_id)
            {    
                sprintf(msg,"Found %s ,ID = 0x%x",bcm5324_phy_def.vendor_name,tmp);
                return(0);
            }
            else
            {
				sprintf(msg,"Found %s ,ID = 0x%x",bcm5324_phy_def.vendor_name,tmp);
				return(0);
				/*
                sprintf(msg,"Found %s ,ID = 0x%x Expecting ID = 0x%x",
							bcm5324_phy_def.vendor_name,tmp,bcm5324_phy_def.phy_id);
                return(W_PHY_UNKNOWN_ID);
				*/
            }
            break;
        }
        case 2:
        {
            micrel = get_micrel_en();
            set_micrel_en(1);
            name = miiphy_get_current_dev();
            if (miiphy_read(name, 25, PHY_PHYIDR1, &val) != 0) //3 is the address
            {
                sprintf(msg,"Cannot ID the PHY device");
                set_micrel_en(micrel);
                return(W_POST_MII_PHY_READ_FAIL);
            }

            tmp = val << 16;

            /* LED4=off, LED3=off */
            miiphy_write(name, 25, 0x1c, 0xb8ee);
            miiphy_write(name, 25, PHY_BMCR, PHY_BMCR_ISO);

            if (miiphy_read(name, 25, PHY_PHYIDR2, &val) != 0)
            {
                sprintf(msg,"Cannot ID the PHY device");
                set_micrel_en(micrel);
                return(W_POST_MII_PHY_READ_FAIL);
            }
            tmp |= val;
            set_micrel_en(micrel);
            if((bcm5481_phy_def.phy_id_mask & tmp) == bcm5481_phy_def.phy_id)
            {
                sprintf(msg,"Found %s ,ID = 0x%x",bcm5481_phy_def.vendor_name,tmp);
                return(0);
            }
            else
            {
				sprintf(msg,"Found %s ,ID = 0x%x",bcm5481_phy_def.vendor_name,tmp);
				return(0);
				/*
                sprintf(msg,"Found %s ,ID = 0x%x Expecting ID = 0x%x",
							bcm5481_phy_def.vendor_name,tmp,bcm5481_phy_def.phy_id);
                return(W_PHY_UNKNOWN_ID);
				*/
            }
            break;
        }
        case 3:
        {
            micrel = get_micrel_en();
            set_micrel_en(1);
            name = miiphy_get_current_dev();
            if (miiphy_read(name, 26, PHY_PHYIDR1, &val) != 0) //3 is the address
            {
                sprintf(msg,"Cannot ID the PHY device");
                set_micrel_en(micrel);
                return(W_POST_MII_PHY_READ_FAIL);
            }
            tmp = val << 16;
            if (miiphy_read(name, 26, PHY_PHYIDR2, &val) != 0)
            {
                sprintf(msg,"Cannot ID the PHY device");
                set_micrel_en(micrel);
                return(W_POST_MII_PHY_READ_FAIL);
            }
            tmp |= val;

	    /* force PHYs */
	    for (addr = 0; addr <= 23; addr++) {
		switch (addr) {
		/* blades */
		default:
		    /* 10/HDX */
		    miiphy_write(name, addr, 0, 0x0040);
		    break;
		/* IOM, LKVM */
		case 0: case 3: case 6: case 10: case 13: case 18: case 21: case 22:
		    /* 100/FDX */
		    miiphy_write(name, addr, 0, 0x2100);
		    break;
		}
	    }

            /* LED4=activity, LED3=on */
            miiphy_write(name, 26, 0x1c, 0x8801);
            miiphy_write(name, 26, 0x1c, 0xb853);

            set_micrel_en(micrel);
            if((bcm5481_phy_def.phy_id_mask & tmp) == bcm5481_phy_def.phy_id)
            {
                 sprintf(msg,"Found %s ,ID = 0x%x",bcm5481_phy_def.vendor_name,tmp);
                 return(0);
            }
            else
            {
				sprintf(msg,"Found %s ,ID = 0x%x",bcm5481_phy_def.vendor_name,tmp);
				return(0);
				/*
                sprintf(msg,"Found %s ,ID = 0x%x Expecting ID = 0x%x",
							bcm5481_phy_def.vendor_name,tmp,bcm5481_phy_def.phy_id);
                return(W_PHY_UNKNOWN_ID);
				*/
            }
            break;
        }
        case 4:
        {
            micrel = get_micrel_en();
            set_micrel_en(0);
            name = miiphy_get_current_dev();
            if (miiphy_read(name, 0, PHY_PHYIDR1, &val) != 0) //3 is the address
            {
                sprintf(msg,"Cannot ID the PHY device");
                set_micrel_en(micrel);
                return(W_POST_MII_PHY_READ_FAIL);
            }
            tmp = val << 16;
            if (miiphy_read(name, 0, PHY_PHYIDR2, &val) != 0)
            {
                sprintf(msg,"Cannot ID the PHY device");
                set_micrel_en(micrel);
                return(W_POST_MII_PHY_READ_FAIL);
            }
            tmp |= val;
            set_micrel_en(micrel);
            if((micrel_phy_def.phy_id_mask & tmp) == micrel_phy_def.phy_id)
            {
                 sprintf(msg,"Found %s ,ID = 0x%x",micrel_phy_def.vendor_name,tmp);
                 return(0);
            }
            else
            {
				sprintf(msg,"Found %s ,ID = 0x%x",micrel_phy_def.vendor_name,tmp);
				return(0);
				/*
                 sprintf(msg,"Found %s ,ID = 0x%x Expecting,ID = 0x%x",
								micrel_phy_def.vendor_name,tmp,micrel_phy_def.phy_id);
                 return(W_PHY_UNKNOWN_ID);
                */
            }
            break;
        }
        default:
            sprintf(msg,"No PHY selected");
            return(0);
            break;
    }
  }

#endif /*CONFIG_CMC_10G*/

