/****************************************************************************

  Copyright (c) 2006, Dell Inc
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

  - Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
    
  - Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution. 

  - Neither the name of Dell Inc nor the names of its contributors
    may be used to endorse or promote products derived from this software 
    without specific prior written permission. 

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. 

*****************************************************************************/

#ifdef CONFIG_CMC_10G	/*only compile in for a CMC_10G build*/


/******************************************************************************
*
*	FILE: util_ethernet.c
*	Use this file to quickly implimit a CMC utility command.
*
*	1) cp  u-boot/common/util_template.c  u-boot/common/util_<newcmd>.c
*	2) Change/modify options as necessary.
*	3) Add command entry to struct "util_cmd_table" contained in cmd_util.c
*	   do not forget to also add   add extern t_usage util_<newcmd>_usage;
*	4) Add "util_<newcmd>.o to   u-boot/common/Makefile
*
******************************************************************************/


/******************************************************************************
*  Include files
******************************************************************************/
#include <configs/cmc_10g.h>
#include <common.h>
#include <command.h>
#include <config.h>
#include <util_defs.h>
#include <util_ret_codes.h>
#include <ppc4xx_enet.h>
#include <miiphy.h>

/******************************************************************************
*  Local Defines
******************************************************************************/


/******************************************************************************
*  Globals
******************************************************************************/


/******************************************************************************
*  Prototypes/Externs
******************************************************************************/
extern s32	mcoptind;
extern char	*mcoptarg[];


/******************************************************************************
*	List all structures defines
******************************************************************************/
t_options util_ethernet_options[] = 
	{
	/*Options*/
	{"i",	"resets x24 switch"},
	{"r",	"resets x24 switch"},
	{"m",	"mux select"},
	{"init","Init 24 port switch"},
	{ NULL, NULL }
	};


t_usage util_ethernet_usage = { 
	"eth", 
					
	/*Syntax*/
	"eth  [-i] [-r] [-m] [-init]\n",

	/*Description*/
	"Ethernet example for adding a new util command.\nExamples:\n\n"
	"util eth <add here>\n", 
	
	util_ethernet_options
	}; /*end t_usage util_ethernet_usage*/


/******************************************************************************
*
*	FUNCTION: util_ethernet()
*	DATE:	
*
******************************************************************************/
s32 util_ethernet(int argc, char *argv[])
{
u32  index = 0, rc = 0;
char msg[MAX_MSG_LEN];		/*message buffer*/
//char addr_str[40];			/*save the address range*/
//struct eth_device *dev;
int reset_x24_flag = 0;	  /* Default to not reset the x24 Broadcom switch. */
int select_mdio_mux_flag = 0;	/* Default to leave the MDIO mux alone. */
int init_mdio_clock_flag = 0;	/* Default to leave the MDIO clock alone. Don't initialize it.*/
//int *gpio1_controller_ptr; /* address used to toggle MDIO mux select gpio.  Bit 31 selects*/
//int select_broadcom = 0x1; /* Bit 31 = 1 selects Broadcom, Bite 31 = 0 selects Micrel*/
//int reset_x24_mux = 0x00001000; /* Bit 19 high not reset, Bit 19 low means in reset*/
//int reset_x24_mux_pulse = 51000; /* at least 50 msec pulse */
/* x24 reset pulse must last at least 50 milliseconds as per hardware designer for CMC */

//addr_str[0]	= '\0';				/*null the string*/
//gpio1_controller_ptr = (int *)0xef60c00; /* address used to toggle MDIO mux select gpio.  Bit 31 selects*/
/* gpio_controller_ptr also controls MDIO mux selection */

int		i;
int		or1;
char	*devname;

//	Get any options
mcoptind = 2;	
while( (index = mcgetopt(argc, argv, "-d -s")) != MCGETOPT_EOF) {
	switch(index) {
	case 0 :
	{
		sys_info_t sysinfo;

		miiphy_listdev();

		get_sys_info (&sysinfo);
		printf("freqOPB=%d\n", sysinfo.freqOPB);
		
		printf("RMMII_FER=0x%x\n", in32(RGMII_FER));
		printf("ZMII_FER=0x%x\n", in32(ZMII_FER));
		printf("ZMII_SSR=0x%x\n", in32(ZMII_SSR));
		printf("EMAC_M1=0x%x\n", in32(EMAC_M1));
		printf("EMAC_M1+0x100=0x%x\n", in32(EMAC_M1+0x100));
		printf("GPIO0_OR=0x%x\n", in32(GPIO0_OR));
		printf("GPIO1_OR=0x%x\n", in32(GPIO1_OR));
		printf("EMAC_STACR=0x%x\n", in32(EMAC_STACR));
		break;
	}
	case 1: /* switch interfaces */
		eth_try_another(0);
		printf("Using %s device\n", getenv("ethact"));
		break;

	default:		/*not good*/
		rc = mcgetopt_error(index, msg, mcoptarg[0]);
		printf("%s %s:%s\n", argv[0], argv[1], msg);
		return(rc);
		break;
	}
}/*end mcgetopt while*/



return(0);
}/* end FUNCTION: util_ethernet()*/


#endif /*CONFIG_CMC_10G*/



