/****************************************************************************

  Copyright (c) 2006, Dell Inc
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

  - Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
    
  - Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution. 

  - Neither the name of Dell Inc nor the names of its contributors
    may be used to endorse or promote products derived from this software 
    without specific prior written permission. 

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. 

*****************************************************************************/



#include <common.h>		/*must be first in file*/
#include <command.h>
#include <asm/processor.h>
#include "../../../cpu/ppc4xx/vecnum.h"
#include "sdram.h"

#ifdef CONFIG_CMC_10G	/*only compile in for a CMC_10G build*/


/******************************************************************************
*
*	FILE: util_mem.c
*	ECC development
*
******************************************************************************/


/******************************************************************************
*  Include files
******************************************************************************/
#include <util_defs.h>
#include <util_ret_codes.h>


/******************************************************************************
*  Local Defines
******************************************************************************/
#define MAX_ECC_BIT				128	/*0-127,128 being a multi uncorrectable*/
#define POST_ECC_BIT_START		64  /*bits 0-63 do not seem to work*/

/******************************************************************************
*  Globals
******************************************************************************/
/*	
*	This structure is filled in by the ECC interrupt routine with register
*	values associated with ECC status that can be processed at a later time.
*/
typedef struct	{	
	u32	status;		/*DDR0_00 holds INTACK and INTSTATUS*/
	u32 synd;		/*DDR0_23 syndrom bytes for ECC event, forced write check*/
	u32 u_addr;		/*DDR0_34 address of uncorrectable event [31:0]*/
	u32 u_data;		/*DDR0_36 data of uncorrectable event [31:0]*/
	u32 c_addr;		/*DDR0_38 address of correctable event [31:0]*/
	u32 c_data31_0;	/*DDR0_40 data of correctable event [31:0]*/
	u32 c_data63_32;/*DDR0_41 data of correctable event [63:32]*/
	} t_ecc_data;

#define MAX_ECC_INT_EVENTS	10
t_ecc_data  g_ecc[MAX_ECC_INT_EVENTS];

u32 g_ecc_index;		/*index into g_ecc struct*/
u32 g_total_ecc_ints;	/*total number of ECC interrupts*/

/*	
*	ECC syndrom codes to force errors.  Index into the array it the bit to 
*	force 0-127.   To force a double bit error use bit 128, or last entry.
*	I had problems with this.  Also see ecc_force_error(), I could not force
*	errors on bits 0-63!!!  See register DDR0_31.  The table on "ECC Syndrome
*	codes" only goes from bit 0-63, I assumed bit 0-63 are the same codes as
*	bits 64-128, but a different position in DDR0_31. See "Parameter 
*	XOR_CHECK_BITS mapping" in the users manual. 
*/
u16 ecc_synd_codes[] = 
	{
	 0x00f4, 0x00f1, 0x00ec, 0x00ea, 0x00e9, 0x00e6, 0x00e5, 0x00e3,	/*00-63*/
	 0x00dc, 0x00da, 0x00d9, 0x00d6, 0x00d5, 0x00d3, 0x00ce, 0x00cb,	
	 0x00b5, 0x00b0, 0x00ad, 0x00ab, 0x00a8, 0x00a7, 0x00a4, 0x00a2,
	 0x009d, 0x009b, 0x0098, 0x0097, 0x0094, 0x0092, 0x008f, 0x008a, 
	 0x0075, 0x0070, 0x006d, 0x006b, 0x0068, 0x0067, 0x0064, 0x0062,
	 0x005e, 0x005b, 0x0058, 0x0057, 0x0054, 0x0052, 0x004f, 0x004a, 
	 0x0034, 0x0031, 0x002c, 0x002a, 0x0029, 0x0026, 0x0025, 0x0023,
	 0x001c, 0x001a, 0x0019, 0x0016, 0x0015, 0x0013, 0x000e, 0x000b, 

	 0xf400, 0xf100, 0xec00, 0xea00, 0xe900, 0xe600, 0xe500, 0xe300,	/*64-127*/
	 0xdc00, 0xda00, 0xd900, 0xd600, 0xd500, 0xd300, 0xce00, 0xcb00, 
	 0xb500, 0xb000, 0xad00, 0xab00, 0xa800, 0xa700, 0xa400, 0xa200,
	 0x9d00, 0x9b00, 0x9800, 0x9700, 0x9400, 0x9200, 0x8f00, 0x8a00, 
	 0x7500, 0x7000, 0x6d00, 0x6b00, 0x6800, 0x6700, 0x6400, 0x6200,
	 0x5e00, 0x5b00, 0x5800, 0x5700, 0x5400, 0x5200, 0x4f00, 0x4a00, 
	 0x3400, 0x3100, 0x2c00, 0x2a00, 0x2900, 0x2600, 0x2500, 0x2300,
	 0x1c00, 0x1a00, 0x1900, 0x1600, 0x1500, 0x1300, 0x0e00, 0x0b00, 
	 0x0300 }; /*force multi-bit errors*/


/******************************************************************************
*  Prototypes/Externs
******************************************************************************/
extern s32	mcoptind;
extern char	*mcoptarg[];

extern void uic2_interrupt( void * parms);

static s32 ecc_bit2offset(u32 bit, u32 *offset, u32 *err_data_bit);


/******************************************************************************
*	List all structures defines
******************************************************************************/
t_options util_mem_options[] = 
	{
	/*Options*/
	{"force","<0xaddr> bit to force <0-127>"},
	{"read", "<0xaddr> read 4 words of data"},
	{"ecc",  "<dump | init | test | off>"},
	{"verbose","display tracking messages"},
	{"debug","display detailed debug details"},
	{"a",    "address range for -ecc test <start:end>|<start#cnt>"},

	{ NULL, NULL }
	};


t_usage util_mem_usage = { 
	"mem", 
					
	/*Syntax*/
	"mem [-ecc init]\n"
	"mem [-ecc off]\n"
	"mem [-ecc dump]\n"
	"mem [-ecc test] [-a <start:end | start#cnt>] [-verbose] [-debug]\n"
	"mem [-force 0xaddr bit]\n"
	"\n",

	/*Description*/
	"This util command will allow the user to init, disable, dump, and test\n"
	"the ECC logic on the CMC.\n\n", 
	
	util_mem_options
	}; /*end t_usage util_mem_usage*/


/******************************************************************************
*
*	FUNCTION: util_mem()
*	DATE:	
*
******************************************************************************/
s32 util_mem(int argc, char *argv[])
{
u32  index = 0, rc = 0;
u32  faddr, fbit;
u32  init_flag = 0, dump_flag = 0, clear_flag = 0, test_flag = 0;
u32  debug_flag = 0, verbo_flag = 0, flag = 0, off_flag = 0;
char msg[MAX_MSG_LEN];		/*message buffer*/
char *cptr1;
u32  i;
u32  saved_ddr20_22;
char addr_str[40];			/*save the address range*/

u32  saddr = CFG_LOAD_ADDR;				/*default if no -a option*/
u32  eaddr = CFG_MEMTEST_END_RAM_EXE;
u32  bcnt  = eaddr - saddr;	
volatile u32  tmp;

addr_str[0]	= '\0';				/*null the string*/

//	Get any options
mcoptind = 2;	
while( (index = mcgetopt(argc, argv, 
	  "-force:2 -read:1 -ecc:1 -verbose -debug -a:1 ")) != MCGETOPT_EOF)
	{
	switch(index) 
		{
		case 0:			/*-force ecc error to a single memory location*/
			faddr = (u32)simple_strtol( mcoptarg[0], &cptr1, 0 );
			if( (s32)cptr1 < (strlen(mcoptarg[0]) + (s32)mcoptarg[0]) ) 
				{	
				printf("%s %s:ERROR: converting \"%s\" int.\n", 
											argv[0], argv[1], mcoptarg[0]);
				return(E_OPTARG_INT_CONVERSION);
				}

			fbit = (u32)simple_strtol( mcoptarg[1], &cptr1, 0 );
			if( (s32)cptr1 < (strlen(mcoptarg[1]) + (s32)mcoptarg[1]) ) 
				{	
				printf("%s %s:ERROR: converting \"%s\" int.\n", 
											argv[0], argv[1], mcoptarg[1]);
				return(E_OPTARG_INT_CONVERSION);
				}

			if( (fbit < 0) || (fbit > MAX_ECC_BIT) )
				{	
				printf("%s %s:ERROR: Bit to force must be between 0 - 127 \n", 
															argv[0], argv[1]);
				return(E_OPTARG_INT_CONVERSION);
				}

			ecc_force_error(faddr, fbit);
			tmp = *(u32*)faddr;					/*read to force error*/
			return(0);
			break;

		case 1:			/*-read 4 words*/
			faddr = (u32)simple_strtol( mcoptarg[0], &cptr1, 0 );
			if( (s32)cptr1 < (strlen(mcoptarg[0]) + (s32)mcoptarg[0]) ) 
				{	
				printf("%s %s:ERROR: converting \"%s\" int.\n", 
											argv[0], argv[1], mcoptarg[0]);
				return(E_OPTARG_INT_CONVERSION);
				}

			printf("Read@ 0x%08lx = 0x%08lx 0x%08lx 0x%08lx 0x%08lx\n", 
				faddr, *(u32*)faddr, *(u32*)(faddr+4), *(u32*)(faddr+8), *(u32*)(faddr+12));
			break;

		case 2:			/*-ecc*/
			if(!strcmp(mcoptarg[0], "init"))
				init_flag = 1;
			else if(!strcmp(mcoptarg[0], "dump"))
				dump_flag = 1;
			else if(!strcmp(mcoptarg[0], "test"))
				test_flag = 1;
			else if(!strcmp(mcoptarg[0], "off"))
				off_flag = 1;
			else
				{
				sprintf(msg, "Invalid -ecc arg of \"%s\"\n", mcoptarg[0]);
				printf("%s %s:ERROR:%s\n", argv[0], argv[1], msg);
				return(E_GETOPT_INVALID_OPT_ARG);
				}
			break;

		case 3:			/*-verbose*/
			flag |= VERBOSE_ON;
			break;

		case 4:			/*-debug*/
			flag |= ECC_DEBUG_ON;
			break;

		case 5 :		/*option "a" address range*/
			strcpy(addr_str, mcoptarg[0]);

			if( (rc = parse_addr(addr_str, &saddr, &eaddr, &bcnt, WIDTH_8, msg)) != 0)
				{
				printf("%s %s:ERROR:RC:0x%x%s\n", argv[0], argv[1], rc, msg);
				return(rc);
				}

			printf("Start address=0x%x End address=0x%x Bytecount=%d\n", 
					saddr, eaddr, bcnt);
			break;

		default:		/*not good*/
			rc = mcgetopt_error(index, msg, mcoptarg[0]);
			printf("%s %s:%s\n", argv[0], argv[1], msg);
			return(rc);
			break;
		}
	}/*end mcgetopt while*/


/****************************************************************************
*	-ecc init
****************************************************************************/
if(init_flag)
	post_ecc_init(msg);

/****************************************************************************
*	-ecc dump
*	Display the contents of the global t_ecc_data struct
****************************************************************************/
if(dump_flag)
	{
	printf("Table g_ecc starts at 0x%x size=0x%x\n", 
		&g_ecc[0].status, sizeof(t_ecc_data) * MAX_ECC_INT_EVENTS);
	printf("Current ECC interrupt data table Total ints=%d index=%d\n", 
													g_total_ecc_ints, g_ecc_index); 
			
	for(i = 0; i < g_ecc_index; ++i)
		{
		printf("%2d: status=0x%08lx  synd=  0x%08lx  u_addr=0x%08lx u_data=0x%08lx\n"
			   "    c_addr=0x%08lx  c_data31_0=0x%08lx  c_data63_32=0x%08lx\n",
				i, g_ecc[i].status, g_ecc[i].synd, g_ecc[i].u_addr, g_ecc[i].u_data, 
				g_ecc[i].c_addr, g_ecc[i].c_data31_0, g_ecc[i].c_data63_32);
		}
	}


/****************************************************************************
*	-ecc test
****************************************************************************/
if(test_flag)
	{
	if(( rc = post_ecc_test(saddr, bcnt, flag, msg)) != 0)
		{
		printf("%s %s:ERROR:%s\n", argv[0], argv[1], msg);
		return(rc);
		}
	}

/****************************************************************************
*	-ecc off
****************************************************************************/
if(off_flag)
	{
	mfsdram(DDR0_22, saved_ddr20_22);	
	mtsdram(DDR0_22, (~DDR0_22_CTRL_RAW_MASK & saved_ddr20_22)); 
	printf("Disabled ECC\n");
	}

return(0);
}/* end FUNCTION: util_mem()*/




  
/*****************************************************************************
*
*	FUNCTION: post_ecc_int_handler()
*	DATE:	6-18-07
*
*	This is the interrupt handler for ECC interrupts.  Simple save necessary 
*	parameters so the upper layer code can figure out what is going on. 
*	I had problems using the global data areas while disabling/enabling ECC.
*	So...save then into stack/registers, then move them into the globals.
*
*****************************************************************************/
s32 post_ecc_int_handler(void)
{
u32 saved_ctrl;
u32 saved_status;
u32 status, synd, u_addr, u_data, c_addr, c_data31_0, c_data63_32;

++g_total_ecc_ints;
mfsdram(DDR0_22, saved_ctrl);	
mfsdram(DDR0_00, saved_status);
mtsdram(DDR0_22, (~DDR0_22_CTRL_RAW_MASK & saved_ctrl)); /*prevent ecc events*/

if(g_ecc_index < MAX_ECC_INT_EVENTS)
	{
	mfsdram(DDR0_00, status);
	mfsdram(DDR0_23, synd);
	mfsdram(DDR0_34, u_addr);
	mfsdram(DDR0_36, u_data);
	mfsdram(DDR0_38, c_addr);
	mfsdram(DDR0_40, c_data31_0);
	mfsdram(DDR0_41, c_data63_32);
	}

/*	Clear the interrupt and restore the CTRL reg*/
mtsdram(DDR0_00, saved_status | DDR0_00_INT_ACK_MASK);	
mtsdram(DDR0_22, saved_ctrl);

if(g_ecc_index < MAX_ECC_INT_EVENTS)
	{/*now it is safe to move into globals*/
	g_ecc[g_ecc_index].status		= status;
	g_ecc[g_ecc_index].synd			= synd;
	g_ecc[g_ecc_index].u_addr		= u_addr;
	g_ecc[g_ecc_index].u_data		= u_data;
	g_ecc[g_ecc_index].c_addr		= c_addr;
	g_ecc[g_ecc_index].c_data31_0	= c_data31_0;
	g_ecc[g_ecc_index].c_data63_32	= c_data63_32;
	++g_ecc_index;
	}

return(0);
}/*end FUNCTION: post_ecc_int_handler() */


/*****************************************************************************
*	FUNCTION:	post_ecc_init()
*	DATE:		6-18-07
*
*	Call this function to initialize the interrupt traps for ECC detection.
*
*****************************************************************************/
s32 post_ecc_init(char *msg)
{
u32	 saved_ctrl;
char *p;

/*	Clear globals and ECC interrupt info areas*/
g_ecc_index			= 0;
g_total_ecc_ints	= 0;
memset(g_ecc, 0, (sizeof(t_ecc_data) * MAX_ECC_INT_EVENTS) );

/*	Install the UIC2 and vector 69 DRR0 trap*/
irq_install_handler(VECNUM_DDR2, (interrupt_handler_t *)post_ecc_int_handler, NULL);
irq_install_handler(VECNUM_UIC2NC, (interrupt_handler_t *)uic2_interrupt, NULL);

/*	Create a usefull info message*/
mfsdram(DDR0_22, saved_ctrl);	
switch(DDR0_22_CTRL_RAW_MASK & saved_ctrl)
	{
	case DDR0_22_CTRL_RAW_ECC_DISABLE:
		p = "ECC detection and correction DISABLED";
		break;
	case DDR0_22_CTRL_RAW_ECC_CHECK_ONLY:
		p = "ECC detection only, NO correction";
		break;
	case DDR0_22_CTRL_RAW_NO_ECC_RAM:
		p = "ECC disabled, no ram";
		break;
	case DDR0_22_CTRL_RAW_ECC_ENABLE:
		p = "ECC detection and correction ENABLED";
		break;
	}

sprintf(msg, "ECC vector init, %s", p);
return(0);
}/*end FUNCTION: post_ecc_init()*/


/*****************************************************************************
*	FUNCTION:	ecc_force_error()
*	DATE:		6-18-07
*
*	Force a ECC error at the specified address.
*	An ECC read operation returns 128 bits/16 bytes/4 words.  
*	127 - 96,  95 - 64, 63 - 32, 31 - 0
*
*	I can force errors as directed by the users manual for bits 127-64.
*	I cannot seem to force error on bits 63-0.  All read and write operations
*	pertaining to forcing and testing ECC errors must occur on 16 byte(128bit)
*	boundaries.  See global define ecc_synd_codes[] for more coments.
*
*****************************************************************************/
s32 ecc_force_error(u32 addr, u32 bit)
{
u32 saved_ctrl;
register volatile u32 fwc;
register volatile u32 *faddr	= (u32*)addr;
register volatile u32 fdata		= *faddr;
register volatile u32 synd		= (u32)ecc_synd_codes[bit];

mfsdram(DDR0_22, saved_ctrl);
mfsdram(DDR0_23, fwc);
fwc |= DDR0_23_FWC_MASK;

/*	Set detect but not correct*/
mtsdram(DDR0_22, 
	(~DDR0_22_CTRL_RAW_MASK & saved_ctrl) | DDR0_22_CTRL_RAW_ECC_CHECK_ONLY);

asm("sync");							/*flush writes*/

mtsdram(DDR0_31, synd);					/*Set bad syndrom byte*/					
mtsdram(DDR0_23, fwc);					/*set force write check*/

*faddr = fdata;							/*force a bad write*/	

mtsdram(DDR0_22, saved_ctrl);
return(0);
}/*end FUNCTION: ecc_force_error()*/


/*****************************************************************************
*	FUNCTION:	post_ecc_test()
*	DATE:		6-20-07
*
*	Force a single bit correctable error and verify the interrupt.
*
*****************************************************************************/
s32 post_ecc_test(u32 start_addr, u32 bcnt, u32 flag, char *msg)
{
u32 rc;
u32 *addr, *saddr, *eaddr;
u32 bad_data;
u32 offset;
u32 orig_bit;
u32 bit;
u32 err_data_bit;
u32 waiting;
u32 saved_data;
u32 corrected_data;
u32 last_int_cnt;


if( (start_addr < CFG_LOAD_ADDR) || (start_addr > CFG_MEMTEST_END_RAM_EXE) )
	{
	sprintf(msg, "Address must be between 0x%08lx and 0x%08lx\n", 
									CFG_LOAD_ADDR, CFG_MEMTEST_END_RAM_EXE);
	rc = E_POST_ECC_TEST_BAD_ADDR_RANGE;
	goto ecc_error_exit;
	}

/*	Addresses MUST be 128bit/16byte/4word aligned, and at least one ECC line*/
if(bcnt < 0x10)
	bcnt = 0x10;

saddr	= (u32*)(start_addr & 0xfffffff0);	
eaddr	= (u32*)((start_addr + bcnt) & 0xfffffff0);	
addr	= saddr;

g_total_ecc_ints	= 0;
last_int_cnt		= 0;


if(VERBOSE_ON & flag)
	printf("Start=0x%08lx end=0x%08lx current=0x%08lx\r", saddr, eaddr, addr);

while(addr < eaddr)
	{
	/*display operator entertainment message every 1K*/
	if((VERBOSE_ON & flag) && !((u32)addr % 1024) )	
		printf("Start=0x%08lx end=0x%08lx current=0x%08lx     \r", 
														saddr, eaddr, addr);

	for(bit = POST_ECC_BIT_START; bit < MAX_ECC_BIT; ++bit)
		{
		/*	Reset previous cnt, get word offset and error bit within the word*/
		g_ecc_index		 = 0;	

		ecc_bit2offset(bit, &offset, &err_data_bit);	
		saved_data = addr[offset];			/*save word we are corrupting*/

		ecc_force_error( (u32)addr, bit);	/*Set bit, read to force error*/
		asm("sync");
		asm("sync");
		asm("sync");
		bad_data = addr[offset];
		asm("sync");
		asm("sync");
		asm("sync");


		if(ECC_DEBUG_ON & flag)
			printf("Addr=0x%08lx bit=%3d saved data=0x%08lx forced data=0x%08lx\n", 
											addr, bit, saved_data, bad_data);

		/*	Wait up to 1000usec for int to propagate*/
		for(waiting = 100; g_total_ecc_ints <= last_int_cnt; --waiting)
			{
			udelay(10);		
			if(!waiting)
				{
				sprintf(msg, 
					"Expected ECC error at 0x%08lx on bit %d did not occur\n", 
															(u32)addr, 	bit);
				rc = E_POST_ECC_NO_EXPECTEED_ERROR;
				goto ecc_error_exit;
				}
			}
#if 0
		/*Interrupt occured verify corrected addrs=starting addr of ECC line*/
		if(g_ecc[0].c_addr != (u32)addr)
			{
			sprintf(msg, "ECC corrected address, exp=0x%08lx  rec=0x%08lx\n", 
												(u32)addr, 	g_ecc[0].c_addr);	
			rc = E_POST_ECC_INVALID_CORRECTED_ADDR;
			goto ecc_error_exit;
			}
#endif

		/*	
		*	Verify the correct bit has been "corrected".  If the origional 
		*	bit is the same level as the "forced" bit, we failed. 
		*/
		if(err_data_bit & saved_data)
			orig_bit = 1;
		else
			orig_bit = 0;

		/*	Forcesd bit MUST be different than origional*/
		if( (  (bad_data & err_data_bit) && orig_bit) ||
		    ((!(bad_data & err_data_bit)) && !orig_bit) )
			{
			sprintf(msg, 
				"Expected bit was not forced, exp=0x%08lx  rec=0x%08lx\n", 
													err_data_bit, bad_data);	
			rc = E_POST_ECC_EXP_BIT_NOT_FORCE_BAD;
			goto ecc_error_exit;
			}

		/*	Verify corrected data*/
		if( (0 == offset) || (2 == offset) )
			corrected_data = g_ecc[0].c_data63_32;
		else
			corrected_data = g_ecc[0].c_data31_0;

		if(corrected_data !=  saved_data)
			{
			sprintf(msg, 
				"Corrected data not equal to origional exp=0x%08lx rec=0x%08lx\n", 
				saved_data, corrected_data);	
			rc = E_POST_ECC_CORRECTED_DATA_NOTEQ_ORIG;
			goto ecc_error_exit;
			}

		/*	Verify single bit corrected status*/
		if( !(g_ecc[0].status & DDR0_00_INT_STATUS_BIT2) )
			{
			sprintf(msg, 
				"Single correctable ECC event did not set status=0x%08lx\n", 
															g_ecc[0].status);	
			rc = E_POST_ECC_SINGLE_CORRECTED_STATUS_NOT_SET;
			goto ecc_error_exit;
			}

		addr[offset] = saved_data;		/*restore data and write correct ECC*/
		++last_int_cnt;					/*bump int cnt*/

		if ( ctrlc() )		/* check for ctrl-c to abort... */
			return(EXIT_BY_CTRLC);
		}/*end bit for*/

	addr += 4;
	}/*end while*/
return(0);


/*	
*	ERROR
*	Make a common error exit (man I hate goto's) so we can display debug info.
*/
ecc_error_exit:

if(VERBOSE_ON & flag)
	{
	printf("\nBit=%d,  Line Addr=0x%08lx,  Word offset=%d,  err_data_bit=0x%08lx\n"
			 "saved_data=0x%08lx,  bad_data=0x%08lx,  delay_cnt=%d total_ints=%d\n", 
			 bit, (u32)addr, offset, err_data_bit, saved_data,  bad_data, waiting, 
			g_total_ecc_ints);
	}

return(rc);
}/*end FUNCTION: post_ecc_test()*/



/*****************************************************************************
*	FUNCTION:	ecc_bit2offset
*	DATE:		6-20-07
*
*	The ECC line is 128 bits or 4 words.  This function takes the bit being 
*	tested and finds the word offset from the start of the ECC line, and the 
*	bit position with in the word.
*	INPUT:	bit - bit to be tested
*			offset - pointer to pass back the word offset.
*			err_data_bit - bit position in the word
*****************************************************************************/
static s32 ecc_bit2offset(u32 bit, u32 *offset, u32 *err_data_bit)
{
/*	Locate the bit that should be set*/
if(bit <= 31)
	*offset = 3;					/*4th word*/
else if(bit <= 63)
	{
	*offset = 2;					/*3rd word*/
	bit -= 32;
	}
else if(bit <= 95)
	{
	*offset = 1;					/*2nd word*/
	bit -= 64;
	}
else
	{
	*offset = 0;					/*1st word*/
	bit -= 96;
	}

*err_data_bit = (0x1 << bit);
return(0);
}/*end FUNCTION: ecc_bit2offset()*/


#endif /*CONFIG_CMC_10G*/



