/****************************************************************************

  Copyright (c) 2006, Dell Inc
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

  - Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
    
  - Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution. 

  - Neither the name of Dell Inc nor the names of its contributors
    may be used to endorse or promote products derived from this software 
    without specific prior written permission. 

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. 

*****************************************************************************/

/*
 * Utility support
 */
#include <common.h>		/*must be first in file*/
#include <command.h>

#ifdef CONFIG_CMC_10G	/*only compile in for a CMC_10G build*/


/******************************************************************************
*
*	FILE: util_switch.c
*
******************************************************************************/


/******************************************************************************
*  Include files
******************************************************************************/
#include <util_defs.h>
#include <util_ret_codes.h>

#include <miiphy.h>

/******************************************************************************
*  Local Defines
******************************************************************************/
#define BCM5324_PSEUDOPHY		(u8) 0x1E
#define BCM5324_PSEUDOPHY_PAGE_REG	(u8) 0x10
#define BCM5324_PSEUDOPHY_ADDR_REG	(u8) 0x11
#define BCM5324_PSEUDOPHY_STAT_REG	(u8) 0x12
#define BCM5324_PSEUDOPHY_DAT0_REG	(u8) 0x18
#define BCM5324_PSEUDOPHY_DAT1_REG	(u8) 0x19
#define BCM5324_PSEUDOPHY_DAT2_REG	(u8) 0x1A
#define BCM5324_PSEUDOPHY_DAT3_REG	(u8) 0x1B

#define BCM5324_PSEUDOPHY_PAGE_MDIO_MASK	(u16)0x0001 /*Set MDIO Access*/

#define BCM5324_PSEUDOPHY_ADDR_NOOP_MASK	(u16) 0x0011 
#define BCM5324_PSEUDOPHY_ADDR_WRITE_MASK	(u16) 0x0001 /*Mask in for write op*/
#define BCM5324_PSEUDOPHY_ADDR_READ_MASK	(u16) 0x0002 /*Mask in for read op*/

/******************************************************************************
*  Globals
******************************************************************************/
const char BCM5324_ETH_NAME[]="ppc_4xx_eth0"; /*This is a horrible way of 
						choosing the device. But the
						only one that doesn't seem to 
						require rewriting uBoot's net
						support */

/******************************************************************************
*  Prototypes/Externs
******************************************************************************/
extern s32	mcoptind;
extern char	*mcoptarg[];

/******************************************************************************
*	List all structures defines
******************************************************************************/
t_options util_switch_options[] = 
	{
	/*Options*/
	{"help",	"Display available actions."},
	{"lab",		"Lab settings = open switch, auto-neg both phys."},
	{"reneg",	"Renegotiate both phys."},
	{"read",	"Read switch register."},
	{"write", 	"Write switch register."},
	{"reset", 	"Reset switch."},
	{ NULL, NULL }
	};


t_usage util_switch_usage = { 
	"switch", 
					
	/*Syntax*/
	"switch  [-help]\n"
	"        [-lab]\n"
	"        [-reneg]\n"
	"        [-read <register>]\n"
	"        [-write <register> <value>]\n"
	"        [-reset]\n",

	/*Description*/
	"This command allows manipulation of the switch.\n\n"
	"Examples:\n"
	"util switch -reneg\n"
	"util switch -read 0x0 0x2\n"
	"util switch -write 0x02 0x88 0x12345678\n",
	
	util_switch_options
	}; /*end t_usage util_i2c_usage*/

/******************************************************************************
*
*       FUNCTION: bcm5324_write()
*       DATE: 03/06/07
*
******************************************************************************/

u32     bcm5324_write(u8 page, u8 addr, u16 *out)
{
        u16 data;

        data = page; data <<= 8;
        data |= BCM5324_PSEUDOPHY_PAGE_MDIO_MASK;

        if(miiphy_write(BCM5324_ETH_NAME,
                  BCM5324_PSEUDOPHY,
                  BCM5324_PSEUDOPHY_PAGE_REG,
                  data)) return 0;

        if( miiphy_write(BCM5324_ETH_NAME,
                 BCM5324_PSEUDOPHY,
                 BCM5324_PSEUDOPHY_DAT0_REG,
                 out[0])) return 0;

        if( miiphy_write(BCM5324_ETH_NAME,
                 BCM5324_PSEUDOPHY,
                 BCM5324_PSEUDOPHY_DAT1_REG,
                 out[1])) return 0;

        if( miiphy_write(BCM5324_ETH_NAME,
                 BCM5324_PSEUDOPHY,
                 BCM5324_PSEUDOPHY_DAT2_REG,
                 out[2])) return 0;

        if( miiphy_write(BCM5324_ETH_NAME,
                 BCM5324_PSEUDOPHY,
                 BCM5324_PSEUDOPHY_DAT3_REG,
                 out[3])) return 0;    

        data = addr;
        data <<= 8;
        data |= BCM5324_PSEUDOPHY_ADDR_WRITE_MASK;

        if(miiphy_write(BCM5324_ETH_NAME,
                  BCM5324_PSEUDOPHY,
                  BCM5324_PSEUDOPHY_ADDR_REG,
                  data)) return 0;

        if(miiphy_read(BCM5324_ETH_NAME,
                 BCM5324_PSEUDOPHY,
                 BCM5324_PSEUDOPHY_ADDR_REG,
                 &data)) return 0;

        while(data & BCM5324_PSEUDOPHY_ADDR_NOOP_MASK) {
                if(miiphy_read(BCM5324_ETH_NAME,
                         BCM5324_PSEUDOPHY,
                         BCM5324_PSEUDOPHY_ADDR_REG,
                         &data)) return 0;
        }

        return 1;
}

/******************************************************************************
*
*       FUNCTION: bcm5324_read()
*       DATE: 03/06/07
*
******************************************************************************/

u32 	bcm5324_read(u8 page, u8 addr, u16 *out)
{
	u16 data;
	
	data = page; data <<= 8;
	data |= BCM5324_PSEUDOPHY_PAGE_MDIO_MASK;

	if(miiphy_write(BCM5324_ETH_NAME, 
		  BCM5324_PSEUDOPHY, 
		  BCM5324_PSEUDOPHY_PAGE_REG,
		  data)) return 0;

	data = addr;
	data <<= 8;
	data |= BCM5324_PSEUDOPHY_ADDR_READ_MASK;
	
	if(miiphy_write(BCM5324_ETH_NAME,
		  BCM5324_PSEUDOPHY,
		  BCM5324_PSEUDOPHY_ADDR_REG,
		  data)) return 0;
	
	if(miiphy_read(BCM5324_ETH_NAME,
		 BCM5324_PSEUDOPHY,
		 BCM5324_PSEUDOPHY_ADDR_REG,
                 &data)) return 0;

	while(data & BCM5324_PSEUDOPHY_ADDR_NOOP_MASK) {
	        if(miiphy_read(BCM5324_ETH_NAME,
                         BCM5324_PSEUDOPHY,
                         BCM5324_PSEUDOPHY_ADDR_REG,
                         &data)) return 0;
	}
 			
	if( miiphy_read(BCM5324_ETH_NAME,
                 BCM5324_PSEUDOPHY,
                 BCM5324_PSEUDOPHY_DAT0_REG,
                 &out[0])) return 0;

        if( miiphy_read(BCM5324_ETH_NAME,           
                 BCM5324_PSEUDOPHY,       
                 BCM5324_PSEUDOPHY_DAT1_REG,       
                 &out[1])) return 0;

        if( miiphy_read(BCM5324_ETH_NAME,           
                 BCM5324_PSEUDOPHY,       
                 BCM5324_PSEUDOPHY_DAT2_REG,       
                 &out[2])) return 0;

        if( miiphy_read(BCM5324_ETH_NAME,           
                 BCM5324_PSEUDOPHY,       
                 BCM5324_PSEUDOPHY_DAT3_REG,       
                 &out[3])) return 0;

	return 1;
}

/******************************************************************************
*
*       FUNCTION: switch_read()
*       DATE: 03/06/07
*
******************************************************************************/

u32	switch_read(char *page, char *addr)
{
	char *cptr1;
	s32 micrel, i;
	u32 ret;
	u16 n_page, n_addr;
	u16 data[4];

        n_page = (u32)simple_strtol( page, &cptr1, 0 );             
        if( (s32)cptr1 < (strlen(page) + (s32)page) )      
        {
                printf("ERROR: \"%s\" cannot be converted to int.\n", page);   
                return(E_REG_INT_CONVERSION);
        }

	n_addr = (u32)simple_strtol( addr, &cptr1, 0 );
        if( (s32)cptr1 < (strlen(addr) + (s32)addr) )
        {
                printf("ERROR: \"%s\" cannot be converted to int.\n", addr);
                return(E_REG_INT_CONVERSION);
        }

	micrel = get_micrel_en();
	set_micrel_en(1); /* BLR - Switch to BCM side for switch comms*/
	ret = bcm5324_read(n_page,n_addr,data);
	set_micrel_en(micrel);

	if(ret) {
		printf("0x");
		for(i=0;i<4;i++) 
			printf("%.4X",data[3-i]);
		printf("\n");
		return 0;
	}
	return -1;
}

/******************************************************************************
*
*       FUNCTION: switch_write()
*       DATE: 03/06/07
*
******************************************************************************/

u32     switch_write(char *page, char *addr, char *data)     
{         
        char *cptr1;
        s32 micrel;
        u32 ret;                
        u16 n_page, n_addr;       
	u32 bigdata;  
        u16 n_data[4];

        n_page = (u32)simple_strtol( page, &cptr1, 0 );
        if( (s32)cptr1 < (strlen(page) + (s32)page) )
        {
                printf("ERROR: \"%s\" cannot be converted to int.\n", page);
                return(E_REG_INT_CONVERSION);
        }

        n_addr = (u32)simple_strtol( addr, &cptr1, 0 );
        if( (s32)cptr1 < (strlen(addr) + (s32)addr) )
        {
                printf("ERROR: \"%s\" cannot be converted to int.\n", addr);
                return(E_REG_INT_CONVERSION);
        }

	bigdata = (u32)simple_strtol( data, &cptr1, 0 );
        if( (s32)cptr1 < (strlen(data) + (s32)data) )
        {                
                printf("ERROR: \"%s\" cannot be converted to int.\n", data);
                return(E_REG_INT_CONVERSION);
        }

	n_data[0] = (u16)(bigdata & 0xFFFF);
	n_data[1] = (u16)(0xFFFF  & (bigdata >>16));
	n_data[2] = 0; /* Until I can figure out how to read in 64bit #s*/
	n_data[3] = 0;
	
        micrel = get_micrel_en();
        set_micrel_en(1); /* BLR - Switch to BCM side for switch comms*/
        ret = bcm5324_write(n_page,n_addr,n_data);
        set_micrel_en(micrel);

        if(ret) {
		printf("Write complete.\n");
		return 0;
	}
        return -1;
}

s32	reneg_script(void); /* BLR - I need to clean up these */
/******************************************************************************
*
*       FUNCTION: config_1gig()
*       DATE:         
*
******************************************************************************/
s32 config_lab()
{
	u16 data[4];
	u16 i;

        data[3] = data[2] = data[1] = 0;
        data[0] = 0x0000;

	/* Disable MAC address aging */
        if(!bcm5324_write((u8)0x02, (u8)0x6, data)) return -1;

	/* Set up all internal Phy's to 100M FULL */
	for(i=0;i<25;i++) {
		miiphy_write(BCM5324_ETH_NAME,i,0x0,0x2100);
	}

	/* Renegotiate external GigE Phy's */
	if(reneg_script()!=0) return -1;

	return 0;
}

/******************************************************************************
*
*       FUNCTION: gigphy_script()
*       DATE:
*
******************************************************************************/
s32 lab_script()
{
	s32 micrel;
	s32 ret;

	set_switch_reset();
	clear_switch_reset();
	micrel = get_micrel_en();
	set_micrel_en(1);
	ret = config_lab();
	set_micrel_en(micrel);

	if(!ret) {
		printf("Configuration complete.\n");
		return 0;
	}

	printf("Switch configuration error.\n");
	return -1;
}

/******************************************************************************
*
*       FUNCTION: print_Speed()   
*       DATE:
*
******************************************************************************/
void 	print_Speed(u8 addr, u32 speed, u32 duplex)
{
	printf("Phy 0x%.2X negotiated to ",addr);

	switch(speed) {
	        case _10BASET   : printf("  10M "); break;        
                case _100BASET  : printf(" 100M "); break;
                case _1000BASET : printf("1000M "); break; 
                default : return; /* Should never happen */
	}

	if(duplex == FULL)
		printf("Full duplex.\n");
	else
		printf("Half duplex.\n");
}

/******************************************************************************
*
*       FUNCTION: bcm5324_portSpeed()
*       DATE:
*
******************************************************************************/
s32     bcm5324_portmatchSpeed(u8 addr)
{
	u8 reg = 0xA0; /* State override register base address */
	u16 val[4];
	u32	speed;
	u32	duplex;


	speed = miiphy_speed(BCM5324_ETH_NAME, addr); /*Won't fail, may lie*/
	duplex = miiphy_duplex(BCM5324_ETH_NAME, addr); /*Same*/

	print_Speed(addr,speed,duplex);

	reg += addr;
	val[3] = val[2] = val [1] = 0;
	val[0] = 0x0041; /* 0x1 means force link up*/

	switch(speed) {
		case _10BASET 	: val[0] |= 0x00; break;
		case _100BASET	: val[0] |= 0x04; break;
		case _1000BASET : val[0] |= 0x08; break;
		default : return -1; /* Should never happen */
	}	
	
	if(duplex == FULL)
		val[0] |= 0x02;
	if(!bcm5324_write(0x0,reg,val)) return -1;
	return 0;
}

/******************************************************************************
*
*       FUNCTION: reneg_script()
*       DATE:
*
******************************************************************************/
s32 reneg_script()
{
	u16 data[4];

	/*Set PHYs in delayed mode*/
	if(miiphy_write(BCM5324_ETH_NAME, 0x19, 0x1C, 0x8E00)) return -1;
        if(miiphy_write(BCM5324_ETH_NAME, 0x1A, 0x1C, 0x8E00)) return -1;

	if(bcm5324_portmatchSpeed(0x19)) return -1; 
	if(bcm5324_portmatchSpeed(0x1A)) return -1;

        /* Read status register twice to ensure link status is valid */
        miiphy_read(BCM5324_ETH_NAME, 0x19, 0x01, &data[0]);
        miiphy_read(BCM5324_ETH_NAME, 0x19, 0x01, &data[0]);
        if(data[0]&0x4)
                miiphy_write(BCM5324_ETH_NAME,0x19,0x1C,0xB8F3);
        else
                miiphy_write(BCM5324_ETH_NAME,0x19,0x1C,0xB803);

        /* Read status register twice to ensure link status is valid */
        miiphy_read(BCM5324_ETH_NAME, 0x1A, 0x01, &data[0]);
        miiphy_read(BCM5324_ETH_NAME, 0x1A, 0x01, &data[0]);
        if(data[0]&0x4)
                miiphy_write(BCM5324_ETH_NAME,0x1A,0x1C,0xB8F3);
        else
                miiphy_write(BCM5324_ETH_NAME,0x1A,0x1C,0xB803);

	return 0;
}

/******************************************************************************
*
*       FUNCTION: switch_renegotiate()
*       DATE:
*
******************************************************************************/
s32 switch_renegotiate()               
{
	s32 micrel;
	s32 ret;

        micrel = get_micrel_en();
        set_micrel_en(1);
        ret = reneg_script();
        set_micrel_en(micrel);

        if(!ret) {
                printf("PHYs renegotiated and propagated to switch..\n");
                return 0;
        }

        printf("Switch configuration error.\n");

	return -1;
}

/******************************************************************************
*
*       FUNCTION: switch_reset() 
*       DATE:
*
******************************************************************************/
void switch_reset()
{
	set_switch_reset();
	clear_switch_reset();
}

/******************************************************************************
*
*	FUNCTION: util_switch()
*	DATE:	
*
******************************************************************************/
s32 util_switch(int argc, char *argv[])
{
u32  index = 0, rc = 0;
char msg[MAX_MSG_LEN];		/*message buffer*/


//	Get any options
mcoptind = 2;	
while( (index = mcgetopt(argc, argv, "-help -lab -reneg -read:2 -write:3 -reset")) != MCGETOPT_EOF)
	{
	switch(index) 
		{
		case 0 :/*help*/
			printf("-help -lab -reneg -read -write -reset\n");
			break;
		case 1 : /*lab*/
			lab_script();
			break;
		case 2 : /*reneg*/
			 switch_renegotiate();
			 break;
		case 3 : /*read*/
			 switch_read(mcoptarg[0],mcoptarg[1]);
			 break;
		case 4 : /*write*/
			 switch_write(mcoptarg[0],mcoptarg[1],mcoptarg[2]);
			 break;
		case 6 : /*reset*/
			 switch_reset();
			 printf("Switch Reset.\n");
			 break;
		default:		/*not good*/
			rc = mcgetopt_error(index, msg, mcoptarg[0]);
			printf("%s %s:%s\n", argv[0], argv[1], msg);
			return(rc);
			break;
		}
	}/*end mcgetopt while*/
return(0);
}/* end FUNCTION: util_switch()*/

#endif /*CONFIG_CMC_10G*/



