/****************************************************************************

  Copyright (c) 2006, Dell Inc
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

  - Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
    
  - Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution. 

  - Neither the name of Dell Inc nor the names of its contributors
    may be used to endorse or promote products derived from this software 
    without specific prior written permission. 

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. 

*****************************************************************************/



#include <common.h>		/*must be first in file*/
#include <command.h>

#ifdef CONFIG_CMC_10G	/*only compile in for a CMC_10G build*/


/******************************************************************************
*
*	FILE: util_template.c
*	Use this file to quickly implimit a CMC utility command.
*
*	Go to "u-boot/common/cmd_util.c" and follow directions on how to add
*	a new command.
*
******************************************************************************/


/******************************************************************************
*  Include files
******************************************************************************/
#include <util_defs.h>
#include <util_ret_codes.h>


/******************************************************************************
*  Local Defines
******************************************************************************/


/******************************************************************************
*  Globals
******************************************************************************/


/******************************************************************************
*  Prototypes/Externs
******************************************************************************/
extern s32	mcoptind;
extern char	*mcoptarg[];


/******************************************************************************
*	List all structures defines
******************************************************************************/
t_options util_template_options[] = 
	{
	/*Options*/
	{"c",	"Example,  single char option, no arg."},
	{"b",	"Example,  single char option, with arg."},
	{"xyz",	"Example of a multicharacter option."},
	{"a",	"Single address,or range. Char \':\' specifies start:end pair,\n"
			"    \'#\' for a start#bytes count.  \"0x\" for hex, else decimal.\n"
			"    -a 0x<startAddress>:0x<endAddress>\n"
			"    -a 0x<startAddress>#0x<byte count>\n"
			"    -a 0x<startAddress, single read>" },

	{ NULL, NULL }
	};


t_usage util_template_usage = { 
	"template", 
					
	/*Syntax*/
	"template  [-c] [-b <text>] [-xyz] [-a <start:end> | <start#count>]\n",

	/*Description*/
	"Template example for adding a new util command.\nExamples:\n\n"
	"util template -c\n"
	"util template -b <any text, no spaces or double quoted>\n"
	"util template -xyz\n"
	"util template -a 0x100#20\n", 
	
	util_template_options
	}; /*end t_usage util_template_usage*/


/******************************************************************************
*
*	FUNCTION: util_template()
*	DATE:	
*
******************************************************************************/
s32 util_template(int argc, char *argv[])
{
u32  index = 0, rc = 0;
char msg[MAX_MSG_LEN];		/*message buffer*/
char addr_str[40];			/*save the address range*/
u32  saddr, eaddr, bcnt;	/*start, end addr, and the byte count*/
char c;

addr_str[0]	= '\0';				/*null the string*/

//	Get any options
mcoptind = 2;	
while( (index = mcgetopt(argc, argv, "-c -b:1 -xyz -a:1 ")) != MCGETOPT_EOF)
	{
	switch(index) 
		{
		case 0 :		/*option "c" */
			printf("Single character option \"c\" entered, no arguments.\n");
			break;

		case 1 :		/*option "b" */
			printf("Single character option \"b\" entered, arg = %s.\n", mcoptarg[0]);
			break;

		case 2 :		/*option "xyz" */
			printf("Multicharacter option \"zyz\" entered, no arguments.\n");
			break;

		case 3 :		/*option "a" address range*/
			strcpy(addr_str, mcoptarg[0]);

			if( (rc = parse_addr(addr_str, &saddr, &eaddr, &bcnt, WIDTH_8, msg)) != 0)
				{
				printf("%s %s:ERROR:RC:0x%x%s\n", argv[0], argv[1], rc, msg);
				return(rc);
				}

			printf("Start address=0x%x End address=0x%x Bytecount=%d\n", 
					saddr, eaddr, bcnt);
			break;


		default:		/*not good*/
			rc = mcgetopt_error(index, msg, mcoptarg[0]);
			printf("%s %s:%s\n", argv[0], argv[1], msg);
			return(rc);
			break;
		}
	}/*end mcgetopt while*/


printf("Execute template specific test here.\n");
return(0);
}/* end FUNCTION: util_template()*/


#endif /*CONFIG_CMC_10G*/



