/*****************************************************************************
*
*     Author: Xilinx, Inc.
*
*
*     This program is free software; you can redistribute it and/or modify it
*     under the terms of the GNU General Public License as published by the
*     Free Software Foundation; either version 2 of the License, or (at your
*     option) any later version.
*
*
*     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
*     COURTESY TO YOU. BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
*     ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION OR STANDARD,
*     XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION IS FREE
*     FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE FOR OBTAINING
*     ANY THIRD PARTY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
*     XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
*     THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY
*     WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM
*     CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND
*     FITNESS FOR A PARTICULAR PURPOSE.
*
*
*     Xilinx hardware products are not intended for use in life support
*     appliances, devices, or systems. Use in such applications is
*     expressly prohibited.
*
*
*     (c) Copyright 2002-2004 Xilinx Inc.
*     All rights reserved.
*
*
*     You should have received a copy of the GNU General Public License along
*     with this program; if not, write to the Free Software Foundation, Inc.,
*     675 Mass Ave, Cambridge, MA 02139, USA.
*
*****************************************************************************/
/****************************************************************************/
/**
*
* @file xiic_l.h
*
* This header file contains identifiers and low-level driver functions (or
* macros) that can be used to access the device.  High-level driver functions
* are defined in xiic.h.
*
* <pre>
* MODIFICATION HISTORY:
*
* Ver   Who  Date     Changes
* ----- ---- -------- -----------------------------------------------
* 1.00b jhl  05/07/02 First release
* 1.01c ecm  12/05/02 new rev
* 2.00  blr  03/03/07 ported to uboot/ppc440epx for 10G CMC
* </pre>
*
*****************************************************************************/

#ifndef XIIC_L_H		/* prevent circular inclusions */
#define XIIC_L_H		/* by using protection macros */

/***************************** Include Files ********************************/
#include <common.h>


/************************** Constant Definitions ****************************/

#define XIL_ACCESS_TYPE volatile u16 *

#define CMC_I2C_PSU123  0xA0001A00
#define CMC_I2C_PSU456  0xA0001C00
#define CMC_I2C_IOMSET1 0xA0001600
#define CMC_I2C_IOMSET2 0xA0001800
#define CMC_I2C_LKVM	0xA0001E00

/*
 * Register offsets in bytes from RegisterBase. Three is added to the
 * base offset to access LSB (IBM style) of the word
 */
#define XIIC_IISR_REG_OFFSET(x) ((XIL_ACCESS_TYPE)(0x020+x))	/* Int. Status Reg    */
#define XIIC_IIER_REG_OFFSET(x) ((XIL_ACCESS_TYPE)(0x028+x)) 	/* Int. Enable Reg    */
#define XIIC_RST_REG_OFFSET(x)  ((XIL_ACCESS_TYPE)(0x040+x))	/* IIC Reset Reg      */
#define XIIC_CR_REG_OFFSET(x)   ((XIL_ACCESS_TYPE)(0x100+x))	/* Control Register   */
#define XIIC_SR_REG_OFFSET(x)   ((XIL_ACCESS_TYPE)(0x104+x))	/* Status Register    */
#define XIIC_DTR_REG_OFFSET(x)  ((XIL_ACCESS_TYPE)(0x108+x))	/* Data Tx Register   */
#define XIIC_DRR_REG_OFFSET(x)  ((XIL_ACCESS_TYPE)(0x10C+x))	/* Data Rx Register   */
#define XIIC_ADR_REG_OFFSET(x)  ((XIL_ACCESS_TYPE)(0x110+x))	/* Address Register   */
#define XIIC_TFO_REG_OFFSET(x)  ((XIL_ACCESS_TYPE)(0x114+x))	/* Tx FIFO Occupancy  */
#define XIIC_RFO_REG_OFFSET(x)  ((XIL_ACCESS_TYPE)(0x118+x))	/* Rx FIFO Occupancy  */
#define XIIC_RFD_REG_OFFSET(x)  ((XIL_ACCESS_TYPE)(0x120+x))	/* Rx FIFO Depth reg  */
#define XIIC_NACK_REG_OFFSET(x) ((XIL_ACCESS_TYPE)(0x124+x))	/* NACK counter reg   */

/* IIC Reset Key*/
#define XIIC_RESET_KEY 	0xA; /* Reset key for reset register */

/* Control Register masks */

#define XIIC_CR_ENABLE_DEVICE_MASK        0x01	/* Device enable = 1      */
#define XIIC_CR_TX_FIFO_RESET_MASK        0x02	/* Transmit FIFO reset=1  */
#define XIIC_CR_MSMS_MASK                 0x04	/* Master starts Txing=1  */
#define XIIC_CR_DIR_IS_TX_MASK            0x08	/* Dir of tx. Txing=1     */
#define XIIC_CR_NO_ACK_MASK               0x10	/* Tx Ack. NO ack = 1     */
#define XIIC_CR_REPEATED_START_MASK       0x20	/* Repeated start = 1     */
#define XIIC_CR_GENERAL_CALL_MASK         0x40	/* Gen Call enabled = 1   */
#define XIIC_CR_START_CFG_MASK            0x80	/* Device enable = 1      */

/* Status Register masks */

#define XIIC_SR_GEN_CALL_MASK             0x01	/* 1=a mstr issued a GC   */
#define XIIC_SR_ADDR_AS_SLAVE_MASK        0x02	/* 1=when addr as slave   */
#define XIIC_SR_BUS_BUSY_MASK             0x04	/* 1 = bus is busy        */
#define XIIC_SR_MSTR_RDING_SLAVE_MASK     0x08	/* 1=Dir: mstr <-- slave  */
#define XIIC_SR_TX_FIFO_FULL_MASK         0x10	/* 1 = Tx FIFO full       */
#define XIIC_SR_RX_FIFO_FULL_MASK         0x20	/* 1 = Rx FIFO full       */
#define XIIC_SR_RX_FIFO_EMPTY_MASK        0x40	/* 1 = Rx FIFO empty      */
#define XIIC_SR_TX_FIFO_EMPTY_MASK        0x80  /* 1 = Tx FIFO Empty      */

/* IPIF Interrupt Status Register masks    Interrupt occurs when...       */

#define XIIC_INTR_ARB_LOST_MASK           0x01	/* 1 = arbitration lost   */
#define XIIC_INTR_TX_ERROR_MASK           0x02	/* 1=Tx error/msg complete */
#define XIIC_INTR_TX_EMPTY_MASK           0x04	/* 1 = Tx FIFO/reg empty  */
#define XIIC_INTR_RX_FULL_MASK            0x08	/* 1=Rx FIFO/reg=OCY level */
#define XIIC_INTR_BNB_MASK                0x10	/* 1 = Bus not busy       */
#define XIIC_INTR_AAS_MASK                0x20	/* 1 = when addr as slave */
#define XIIC_INTR_NAAS_MASK               0x40	/* 1 = not addr as slave  */
#define XIIC_INTR_TX_HALF_MASK            0x80	/* 1 = TX FIFO half empty */
#define XIIC_INTR_STRT_ERR3_MASK	  0x100 /* 1 = Error, 0 = No Error*/
#define XIIC_INTR_STRT_ERR2_MASK	  0x200 /* 1 = Error, 0 = No Error*/
#define XIIC_INTR_STRT_ERR1_MASK	  0x400 /* 1 = Error, 0 = No Error*/

#define XIIC_TX_ADDR_SENT             0x00
#define XIIC_TX_ADDR_MSTR_RECV_MASK   0x02

/* The following constants specify the depth of the FIFOs */

#define IIC_RX_FIFO_DEPTH         16	/* Rx fifo capacity               */
#define IIC_TX_FIFO_DEPTH         16	/* Tx fifo capacity               */

/* The following constants specify groups of interrupts that are typically
 * enabled or disables at the same time
 */
#define XIIC_TX_INTERRUPTS                                          \
            (XIIC_INTR_TX_ERROR_MASK | XIIC_INTR_TX_EMPTY_MASK |    \
             XIIC_INTR_TX_HALF_MASK)

#define XIIC_TX_RX_INTERRUPTS (XIIC_INTR_RX_FULL_MASK | XIIC_TX_INTERRUPTS)

/* The following constants are used with the following macros to specify the
 * operation, a read or write operation.
 */
#define XIIC_READ_OPERATION  1
#define XIIC_WRITE_OPERATION 0

/* The following constants are used with the transmit FIFO fill function to
 * specify the role which the IIC device is acting as, a master or a slave.
 */
#define XIIC_MASTER_ROLE     1
#define XIIC_SLAVE_ROLE      0

/**************************** Type Definitions ******************************/

/***************** Macros (Inline Functions) Definitions ********************/

/************************** Function Prototypes *****************************/

u32 	xiic_init(u32 dev);
u32	xiic_iic_send(u32 dev, u8 address, u8 *data, u8 len, u32 rs);
u32 	xiic_iic_recv(u32 dev, u8 address, u8 *data, u8 len, u32 rs);

#endif				/* end of protection macro */
