
/*
 * CMC_10G addition for recover support
 */
#include <common.h>		/*must be first in file*/
#include <command.h>


/*****************************************************************************
*
*	Code in this file is the command interface to the CMC_10G recover 
*	subcomands which provide support for fwupdate recovery along with debug 
*	activities in dev, lab, and at PCBA test.
*
*	DATE:	2-28-07
*
*****************************************************************************/
#ifdef CONFIG_CMC_10G	/*only compile in for a CMC_10G build*/

#include <util_defs.h>
#include <util_ret_codes.h>


/*****************************************************************************
*	Define the command table for the recover command.  Any subcommand that is
*	to be called from the "recover" command must be added here. To add a 
*	subcommand:
*	1) Add the extern t_usage struct below.
*	2) Add the command prototype in util_defs.h.
*	3) Add the command to recover_cmd_table below.
*	4) Add the command into an existing file, or create a new one. If a new 
*      file is created add it in "u-boot/board/amcc/cmc_10g"
*	5) Add the recover_<new cmd>.o to  "u-boot/board/amcc/cmc_10g/Makefile"
*	6) Create a "t_usage <testname>Usage" struct in same file as the command.
*****************************************************************************/
/*
typedef struct  {
    char *test_name;        //pointer to char string of test name
    s32  (*test_func)(int argc, char*argv[]);	//pointer to the test function
	t_usage *usage;			//pointer to the t_usage usage struct
    } t_util_cmd_table;                       
*/

/*****************************************************************************
*	List external references to the command usage statment.
*****************************************************************************/
extern u32	 g_post_rc;
extern char  g_post_msg[];


extern t_usage recover_fwupdate_usage;
extern t_usage recover_getniccfg_usage;
extern t_usage recover_ping_usage;
extern t_usage recover_reset_usage;
extern t_usage recover_setniccfg_usage;

t_util_cmd_table   recover_cmd_table[] = {
	{ "fwupdate",		recover_fwupdate,	 &recover_fwupdate_usage},
	{ "getniccfg",		recover_getniccfg,	 &recover_getniccfg_usage},
	{ "ping",			recover_ping,	     &recover_ping_usage},
	{ "reset",			recover_reset,		 &recover_reset_usage},
	{ "setniccfg",		recover_setniccfg,	 &recover_setniccfg_usage},

	{ NULL, NULL, NULL} };


/*****************************************************************************
*
*	FUNCTION:	do_recover()
*	DATE:		Ported 2-28-07
*
*	The recover command interface provides a user command interface that has 
*	access to its own command table containing fwupdate relatedcommands. 
*	
****************************************************************************/
int do_recover ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
u32	rc = 0, i, j, tmp, max_op_len;
s32	not_found, tab_i;
s8	mbuf[100];
t_util_cmd_table	*recover_cmd_tab = recover_cmd_table;
	

if( (argc == 1) || ( (argc == 2 ) && (!strcmp(argv[1], "help")) ) )
	{
	/* 
	*	The user can enter "recover"  or  "recover help" to display a simple 
	*	syntax/synopsis usage statment from each of the recover subcommands.
	*	found in the recover command table.
	*/
	printf("\n** Recover Reason **:\n");
	printf("%s\nRC = 0x%x\n\n", g_post_msg, g_post_rc);

	printf("Available \"%s\" subcommands:\n\n", argv[0]);

	for(tab_i = 0; recover_cmd_tab[tab_i].usage != NULL; ++tab_i)	
		{
		printf("%s\n", recover_cmd_tab[tab_i].usage->syntax);
		}

	return(0);
	}
else if( (argc >= 3) && (!strcmp(argv[1], "help")) )
	{
	/*
	*	The user can enter "recover help <subcommand>" to get a detailed
	*	description/man page of the indivitual subcommand.
	*/
	not_found = 1;
	tab_i = 0;
	while(not_found)
		{
		not_found = strcmp(recover_cmd_tab[tab_i].test_name, argv[2]);
		if(not_found)
			++tab_i;

		if (recover_cmd_tab[tab_i].test_name == NULL)	
			{
			sprintf(mbuf, "%s: Error...subcommand \"%s\" is not found.\n", 
							argv[0], argv[2]);
			printf("%s", mbuf);
			return(E_RECOVER_SUBCMD_NOT_FOUND);
			}
		}

	/* Display the man page description*/
	printf("\n\nNAME:\n%s\n\nSYNOPSIS:\n%s\n\nDESCRIPTION:\n%s\n\n",
			recover_cmd_tab[tab_i].usage->cmd_name,
			recover_cmd_tab[tab_i].usage->syntax,
			recover_cmd_tab[tab_i].usage->description);

	printf("OPTIONS:");
	for(i = 0; recover_cmd_tab[tab_i].usage->options[i].opt_ch != NULL; ++i)
		{	
		/*
		*	Calc number of spaces needed to keep option descriptions aligned.
		*	(it just looks better...why...picky....I know...its a curse)
		*/ 
		for(j = 0, tmp = 0, max_op_len = 0; 
			recover_cmd_tab[tab_i].usage->options[j].opt_ch != NULL;   ++j)
			{
			tmp = strlen(recover_cmd_tab[tab_i].usage->options[j].opt_ch);
			if(tmp > max_op_len)
				max_op_len = tmp;
			}

		// Display the option char/word
		printf("\n-%s",	recover_cmd_tab[tab_i].usage->options[i].opt_ch);
		tmp = strlen(recover_cmd_tab[tab_i].usage->options[i].opt_ch);

		// Calc number of spaces needed to keep descriptions aligned
		max_op_len = (max_op_len - tmp) + 1;
		while(max_op_len--)
			printf(" ");
		
		// Now print the description
		printf("%s", recover_cmd_tab[tab_i].usage->options[i].opt_msg);
		}
	printf("\n\n");


	return(0); /*exit "recover help" */
	}

/*
*	The recover subcommands are diverse enough they do not have common options.
*	Subcommands themselvs are required to do the getopts parsing if they need it.
*
*	The recover subcommand is in the second argument. Search the command table.
*/
not_found = 1;
tab_i = 0;
while(not_found)
    {
    not_found = strcmp(recover_cmd_tab[tab_i].test_name, argv[1]);
    if(not_found)
        ++tab_i;
    if (recover_cmd_tab[tab_i].test_name == NULL)	
		break;
    }

if(not_found)
	{	
	sprintf(mbuf, "recover: Error...subcommand /'%s/' is not found.\n", argv[1]);
	printf("%s", mbuf);
	return(E_RECOVER_SUBCMD_NOT_FOUND);
	}
else
	{
	/*
	*	The recover subcommand has been found, remove the "recover" from the 
	*	arg list before calling the subcommand. 
	*/
	rc = recover_cmd_tab[tab_i].test_func(argc, argv);
	}

return(rc);					
}/*end FUNCTION: do_recover() */

/* -------------------------------------------------------------------- */

U_BOOT_CMD(
	recover,	10,	1,	do_recover,
	"recover - run a recover subcommand.\n",
		  
	"\t enter \"recover\" for a list of subcommands available\n"
	"\t recover help <subcommand>  -display detailed description of subcommand \n"
);


#endif	/* CONFIG_CMC_10G */


