/*
 * (C) Copyright 2003
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * CMC_10G addition for utility support
 */
#include <common.h>		/*must be first in file*/
#include <command.h>


/*****************************************************************************
*
*	Code in this file is the command interface to the CMC_10G utility 
*	subcomands which provide support for debug activities in dev, lab, and at 
*	PCBA test. Some code may seem like a duplication of existing u-boot code.  
*	However "util" is a port of RACMON code used on a previous product.
*
*	U-boot seems to have positional arguments, existing util commands used 
*	getopt options.  No attempt will be made to convert over.  The end users
*	(us) are use to it, so leave it alone.
*
*	This CMC_10G code is all HW specific and totally focused on the CMC.  I 
*	cannot imagine any of this code being sent to Denx as a u-boot patch.
*	(Who else would want it)
*
*	DATE:	8-31-06
*
*****************************************************************************/
#ifdef CONFIG_CMC_10G	/*only compile in for a CMC_10G build*/

#include <util_defs.h>
#include <util_ret_codes.h>


/*****************************************************************************
*	Define the command table for the util command.  Any subcommand that is to 
*	be called from the "util" command must be added here. To add a subcommand:
*	1) Add the extern t_usage struct below.
*	2) Add the command prototype in util_defs.h.
*	3) Add the command to util_cmd_table below.
*	4) Add the command into an existing file, or create a new one. If a new 
*      file is created add it in "u-boot/board/amcc/cmc_10g"
*	5) Add the util_<new cmd>.o to  "u-boot/board/amcc/cmc_10g/Makefile"
*	6) Create a "t_usage <testname>Usage" struct in same file as the command.
*****************************************************************************/
/*
typedef struct  {
    char *test_name;        //pointer to char string of test name
    s32  (*test_func)(int argc, char*argv[]);	//pointer to the test function
	t_usage *usage;			//pointer to the t_usage usage struct
    } t_util_cmd_table;                       
*/

/*****************************************************************************
*	List external references to the command usage statment.
*****************************************************************************/
extern t_usage util_chas_usage;
extern t_usage util_cpld_usage;
extern t_usage util_ethernet_usage;
extern t_usage util_flash_usage;
extern t_usage util_fpga_usage;
extern t_usage util_fru_usage;
extern t_usage util_i2c_usage;
extern t_usage util_mem_usage;
extern t_usage util_post_usage;
extern t_usage util_reg_usage;
extern t_usage util_reset_usage;
extern t_usage util_script_usage;
extern t_usage util_switch_usage;
extern t_usage util_template_usage;



t_util_cmd_table util_cmd_table[] = {
	{ "chas",			util_chas,			 &util_chas_usage},
	{ "cpld",			util_cpld,			 &util_cpld_usage},
	{ "eth",			util_ethernet,		 &util_ethernet_usage},
	{ "flash",			util_flash,			 &util_flash_usage}, 
	{ "fpga",			util_fpga,			 &util_fpga_usage},
	{ "fru",			util_fru,			 &util_fru_usage},
	{ "i2c",			util_i2c,			 &util_i2c_usage},
	{ "mem",			util_mem,			 &util_mem_usage},
	{ "post",			util_post,			 &util_post_usage},
	{ "reg",			util_reg,			 &util_reg_usage},
	{ "reset",			util_reset,			 &util_reset_usage},
	{ "script",			util_script,		 &util_script_usage},
	{ "switch",			util_switch,		 &util_switch_usage},
	{ "template",		util_template,		 &util_template_usage},
	{ NULL, NULL, NULL} };


/*****************************************************************************
*
*	FUNCTION:	do_util()
*	DATE:		Ported 8-31-06
*
*	The util command interface provides a user command interface that has 
*	access to its own command table containing diagnostic commands and tests. 
*	
****************************************************************************/
int do_util ( cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
u32	rc = 0, i, j, tmp, max_op_len;
s32	not_found, tab_i;
s8	mbuf[100];
t_util_cmd_table	*util_cmd_tab = util_cmd_table;
	

if( (argc == 1) || ( (argc == 2 ) && (!strcmp(argv[1], "help")) ) )
	{
	/* 
	*	The user can enter "util"  or  "util help" to display a simple 
	*	syntax/synopsis usage statment from each of the util subcommands.
	*	found in the util command table.
	*/
	printf("Available \"%s\" subcommands:\n\n", argv[0]);

	for(tab_i = 0; util_cmd_tab[tab_i].usage != NULL; ++tab_i)	
		{
		printf("%s\n", util_cmd_tab[tab_i].usage->syntax);
		}

	return(0);
	}
else if( (argc >= 3) && (!strcmp(argv[1], "help")) )
	{
	/*
	*	The user can enter "util help <subcommand>" to get a detailed
	*	description/man page of the indivitual subcommand.
	*/
	not_found = 1;
	tab_i = 0;
	while(not_found)
		{
		not_found = strcmp(util_cmd_tab[tab_i].test_name, argv[2]);
		if(not_found)
			++tab_i;

		if (util_cmd_tab[tab_i].test_name == NULL)	
			{
			sprintf(mbuf, "%s: Error...subcommand \"%s\" is not found.\n", 
							argv[0], argv[2]);
			printf("%s", mbuf);
			return(E_UTIL_SUBCMD_NOT_FOUND);
			}
		}

	/* Display the man page description*/
	printf("\n\nNAME:\n%s\n\nSYNOPSIS:\n%s\n\nDESCRIPTION:\n%s\n\n",
			util_cmd_tab[tab_i].usage->cmd_name,
			util_cmd_tab[tab_i].usage->syntax,
			util_cmd_tab[tab_i].usage->description);

	printf("OPTIONS:");
	for(i = 0; util_cmd_tab[tab_i].usage->options[i].opt_ch != NULL; ++i)
		{	
		/*
		*	Calc number of spaces needed to keep option descriptions aligned.
		*	(it just looks better...why...picky....I know...its a curse)
		*/ 
		for(j = 0, tmp = 0, max_op_len = 0; 
			util_cmd_tab[tab_i].usage->options[j].opt_ch != NULL;   ++j)
			{
			tmp = strlen(util_cmd_tab[tab_i].usage->options[j].opt_ch);
			if(tmp > max_op_len)
				max_op_len = tmp;
			}

		// Display the option char/word
		printf("\n-%s",	util_cmd_tab[tab_i].usage->options[i].opt_ch);
		tmp = strlen(util_cmd_tab[tab_i].usage->options[i].opt_ch);

		// Calc number of spaces needed to keep descriptions aligned
		max_op_len = (max_op_len - tmp) + 1;
		while(max_op_len--)
			printf(" ");
		
		// Now print the description
		printf("%s", util_cmd_tab[tab_i].usage->options[i].opt_msg);
		}
	printf("\n\n");


	return(0); /*exit "util help" */
	}

/*
*	The utility subcommands are diverse enough they do not have common options.
*	Subcommands themselvs are required to do the getopts parsing if they need it.
*
*	The util subcommand is in the second argument. Search the command table.
*/
not_found = 1;
tab_i = 0;
while(not_found)
    {
    not_found = strcmp(util_cmd_tab[tab_i].test_name, argv[1]);
    if(not_found)
        ++tab_i;
    if (util_cmd_tab[tab_i].test_name == NULL)	
		break;
    }

if(not_found)
	{	
	sprintf(mbuf, "util: Error...subcommand /'%s/' is not found.\n", argv[1]);
	printf("%s", mbuf);
	return(E_UTIL_SUBCMD_NOT_FOUND);
	}
else
	{
	/*
	*	The util subcommand has been found, remove the "util" from the arg 
	*	list before calling the subcommand. 
	*/
	rc = util_cmd_tab[tab_i].test_func(argc, argv);
	}

return(rc);					
}/*end FUNCTION: do_util() */

/* -------------------------------------------------------------------- */

U_BOOT_CMD(
	util,	10,	1,	do_util,
	"util    - run a Dell developed utility subcommand.\n",

	"\t enter \"util\" for a list of subcommands available\n"
	"\t util help <subcommand>  -display detailed description of subcommand \n"
);


#endif	/* CONFIG_CMC_10G */


